/*  
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_a2l.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing the low level parsing API for A2L files.
 *
 * File containing the low level parsing API for A2L files.
*/
#ifndef CS_A2L_H
#define CS_A2L_H

#include <inttypes.h>
#include "cs_defines.h"
#include "cs_debug.h"
#include "cs_parser.h"

/** \def CS_A2L_SIGNALMAX
    Maximal possible signal definition
*/
#define CS_A2L_SIGNALMAX 1024

/** \def CS_A2L_VERSION_MAJOR
    Major Version of A2L stack
*/
#define CS_A2L_VERSION_MAJOR 0

/** \def CS_A2L_VERSION_MINOR
    Minor Version of A2L stack
*/
#define CS_A2L_VERSION_MINOR 1

/** \def CS_A2L_VERSION
    Get version word from major minor version
*/
#define CS_A2L_VERSION(a, b) (((a) << 8) + (b))

/** \def CS_A2L_VERSION_MAGIC
    A2L interface version word.
 */
#define CS_A2L_VERSION_MAGIC CS_A2L_VERSION(CS_A2L_VERSION_MAJOR, CS_A2L_VERSION_MINOR)

/** \def CS_A2L_CHECKVERSION
    Check A2L stack compatibility
    \retval 0 if stack is compatible
    \retval 1 if stack is incompatible
*/
#define CS_A2L_CHECKVERSION ((CS_A2L_VERSION_MAGIC == cs_a2l_libgetversion())?0:1)

/** \def CS_A2L_RECORD_SIZEMAX
    Maximal record item size
*/
#define CS_A2L_RECORD_SIZEMAX 32

/** \def CS_A2L_MS_SIZEMAX
    Maximal memory segment count
*/
#define CS_A2L_MS_SIZEMAX 8

/** \def CS_A2L_AM_SIZEMAX
    Maximal Address mapping count
*/
#define CS_A2L_AM_SIZEMAX 8

/** \enum cs_a2l_bo_t
    Byteorder definition
*/
typedef enum{
  A2L_BO_INTEL = 0x00, /**< Byteorder Intel*/
  A2L_BO_MOTOROLLA = 0x01, /**< Byteorder Motorolla*/
  A2L_BO_UNKNOWN = 0x02    /**< Not defined*/
} cs_a2l_bo_t;

/** \enum cs_a2l_feature_t
    Flag if optional features are supported
*/
typedef enum{
    A2L_FEATURE_UNKNOWN = 0,
    A2L_FEATURE_DISABLED,  
    A2L_FEATURE_ENABLED
}cs_a2l_feature_t;

/** \enum cs_a2l_ts_unit_t
    Timestamp units
*/
typedef enum{
  A2L_TS_UNIT_1NS = 0x00,  /**< Timestamp unit 1 ns*/
  A2L_TS_UNIT_10NS = 0x01,  /**< Timestamp unit 10 ns*/
  A2L_TS_UNIT_100NS = 0x02,  /**< Timestamp unit 100 ns*/
  A2L_TS_UNIT_1US = 0x03,  /**< Timestamp unit 1 micro s*/
  A2L_TS_UNIT_10US = 0x04,  /**< Timestamp unit 10 micro s*/
  A2L_TS_UNIT_100US = 0x05,  /**< Timestamp unit 100 micro s*/
  A2L_TS_UNIT_1MS = 0x06,  /**< Timestamp unit 1 ms*/
  A2L_TS_UNIT_10MS = 0x07,  /**< Timestamp unit 10 ms*/
  A2L_TS_UNIT_100MS = 0x08,  /**< Timestamp unit 100 ms*/
  A2L_TS_UNIT_1S = 0x09,  /**< Timestamp unit 1 s*/
  A2L_TS_UNIT_1PS = 0x0A,  /**< Timestamp unit 1 picos*/
  A2L_TS_UNIT_10PS = 0x0B,  /**< Timestamp unit 10 picos*/
  A2L_TS_UNIT_100PS = 0x0C,  /**< Timestamp unit 100 picos*/
}cs_a2l_ts_unit_t;

/** \enum cs_a2l_ag_type_t
    Address granulation type
*/
typedef enum{
    A2L_AG_BYTE = 0x00,           /**< Byte granulation*/
    A2L_AG_WORD = 0x02,           /**< Word granulation*/
    A2L_AG_DWORD = 0x04,          /**< Double Word granulation*/
    A2L_AG_RESERVED = 0x06,       /**< Reserved granulation*/
}cs_a2l_ag_t;

/** /enum cs_a2l_chksum_type_t
    Checksum type
*/
typedef enum{
  A2L_CHKSUM_TYPE_UNKNOWN = 0,    /**< Type is unknown*/
  A2L_CHKSUM_TYPE_ACTIVEPAGE,     /**< Active page*/
}cs_a2l_chksum_type_t;

/** \struct cs_a2l_chksum_t
    Storage structure for checksum calculation
*/
typedef struct{
  uint32_t algorythm;             /**< Algorythm identifier*/
  uint32_t limit;                 /**< Limit*/
  cs_a2l_chksum_type_t type;     /**< Page type*/
}cs_a2l_chksum_t;

/** \struct cs_a2l_list_t
    List of multiple search sections
*/
typedef struct{
  int no_result;                  /**< Number of sections found*/
  cs_parser_result_t *result;     /**< Result data list, dynamically allocated*/
  char **resultname;              /**< Result names, dynamically allocated*/
  char **resultlabel;             /**< Result labels, dynamically allocated*/
  char sectiontype[CS_PARSER_CHARBUF]; /**< Storage of sectiontype*/
}cs_a2l_list_t;

/** \enum cs_a2l_ms_type_t
    Memory segment type
*/
typedef enum{
  A2L_MS_TYPE_UNKNOWN = 0,             /**< Unknown type*/
  A2L_MS_TYPE_DATA,                    /**< Data segment, R+W access*/
  A2L_MS_TYPE_CODE,                    /**< Code segment, RO access*/
  A2L_MS_TYPE_VARIABLES,               /**< Variables segment, R+W access*/
}cs_a2l_ms_type_t;

/** \enum cs_a2l_ms_store_t
    Memory segment storage type
*/
typedef enum{
  A2L_MS_STORE_UNKNOWN = 0,           /**< Unknwon storage*/
  A2L_MS_STORE_FLASH,                 /**< Storage type flash*/
  A2L_MS_STORE_RAM,                   /**< Storage type RAM*/
}cs_a2l_ms_store_t;

/** \enum cs_a2l_ms_place_t
    memmory segment placement
*/
typedef enum{
  A2L_MS_PLACE_UNKNOWN = 0,           /**< Unknown placement*/
  A2L_MS_PLACE_EXTERN,                /**< External placement*/
  A2L_MS_PLACE_INTERN,                /**< Internal placement*/
}cs_a2l_ms_place_t;

/** \struct cs_a2l_ms_t
    Definition of an memory segment
*/
typedef struct{
  char name[CS_CHARBUF];          /**< Segment name*/
  char comment[CS_CHARBUF];       /**< Segment comment*/
  cs_a2l_ms_type_t type;          /**< Segment type*/   
  cs_a2l_ms_store_t store;        /**< Segment storage*/
  cs_a2l_ms_place_t place;        /**< Segment placement*/
  uint32_t start;                 /**< Start adress of segment*/
  uint32_t len;                   /**< Length of segment*/
}cs_a2l_ms_t;

/** \struct cs_a2l_am_t
    Address mapping information
*/
typedef struct{
  uint32_t start;                 /**< Start address of mapping*/
  uint32_t len;                   /**< Length of mapping area*/
  uint32_t offset;                /**< Offset of mapped address*/
}cs_a2l_am_t;

/** \struct cs_a2l_modpar_t
    Optional parameter
*/
typedef struct{
  int ms_count;                        /**< Number of defined Memory segments*/
  cs_a2l_ms_t ms[CS_A2L_MS_SIZEMAX];   /**< Memory segment storage array*/
  int am_count;                        /**< Number of defined address mapping areas*/
  cs_a2l_am_t am[CS_A2L_AM_SIZEMAX];   /**< Address mapping storage array*/
}cs_a2l_modpar_t;

/** \enum cs_a2l_compu_tab_type_t
    Type of table
*/
typedef enum{
  A2L_COMPU_TAB_TYPE_UNKNOWN = 0,  /**< Unknown type*/
  A2L_COMPU_TAB_TYPE_VTAB,         /**< Value to string tab*/
  A2L_COMPU_TAB_TYPE_VTAB_RANGE    /**< Value range to string*/
}cs_a2l_compu_tab_type_t;

/** \struct cs_a2l_compu_vtab_elem_t
    Signal lookup table element
*/
typedef struct{
  int id_start;                     /**< ID of start element*/
  int id_end;                       /**< ID of end element*/
  char label[CS_CHARBUF];           /**< Label of lement*/
}cs_a2l_compu_vtab_elem_t;

/** \struct cs_a2l_compu_vtab_t
    Signal lookup table 
*/
typedef struct{
  char name[CS_CHARBUF];            /**< Name or identifier of the method*/
  char comment[CS_CHARBUF];         /**< Optional cmooment of the method*/
  cs_a2l_compu_tab_type_t ttype;    /**< Table type*/
  int number;                       /**< Number of lookup entries*/
  cs_a2l_compu_vtab_elem_t *elems;  /**< Element array*/
  char edefault[CS_CHARBUF];        /**< Default label*/
}cs_a2l_compu_vtab_t;

/** \enum cs_a2l_compu_method_algo_t
    Algorythm type for compu method
*/
typedef enum{
  A2L_CM_UNDEF = 0,              /**< Method not defined*/
  A2L_CM_IDENT,                  /**< No function defined*/
  A2L_CM_FUNCTION,               /**< External function*/
  A2L_CM_LINEAR,                 /**< Linear calculation, Scale and offset*/
  A2L_CM_RAT,                    /**< Rational function*/
  A2L_CM_TAB_INTP,               /**< Table look up with interpolation*/
  A2L_CM_TAB_NOINTP,             /**< Table look up without interpolation*/
  A2L_CM_TAB_VERB,               /**< Verbal Table look up*/
}cs_a2l_compu_method_algo_t;


/** \struct cs_a2l_compu_method_t
    Storage for value calculation
*/
typedef struct{
  char name[CS_CHARBUF];            /**< Name or identifier of the method*/
  char comment[CS_CHARBUF];         /**< Optional cmooment of the method*/
  char unit[CS_CHARBUF];            /**< Unit of value*/
  char format[CS_CHARBUF];          /**< Output format string*/
  double coeff[CS_CHARBUF];         /**< Coeffizient value array for signal calculation*/
  char formular[CS_CHARBUF];        /**< Formular in string format*/
  char compuvtab[CS_CHARBUF];       /**< optional CompuVTab name*/
  cs_a2l_compu_method_algo_t algo;  /**< Basic algorythm type of the method*/
  cs_a2l_compu_vtab_t tab;          /**< Lookuptable if type TAB_VERB*/
  double min;                       /**< Minimal allowed value*/
  double max;                       /**< Maximal allowed value*/
}cs_a2l_compu_method_t;

/** \enum cs_a2l_data_type_t
    Data interpretation type
*/
typedef enum{
  A2L_UBYTE = 0,           /**< Unsigned char type*/
  A2L_SBYTE,               /**< Signed char type*/
  A2L_UWORD,               /**< Unsigned short type*/
  A2L_SWORD,               /**< Signed short type*/
  A2L_UINT,                /**< Unsigned int type*/
  A2L_SINT,                /**< Signed int type*/
  A2L_UINT64,              /**< Long unsigned int type*/
  A2L_SINT64,              /**< Long signed int type*/
  A2L_FLOAT,               /**< Singe precision floattype*/
  A2L_FLOAT64,             /**< Double precision floattype*/
}cs_a2l_data_type_t;

/** \struct cs_a2l_measurement_t
    Storage structure for signal dependent information
*/
typedef struct{
  char name[CS_CHARBUF];/**< Signal name*/
  char comment[CS_CHARBUF];/**< Comment name*/
  char conv[CS_CHARBUF];/**< Compu method name*/
  char format[CS_CHARBUF]; /**< Format description*/
  cs_a2l_bo_t BO;          /**< Byte order if overwritten in signal section*/
  int daq_idx;             /**< DAQ idx to register at*/
  int odt_idx;             /**< ODT idx to register at*/
  int entry_idx;           /**< Entry idx to register at*/
  int ev_idx;              /**< Event handler to register at*/
  int prescaler;           /**< Requested prescaler factor*/
  int size;                /**< Size of entry in AG*/
  double min;                 /**< Minimal value*/
  double max;                 /**< Maximal value*/
  int resolution;          /**< Signal resolution*/
  int aqurancy;            /**< Signal aqurancy*/
  uint8_t bitoffset;       /**< For Bit values, bit offset*/
  uint32_t address;        /**< Address of entry*/
  uint8_t extension;       /**< Address extension of entry*/
  double *exvalue;         /**< External value reference,
			      to store last received value
			   */
  double *exts;            /**< External timestamp reference, 
			      to store last received timestamp value
			   */
  void (*extcallback)(double value, double ts, void *priv); /** External callback, 
								is triggered when a value was received
							    */
  void *extcallbackpriv;      /**< External callback private data structure*/
  cs_a2l_compu_method_t cm;/**< Calculation method to convert value*/  
  cs_a2l_data_type_t dtype;/**< Data type used for conversion*/
}cs_a2l_measurement_t;

/** \enum cs_a2l_record_type_t
    Record type
*/
typedef enum{
  A2L_RECORD_TYPE_UNKNOWN = 0,
  A2L_RECORD_TYPE_FNC,
}cs_a2l_record_type_t;

/** \enum cs_a2l_record_access_t
    Record access
*/
typedef enum{
  A2L_RECORD_ACCESS_UNKNOWN = 0,
  A2L_RECORD_ACCESS_DIRECT,
}cs_a2l_record_access_t;

/** \enum cs_a2l_record_orientation_t
    Record orientation
*/
typedef enum{
  A2L_RECORD_ORIENT_UNKNOWN = 0,
  A2L_RECORD_ORIENT_COLUMN,
  A2L_RECORD_ORIENT_ROW,
  A2L_RECORD_ORIENT_INDEXINCR
}cs_a2l_record_orientation_t;

/** \struct cs_a2l_record_t
    Storage structure for record layout
*/
typedef struct{
  char name[CS_CHARBUF];                        /**< Record name*/
  int size;                                     /**< Size of record*/
  cs_a2l_record_type_t rtype[CS_A2L_RECORD_SIZEMAX]; /**< Record type*/
  cs_a2l_data_type_t dtype[CS_A2L_RECORD_SIZEMAX]; /**< Data type*/
  cs_a2l_record_access_t access[CS_A2L_RECORD_SIZEMAX];   /**< Access type*/
  cs_a2l_record_orientation_t orientation[CS_A2L_RECORD_SIZEMAX]; /**< Orientation*/
  int position[CS_A2L_RECORD_SIZEMAX];                  /**< Element position*/
}cs_a2l_record_t;

/** \enum cs_a2l_characteristic_type_t
    Basis characteristic type
*/
typedef enum{
  A2L_CHARACTERISTIC_TYPE_UNKNOWN = 0,     /**< Type is unknown*/
  A2L_CHARACTERISTIC_TYPE_VALUE,           /**< Value type*/
  A2L_CHARACTERISTIC_TYPE_VALBLK,          /**< Block type*/
  A2L_CHARACTERISTIC_TYPE_CURVE,           /**< Curve type*/
  A2L_CHARACTERISTIC_TYPE_ASCII,           /**< Curve type*/
}cs_a2l_characteristic_type_t;

/** \enum cs_a2l_characteristic_access_t
    Basis characteristic access type
*/
typedef enum{
  A2L_CHARACTERISTIC_ACCESS_RW = 0,     /**< Read and write*/
  A2L_CHARACTERISTIC_ACCESS_RO          /**< Read only*/
}cs_a2l_characteristic_access_t;

/** \struct cs_a2l_chracteristic_t
    Storage structure for characteristics
*/
typedef struct{
  char name[CS_CHARBUF];   /**< Signal name*/
  char comment[CS_CHARBUF];/**< Comment name*/
  char unit[CS_CHARBUF];   /**< Unit*/
  char conv[CS_CHARBUF];   /**< Compu method name*/
  char recordname[CS_CHARBUF]; /**< Record name*/
  char format[CS_CHARBUF]; /**< Format description*/
  cs_a2l_characteristic_type_t type; /**< Value type*/
  cs_a2l_characteristic_access_t access; /**< Access mode*/
  cs_a2l_bo_t BO;          /**< Byte order if overwritten in signal section*/
  double maxdiff;          /**< Max difference*/
  double min;              /**< Minimal value*/
  double max;              /**< Maximal value*/
  uint32_t address;        /**< Address of entry*/
  uint8_t extension;       /**< Address extension of entry*/
  cs_a2l_compu_method_t cm;/**< Calculation method to convert value*/  
  cs_a2l_record_t record; /**< Record description*/
  cs_a2l_ag_t AG;         /**< Adress granulation, copied from cs_a2l_t*/
}cs_a2l_characteristic_t;

/** \struct cs_a2l_t
    Storage structure for A2L-file parsing
*/
typedef struct{
  FILE *fs;               /**< FILE stream*/
  char *filename;         /**< Filename to parse*/
  long filelen;           /**< File length*/
  void *map;              /**< File map pointer*/
  char *buffer;           /**< Cleaned up raw file content*/        
  cs_parser_result_t project; /**<Storage for defined PROJECT section*/
  cs_parser_result_t module; /**<Storage for defined MODULE section*/
  cs_parser_result_t iface; /**<Storage for defined IF_DATA section*/
  cs_parser_result_t protocol; /**<Storage for defined PROTOCOL section*/
  int signalcounter;      /**< Counter of valid signals*/
  cs_a2l_measurement_t signals[CS_A2L_SIGNALMAX]; /**< Storage array for selected signals*/
  cs_a2l_ag_t AG;         /**< Adress granulation*/
  cs_a2l_bo_t BO;         /**< Byte order defines in MOD_COMMON*/
  cs_a2l_modpar_t modpar; /**< Optional paremeter*/
  char ifacename[CS_CHARBUF]; /**< Interface definition name, ex: ASAP1B_CCP*/
}cs_a2l_t;


#ifdef __cplusplus
extern "C" {
#endif

  /** Set debug level
      \retval Active Debuglevel
  */  
  cs_debug_level_t cs_a2l_debuglevel(
				     cs_debug_level_t level  /**< Debug level*/
				     );

  /** Set debugging callback routine
      Default: Function registered which prints debugging to STDOUT
  */
  void cs_a2l_debug_setcallback(
				void *ctx,                             /**< Callback private data*/
				void (*callb)(void *ctx,
					      cs_debug_sub_t sub,
					      cs_debug_level_t dl,
					      const char *func,
					      const int line,
					      const char *fmt,
					      va_list ap) /**< Callback routine*/
				);
  
  /** Get A"L stack version bytes*/
  uint16_t cs_a2l_libgetversion();

  /** Get number of bytes of value representation depending on AG
      \retval number of bytes
  */
  int cs_a2l_get_ag_bytes(
			  cs_a2l_ag_t AG           /**< Data Granulation type*/
			  );
  
  /** Convert string to type information
      \retval 1 Failure
   */
  int cs_a2l_StringToType(
			  char* typestring,        /**< Type string*/
			  cs_a2l_ag_t AG,          /**< Adress granulation*/
			  cs_a2l_data_type_t *dtype, /**< Data type feedback*/
			  int *agsize              /**< Data type size in AG bytes*/
			  );

  /** Get Type size in AG
      \retval 0 Failure
  */
  int cs_a2l_TypeToAG(
		      cs_a2l_data_type_t dtype     /**< Data type*/
		      );

  /** Printout section list*/
  void cs_a2l_printlist(
			cs_a2l_list_t *list        /**< List pointer*/
			);

  /** Init working structure*/
  void cs_a2l_init(
		   cs_a2l_t *data  /**< Working storage structure*/
		   );


  /** Open A2L-File and tries to map it to memory
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_a2l_open(
		  cs_a2l_t *data,  /**< Working storage structure*/
		  char *filename       /**< Filename to open*/
		  );


  /** Close A2L-File and reset working structure
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_a2l_close(
		   cs_a2l_t *data  /**< Working storage structure*/
		   );


  /** Search specific section in A2L file
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_a2l_search_section(
			    cs_parser_result_t *result, /**< Regex data structure pointer*/
			    char *sectionname,   /**< Section ident to search for, may not be NULL, search for \begin IDENT ... \end IDENT*/
			    char *sectionsubname /**< Section Name properties, may be NULL, search for \begin IDENT NAME ... \end IDENT*/
			    );

  /** Get ASAP2 Version from file*/
  int cs_a2l_getversion(
			cs_a2l_t *data,   /**< Working storage structure*/
			int *major,           /**< Version Major*/
			int *minor            /**< Version Minor*/
			);

  /** Search for sections and list them in a list
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_a2l_list_sections(
			   cs_parser_result_t *result,  /**< Area to search in*/
			   cs_a2l_list_t *list,      /**< Section list storage pointer*/
			   char *sectiontype             /**< Section type to search*/
			   );

  /** Free list
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_a2l_list_free(
		       cs_a2l_list_t *list    /**< Section list storage pointer*/
		       );

  /** Get specific Project section
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_a2l_GetProject(
		       cs_a2l_t *data,  /**< Working storage structure*/
		       char * project   /**< Project name*/
		       );

  /** Get specific Module section
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_a2l_GetModule(
		       cs_a2l_t *data,  /**< Working storage structure*/
		       char * module        /**< Module name*/
		       );

  /** Get Common system settings
      \retval 0 No failure
      \retval 1 failure
  */  
  int cs_a2l_GetModCommon(
			  cs_a2l_t *data    /**< Working storage structure*/
			  );

  /** Get Interface data section from module section
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_a2l_GetInterface(
			  cs_a2l_t *data,  /**< Working storage structure*/
			  char * iface         /**< Interface name (XCP,...)*/
			  );

  /** Get MOD_PAR information
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_a2l_GetModPar(
		       cs_a2l_t *data,          /**< Working storage structure*/ 
		       char * name              /**< Section name*/
		       );
  
  /** Recalculate address based on MOD_PAR ADDRESS_MAPPING
      \retval Recalculated address
  */
  uint32_t cs_a2l_CalcAddress(
			      cs_a2l_t *data,   /**< Working storage structure*/
			      uint32_t address  /**< Memory address*/
			      );

  /** Get MEASUREMENT information
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_a2l_GetMeasurement(
			    cs_a2l_t *data,                /**< Working storage structure*/
			    cs_a2l_measurement_t *me,          /**< Signal storage structure*/
			    char *name                         /**< Signal name to look for*/
			    );

  /** Get compu method information
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_a2l_GetCompuMethod(
			    cs_a2l_t *data,         /**< Working storage structure*/
			    cs_a2l_compu_method_t *cm,  /**< Compu method storage structure*/
			    char *name                  /**< Name of compu method*/
			    );
  /** Get RECORD_LAYOUT information
      \retval 1 Failure
  */
  int cs_a2l_GetRecord(
		       cs_a2l_t *data,              /**< Working storage structure*/
		       cs_a2l_record_t *record,     /**< Record data storage*/
		       char *name                   /**< Record to search for*/
		       );

  /** Get CHARACTERISTIC information
      \retval 1 Failure
  */
  int cs_a2l_GetCharacteristic(
			       cs_a2l_t *data,              /**< Working storage structure*/
			       cs_a2l_characteristic_t *ch, /**< Characteristic storage structure*/
			       char *name                   /**< Characteristic to search for*/
			       );

  /** Reset signal information structure
   */
  void cs_a2l_reset_measurement(
				cs_a2l_measurement_t *me  /**< Signal information structure*/
				);

  /** Reset compu method structure
   */
  void cs_a2l_reset_compumethod(
				cs_a2l_compu_method_t *cm  /**< Compu method structure*/
				);

  /** Reset record structure
   */
  void cs_a2l_reset_record(
			   cs_a2l_record_t *record       /**< Record data structure*/
			   );

  /** Reset characteristic structure
   */
  void cs_a2l_reset_characteristic(
				   cs_a2l_characteristic_t *ch /**< Characteristic storage structure*/
				   );

  /** Test if flag name exist
      \retval 1 Failure or does not exist
      \retval 0 Flag exists
  */
  int cs_a2l_getflag(
		     cs_parser_result_t result,      /**< Area to search in*/
		     char *name                      /**< Flag name*/
		     );

  /** Get double value vector
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_a2l_getdoublevaluevector(
				  cs_parser_result_t result,  /**< Area to search in*/
				  char *name,                 /**< Value name*/
				  double *value,              /**< Vector store array*/
				  int lenmax                  /**< Max array size*/
				  );

  /** Get unit32 value
      \retval result
  */
  uint32_t cs_a2l_getuintvalue(
			       cs_parser_result_t result,      /**< Area to search in*/
			       char *name,                     /**< Value name*/
			       uint8_t *error                  /**< Error flag 0=Success, 1=Failure*/
			       );

  /** Get string value
      \retval 1 Failure
  */
  int cs_a2l_getstringvalue(
			    cs_parser_result_t result,         /**< Area to search in*/
			    char *name,                        /**< Name of variable*/
			    char* value,                       /**< Value storage buffer*/
			    int maxlen                         /**< Maximal buffer len*/
			    );

  /** Parse time definition
   */
  void cs_a2l_GetTsUnit(
			cs_parser_result_t ltime,                /**< Area to search in*/
			cs_a2l_ts_unit_t *value                  /**< Result value*/
			);
  
  /** Search for Section,optional with label
      \retval 0 Ok
      \retval 1 Not found or failure
  */
  int cs_a2l_getsectionheader(
			      cs_parser_result_t *result,      /**< Area to search in*/
			      char *name,                      /**< Section name*/
			      char *label                      /**< Optional section Label, othercase NULL*/
			      );
  
  /** Print out MOD_PAR information to STDOUT
   */
  void cs_a2l_print_modpar(
			   cs_a2l_modpar_t *mp                /**< MOD_PAR structure*/
			   );

  /** Print out compu method information to STDOUT
   */
  void cs_a2l_print_compumethod(
				cs_a2l_compu_method_t *cm /**< Compu method storage structure*/
				);

  /** Print out lookup table
   */
  void cs_a2l_print_compuvtab(
			      cs_a2l_compu_vtab_t *tab    /**< Lookuptable storage*/
			      );
  
  /** Print signal information to STDOUT
   */
  void cs_a2l_print_measurement(
				cs_a2l_measurement_t *me /**< Signal information storage structure*/
				);

  /** Print record information
   */
  void cs_a2l_print_record(
			   cs_a2l_record_t *re           /**< Record information structure*/
			   );
  
  /** Print characteristic information
   */
  void cs_a2l_print_characteristic(
				   cs_a2l_characteristic_t *ch /**< Characteristic information structure*/
				   );

  /** Print timestamp unit to STDOUT
   */
  void cs_a2l_print_ts(
		       cs_a2l_ts_unit_t ts_unit,            /**< Timestamp unit*/
		       char *pretxt                         /**< Pretext*/
		       );


  /** Create Cumpu method manually
      \retval 1 Failure
  */
  int cs_a2l_compu_method_linear_create(
					cs_a2l_compu_method_t* cm, /**< Storage structure*/
					char *name,                /**< Name*/
					char *label,               /**< Label*/
					char *unit,                /**< Unit*/
					char *format,              /**< Format*/
					double scale,              /**< Scale factor*/
					double offset              /**< Offset factor*/
					);


  /** Calculate compu method raw to double
      \retval 1 Failure
  */
  int cs_a2l_compu_method_r2d(
			      cs_a2l_compu_method_t cm,           /**< Compu method*/
			      double rawvalue,                    /**< Raw value*/
			      double *value                       /**< Storage pointer*/
			      );

  /** Calculate compu method double to raw
      \retval 1 Failure
  */
  int cs_a2l_compu_method_d2r(
			      cs_a2l_compu_method_t cm,           /**< Compu method*/
			      double *rawvalue,                   /**< Raw value storage*/
			      double value                        /**< Value*/
			      );

  /** Convert raw buffer to double value
      \retval 1 Failure
  */
  int cs_a2l_convert_to_double(			
			       cs_a2l_bo_t endianess,                  /**< Remote ECU Byteorder 1 => Motorolla, 0 => Intel*/
			       uint8_t *udata,                     /**< Raw data pointer*/
			       int size,                           /**< Raw data size*/
			       cs_a2l_compu_method_t cm,           /** Compu method description*/
			       cs_a2l_data_type_t dtype,           /**< Data conversion type*/
			       double *value                       /**< Return pointer*/
						);

  /**Convert double value to raw buffer
     \retval 1 Failure
  */
  int cs_a2l_convert_from_double(			
				 cs_a2l_bo_t endianess,                /**< Remote ECU Byteorder 1 => Motorolla, 0 => Intel*/
				 uint8_t *udata,                   /**< Raw data pointer*/
				 int size,                         /**< Raw data size*/
				 cs_a2l_compu_method_t cm,         /** Compu method description*/
				 cs_a2l_data_type_t dtype,         /**< Data conversion type*/
				 double value                      /**< Value to convert*/
							);


  /**Convert label to value
     \retval 1 Failure
  */
  int cs_a2l_characteristic_l2d(
				cs_a2l_characteristic_t *ch,       /**< Characteristic storage structure*/
				char *label,                       /**< Label to convert*/
				int len,                           /**< Label length*/
				double *value                      /**< Converted value*/
				);

  /**Convert value to label
     \retval 1 Failure
  */
  int cs_a2l_characteristic_d2l(
				cs_a2l_characteristic_t *ch,       /**< Characteristic storage structure*/
				char *label,                       /**< Converted value*/
				int *len,                          /**< Max label length*/
				double value                       /**< Value to convert*/
				);

  /** Check and modify value to limits
      \retval 1 Failure
  */
  int cs_a2l_characteristic_limit(
				  cs_a2l_characteristic_t *ch,     /**< Characteristic storage structure*/
				  double *value                    /**< Value to check*/
				  );

#ifdef __cplusplus
    }
#endif


#endif
