/*  XCP CAN Stack for SocketCAN
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_ccp.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing the CCP low level API.
 *
 * This file contain the CCP low level API. It is based on ASAM CCP specification.
 * The implementation is not complete and based on the Spec 2.1 version.
 * Actual focus is running DAQ signal handling.
 */

#ifndef CS_CCP_H
#define CS_CCP_H

#include <inttypes.h>

#include "cs_ccp_a2l.h"
#include "cs_socketcan.h"
#include "cs_ccp_codes.h"


/** \struct cs_ccp_odt_entry
    ODT entry
*/
typedef struct{
  uint8_t data[CCP_DATALENGTH];/**< Raw data storage*/
  cs_a2l_measurement_t me; /**< Signal dependent information*/
}cs_ccp_odt_entry_t;

/** \struct cs_ccp_odt_t
    ODT structure
*/
typedef struct{
    int no_entry;              /**< Number of ODT entries in list*/
    cs_ccp_odt_entry_t entry[CCP_ODTENTRYMAX]; /**< Storage pointer*/
    int datasize;              /**< Sum of data bytes*/
    double last;               /**< Timestamp of last received ODT*/
}cs_ccp_odt_t;

/** \struct cs_ccp_daq_t
    DAQ structure
*/
typedef struct{
    int no_odt;                /**< Number of ODTs in DAQ*/
    int no_active;             /**< Number of ODT used in configuration*/
    uint8_t *odt_active;           /**< Array of used ODT ids*/
    cs_ccp_odt_t *odt;         /**< Storage pointer*/
    int prescaler;             /**< Prescaler*/
    int ev;                    /**< Evcent channel number*/
    int resume;                /**< 1 => DAQ is prepared for resume mode*/
    int running;               /**< 1 => DAQ is running*/
    int selected;              /**< 1 => DAQ is selected*/
    uint8_t first_pid;        /**< First PID of DTO*/
    uint32_t canid;            /**< DAQ CAN id*/
    uint16_t pid;              /**< actual PID of DTO*/
    double ts_ref;             /**< Timestamp of first received ODT*/ 
    uint32_t ts_old;           /**< Storage for old received timestamp*/
    uint64_t counter;          /**< Recv or send counter*/
    double period;             /**< Sampling period time in s*/
    double ts_expect;          /**< Expected timestamp for next ODT package*/
}cs_ccp_daq_t;

/** \struct cs_ccp_daq_list_t
    DAQ list structure
*/
typedef struct{
    cs_ccp_daq_t daq[CCP_DAQMAX];        /**< Storage pointer*/
    uint8_t daq_active[CCP_DAQMAX];      /**< Active DAQs*/
    int no_active;            /**< Number of active DAQs*/
    double clock_shift;       /**< Clock shift between master and slave*/
    double clock_base;        /**< Clock base timestamp of slave*/
    double daq_fac_recvtimeout; /**< Time factor for DAQ timeout detection*/
}cs_ccp_daq_list_t;


/** \struct cs_ccp_cmd_entry_t
    Storage structure for cmd specific information
*/
typedef struct{
  uint32_t timeout;      /**< Timeout to ack*/
  uint8_t optional;      /**< Is optional*/
  uint8_t isoperational; /**< Is and optional cmd but activated by configuration*/
  uint8_t supported;     /**< Marker for recieve frame*/
  char name[CS_CHARBUF];            /**< Name of cmd*/
}cs_ccp_cmd_t;

/** \enum cs_ccp_return_category_t
    Error code category
*/
typedef enum{
  CCP_RETURN_CATEGORY_NOERROR = 0,  /**< No error*/
  CCP_RETURN_CATEGORY_TIMEOUT,      /**< Retry*/
  CCP_RETURN_CATEGORY_C0,           /**< Warning*/
  CCP_RETURN_CATEGORY_C1,           /**< Wait*/
  CCP_RETURN_CATEGORY_C2,           /**< Reinit*/
  CCP_RETURN_CATEGORY_C3,           /**< Terminate*/
}cs_ccp_return_category_t;

/** \struct cs_ccp_return_entry_t
    Storage structure for return code specific information
*/
typedef struct{
  uint8_t source;      /**< Return code source, 0=> CCP Standard, >0 CCP API Stack*/
  cs_ccp_return_category_t category; /**< Return code category*/
  char name[CS_CHARBUF];           /**< Name of return code*/
}cs_ccp_return_t;

/** \struct cs_ccp_pkg_t
    Package transfer structure
*/
typedef struct{
    uint8_t data[CCP_DATALENGTH];           /**< Package data segment*/
    int dl;                                 /**< Package data segment length*/
    struct timeval tv;                      /**< timestamp*/
}cs_ccp_pkg_t;

/** \enum cs_ccp_indicator_t
    Event indicator bitmask
*/
typedef enum{
  CCP_INDICATOR_NONE = 0,           /**< No indication*/
  CCP_INDICATION_KEYREQUEST=0x01,   /**< Key request*/
  CCP_INDICATION_SESSIONSTATUS=0x02,/**< Session status request*/
  CCP_INDICATION_COLDRESTART=0x04,  /**< Cold restart request*/
  CCP_INDICATION_CALINIT=0x08,      /**< CAL data init request*/
  CCP_INDICATION_DAQINIT=0x10,      /**< DAQ data init request*/
  CCP_INDICATION_CODEUPDATE=0x20   /**< Code update request*/
}cs_ccp_indicator_t;

/** Basic CCP working structure*/
typedef struct{
  cs_socketcan_t *socket;          /**< CAN interface socket*/
  cs_ccp_a2l_t ccpa2l;             /**< A2L storage data pointer*/
  uint8_t wc;                      /**< CCP working counter*/
  uint8_t lerror;                  /**< Latest stored error code*/
  char lerrorstr[CS_CHARBUF];      /**< Latest stored error code*/
  struct{
    uint32_t address;              /**< Memory access address*/
    uint8_t extension;             /**< Memory extension*/
  }mta;
  cs_ccp_cmd_t cmd[CCP_CMD_BUF]; /**< Cmd lookup buffer*/
  cs_ccp_return_t error[CCP_ERR_BUF];  /**< Error code decoding table*/
  cs_ccp_indicator_t indicator;    /**< Asynchron indication bitmask*/
  cs_ccp_daq_list_t daqlist;       /**< DAQ management structure*/
  uint8_t resource_availability_mask; /**< Resource availability mask*/
  uint8_t resource_protection_mask; /**< Resource protection mask*/
  uint32_t stim;                    /**< Configure message seperation time in ms, used in API*/
  uint8_t slaveid[CCP_EXCHANGEID_SIZE]; /**< Slave id,read out by cmd exchange id*/
  uint8_t ignore_feedback;         /**< Flag if high ignore slave feedback*/
  int ignore_counter;              /**< Ignore counter*/
}cs_ccp_t;


#ifdef __cplusplus
extern "C" {
#endif

  /** Set debug level
      \retval Active Debuglevel
  */  
  cs_debug_level_t cs_ccp_debuglevel(
				     cs_debug_level_t level  /**< Debug level*/
				     );

  /** Set debugging callback routine
      Default: Function registered which prints debugging to STDOUT
  */
  void cs_ccp_debug_setcallback(
				void *ctx,                             /**< Callback private data*/
				void (*callb)(void *ctx,
					      cs_debug_sub_t sub,
					      cs_debug_level_t dl,
					      const char *func,
					      const int line,
					      const char *fmt,
					      va_list ap) /**< Callback routine*/
				);
  
  /** Get verion magic of CCP stack
      \retval uint16 high byte is major version , low byte is minor version
  */
  uint16_t cs_ccp_getversion();

  /** Get last failure code
      \retval 0 No failure
  */
  uint8_t cs_ccp_getlasterror(
			      cs_ccp_t *data         /**< CCP stack structure*/ 
			      );

  /** Get last failure code string
  */
  char *cs_ccp_getlasterrorstring(
				  cs_ccp_t *data         /**< CCP stack structure*/ 
				  );


  /** Basic init of CCP stack
      \retval 0 No failure
      \retval 1 Failure
  */
    int cs_ccp_init(
		    cs_ccp_t *data         /**< CCP stack structure*/ 
		    );

  /** Close physical layer
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_close(
		     cs_ccp_t *data         /**< CCP stack structure*/ 
		     );

  /** Register callback value and function for ODT entry
      Only for internal usage, normal user should take  cs_ccp_entry_reg_signal
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_entry_reg_value(			
			     cs_ccp_t *data,        /**< CCP stack structure*/
			     uint8_t daq,               /**< DAQ list number*/
			     uint8_t odt,               /**< ODT number*/
			     uint8_t entry,             /**< ODT Entry number*/
			     int size,                  /**< Element size in Bytes*/
			     double *value,         /**< Optional: Feedback value pointer, maybe NULL*/
			     cs_a2l_bo_t BO,        /**< signal byte order*/
			     cs_a2l_compu_method_t cm, /** Compu method description*/
			     cs_a2l_data_type_t dtype, /**< Data conversion type*/
			     double *ts,             /**< Otional: Feedback timestamp counter, maybe NULL*/
			     void (*extcallback)(double value, double ts, void *priv), /**Optional: Receive callback function for new value
											  Arguments: value -> signal value
											  ts    -> signal timestamp
											  priv  -> private data structure,
											  maybe NULL
										       */
			     void *priv              /**< Optional: Private data structure storage pointer, 
							for use in callback function call,
							maybe NULL
						     */
						);


  /** Register callback value and function for ODT entry
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_entry_reg_signal(			
			     cs_ccp_t *data,            /**< CCP stack structure*/
			     uint8_t daq,               /**< DAQ list number*/
			     uint8_t odt,               /**< ODT number*/
			     uint8_t entry,             /**< ODT Entry number*/
			     char *signal,              /**< Signal name*/
			     double *value,             /**< Optional: Feedback value pointer, maybe NULL*/
			     double *ts,                /**< Optional: Feedback timestamp counter, maybe NULL*/
			     void (*extcallback)(double value, double ts, void *priv), /**Optional: Receive callback function for new value
											  Arguments: value -> signal value
											  ts    -> signal timestamp
											  priv  -> private data structure,
											  maybe NULL
										       */
			     void *priv                  /**< Optional: Private data structure storage pointer, 
							for use in callback function call,
							maybe NULL
						     */
						);


  /** Set DAQ sampling period, this is used for timeout detection
      \retval 0 No failure
      \retval 1 Failure
  */      
  int cs_ccp_daq_setperiod(			
			   cs_ccp_t *data,            /**< CCP stack structure*/
			   uint8_t daq,               /**< DAQ id*/
			   double period              /**< period in s*/
						);
  
  /** Set prescaler value for DAQ, used by CCP API
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_daq_setprescaler(			
			    cs_ccp_t *data,           /**< CCP stack structure*/
			    uint8_t daq,              /**< DAQ id*/
			    uint16_t prescaler        /**< Prescaler >= 1*/
						);
  /** Get max ODT id for DAQ, used by CCP API
      \retval 0xFF Failure
  */
  uint8_t cs_ccp_odt_getmaxactive(
				  cs_ccp_t *data,           /**< CCP stack structure*/
				  uint8_t daq               /**< DAQ id*/
				  );

  /** Check for running DAQ entries if there was an receive timeout.
      An timeout happens when the last ODT recv timestamp  plus
      (DAQ period multiplied by fac_timeout) is lesser than actual time
      \retval 0 No timeout
      \retval 1 There was an failure or timeout
  */
  int cs_ccp_daq_istimeout(			
			   cs_ccp_t *data,       /**< CCP stack structure*/
			   double fac_timeout    /**< Timeout factor, has to be >=1.0, < 1.0 the check is deactivated*/
						);

  /** Receive Data DTO, EV, SERV, ...
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_receive(
		     cs_ccp_t *data       /**< CCP stack structure*/  
		     );
  
  /** Print out session status
   */
  void cs_ccp_print_status(
			   uint8_t status  /**< Seesion status*/
			   );
  /** Print slave resources
   */
  void cs_ccp_print_resource(
			     cs_ccp_t *data  /**< CCP stack structure*/
			     );
  
  /** Open and prepare CAN interface
      \retval 0 No failure
      \retval 1 Failure
   */
  int cs_ccp_can_open(
		      cs_ccp_t *data,      /**< CCP stack structure*/ 
		      char *iface          /**< Interface name*/
		      );

  /** Close CAN interface
      \retval 0 No failure
      \retval 1 Failure
   */
  int cs_ccp_can_close(
		       cs_ccp_t *data      /**< CCP stack structure*/ 
		       );
  
  /** If set to zero, every cmd will wait for ack
      If set to one, cmds without data response will skip waiting for slave ack
      Note dangerous
  */
  /*************************************************************************************************/
  void cs_ccp_set_ingnoreresponse(
				  cs_ccp_t *data, 
				  uint8_t ignore
				  );
  /*************************************************************************************************/

  /** Set cmd seperation time
   */
  void cs_ccp_set_stim(
		       cs_ccp_t *data,      /**< CCP stack structure*/
		       uint32_t stim        /**< STIM time in ms*/
		       );
  
  /** Wait with async data processing       
   */
  void cs_ccp_wait_stim(
			cs_ccp_t *data      /**< CCP stack structure*/
			);

  /** Connect to slave
      \retval 0 No failure
      \retval 1 Failure
   */
  int cs_ccp_cmd_connect(
			 cs_ccp_t *data    /**< CCP stack structure*/ 
			 );

  /** Disconnect from slave
      \retval 0 No failure
      \retval 1 Failure
   */
  int cs_ccp_cmd_disconnect(
			    cs_ccp_t *data,        /**< CCP stack structure*/ 
			    uint8_t endofsession   /**< 0=temporary, 1=endofsession*/
			    );
  
  /** Get CCP Version information
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_cmd_ccpversion(
			    cs_ccp_t *data,          /**< CCP stack structure*/
			    uint8_t *masterversion,  /**< Feedback Slave Major version*/
			    uint8_t *releaseversion  /**< Feedback Slave Minor version*/
			    );
  
  /** Get Configuration id and Resource information
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_cmd_exchangeid(
			    cs_ccp_t *data,                         /**< CCP stack structure*/
			    uint8_t *length,                        /**< ID length, for next upload cmd*/
			    uint8_t *qualifier,                     /**< Qualifier*/
			    uint8_t *resource_availability_mask,    /**< Resource availabbility mask*/
			    uint8_t *resource_protection_mask       /**< Resource protection_mask*/
			    );

  /** Get Session status
      See CCP_SSTATUS_*** defines to analyze
      \retval 0 No failure
      \retval 1 Failure
  */ 
int cs_ccp_cmd_getsessionstatus(
				cs_ccp_t *data,    /**< CCP stack structure*/
				uint8_t *status    /**< Slave Status*/
				);

  /** Set memory transfer adress
      \retval 0 No failure
      \retval 1 Failure
   */
  int cs_ccp_cmd_setmta(
			cs_ccp_t *data,             /**< CCP stack structure*/
			uint8_t mtanumber,          /**< 0=default, 1=for move cmd*/
			uint32_t mta,               /**< Memory adress*/
			uint8_t adressextension     /**< Adress extension*/
			);

  /** Download memory block to slave
      Note: Endianess conversion has to be done externally
      \retval 0 No failure
      \retval 1 Failure
  */ 
  int cs_ccp_cmd_dnload(
			cs_ccp_t *data,             /**< CCP stack structure*/
			uint8_t *tdata,             /**< Data pointer*/
			int size                    /**< Size of data array (1-5)*/
			);

  /** Upload memory block to slave
      Note: Endianess conversion has to be done externally
      \retval 0 No failure
      \retval 1 Failure
  */ 
  int cs_ccp_cmd_upload(
			cs_ccp_t *data,             /**< CCP stack structure*/
			uint8_t *tdata,             /**< Data pointer*/
			int size                    /**< Size of data array (1-5)*/
			);
  /** Get DAQ Size
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_cmd_getdaqsize(
			    cs_ccp_t *data,         /**< CCP stack structure*/
			    uint8_t daq,            /**< DAQ id*/
			    uint32_t dto_id,        /**< DTO id*/
			    uint8_t *daq_size,      /**< Return size of DAQ*/
			    uint8_t *pid_first      /**< First PID of DAQ*/
			    );

  /** Set Write pointer for DAQ configuration
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_cmd_setdaqptr(
			   cs_ccp_t *data,          /**< CCP stack structure*/
			   uint8_t daq,             /**< DAQ id*/
			   uint8_t odt,             /**< ODT id*/
			   uint8_t element          /**< Element in ODT Table*/
			   );

  /** Write ODT entry
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_cmd_writedaq(
			  cs_ccp_t *data,          /**< CCP stack structure*/
			  int size,                /**< Size of element (1,2,4)*/
			  uint8_t addr_extension,  /**< Address extension*/
			  uint32_t addr            /**< Address*/
			  );

  /** Start, Stop, Prepare DAQ transmission
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_cmd_startstop(
			   cs_ccp_t *data,           /**< CCP stack structure*/
			   uint8_t mode,             /**< 0=Stop, 1=Start, 2=Prepare for synchron start*/
			   uint8_t daq,              /**< DAQ id*/
			   uint8_t odt_last,         /**< ODT to transmit*/
			   uint8_t event,            /**< Event channel*/
			   uint16_t prescaler        /**< Prescaler >= 1*/
			   );
  /** Start all prepared, Stop all, DAQ transmission
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_cmd_startstopall(
			   cs_ccp_t *data,           /**< CCP stack structure*/
			   uint8_t mode              /**< 0=Stop all, 1=Start selected*/
			   );


  /** Test if slave is available
      \retval 0 Receive Ack from Slave
      \retval 1 Failure or timeout
   */
  int cs_ccp_cmd_test(
		      cs_ccp_t *data       /**< CCP stack structure*/
		      );
  
  /** Build checksum
      \retval 0 Receive Ack from Slave
      \retval 1 Failure or timeout
  */
  int cs_ccp_cmd_buildchksum(
			     cs_ccp_t *data,      /**< CCP stack structure*/
			     int blocksize,       /**< Blocksize in Bytes*/
			     uint32_t *checksum   /**< Checksum*/
			     );

#ifdef __cplusplus
}
#endif

#endif
