/*  CCP CAN Stack for SocketCAN
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_ccp_a2l.h
 * @author Andreas Stewering-Bone
 * @date 5 June 2015
 * @brief File containing the CCP A2L API.
 *
 * This file contain the CCP A2L API.
*/

#ifndef CS_CCP_A2L_H
#define CS_CCP_A2L_H

#include <pcre.h>

#include "cs_a2l.h"
#include "cs_ccp_codes.h"

/** \enum cs_ccp_a2l_daq_mode_t
    CCP DAQ transfer mode
*/
typedef enum{
  CCP_DAQ_MODE_UNKNOWN = 0,  /**< Unknown*/
  CCP_DAQ_MODE_BURST         /**< DAQ mode BURST*/
}cs_ccp_a2l_daq_mode_t;

/** \enum cs_ccp_a2l_consistency_t
    Data consistency mode
*/
typedef enum{
  CCP_CONSISTENCY_UNKNOWN = 0,  /**< Unknown*/
  CCP_CONSISTENCY_DAQ,          /**< DAQ level*/
  CCP_CONSISTENCY_ODT           /**< ODT level*/
}cs_ccp_a2l_consistency_t;

/** \enum cs_ccp_a2l_addrextension_t
    Address extension mode
*/
typedef enum{
  CCP_ADDREXTENSION_UNKNOWN = 0,  /**< Unknown*/
  CCP_ADDREXTENSION_DAQ,          /**< DAQ*/
}cs_ccp_a2l_addreextension_t;

/** \struct cs_ccp_a2l_event_t
    Event channel definition structure
*/
typedef struct{
  uint8_t id;               /**< ID of event*/
  char name[CS_CHARBUF];    /**< Name*/
  char shortname[CS_CHARBUF];    /**< Short name*/
  int cse;                  /**< No information about*/
  int samplerate;      /**< Sample rate in ms*/
}cs_ccp_a2l_event_t;

/** \struct cs_ccp_a2l_source_t
    DAQ List definition structure
*/
typedef struct{
  char name[CS_CHARBUF];            /**< Name*/
  char displayident[CS_CHARBUF];    /**< Display identifier*/
  int period;                       /**< Period in ms, normally 0*/
  int rate;                         /**< Rate in ms, normally 0*/
  int id;                           /**< DAQ List id*/
  int length;                       /**< Number of possible ODT entries*/
  uint32_t can_id;                  /**< CAN id*/
  uint8_t reduction_allowed;        /**< Reduction allowed, 0 => No, 1=> Yes*/
  cs_ccp_a2l_event_t *event;        /**< Pointer to event*/
  uint8_t first_pid;                /**< First pid for DAQ transmission*/
}cs_ccp_a2l_source_t;

/** \struct cs_ccp_a2l_t
    Storage structure for A2L-file parsing
*/
typedef struct{
  cs_a2l_t a2l;           /**< A2L file interface*/
  uint16_t ccp_version;   /**< CCP Version id*/
  uint16_t blob_version;  /**< Blob version id*/
  uint32_t can_send_id;   /**< CAN send id*/
  uint32_t can_recv_id;   /**< CAN recv id*/
  uint16_t station_id;    /**< Station id*/
  uint32_t baudrate;      /**< Baudrate*/
  cs_a2l_bo_t BO;         /**< Byte order*/
  cs_ccp_a2l_daq_mode_t daqmode;  /**< DAQ transfer mode*/
  cs_ccp_a2l_consistency_t consistency;/**< Data consistency level*/
  cs_ccp_a2l_addreextension_t addrextension; /**< Address extension mode*/
  uint8_t cmd_optional[CS_CHARBUF]; /**< Ids of optional cmd*/
  cs_ccp_a2l_event_t *events; /**<Event channel definition*/
  int no_events;          /**< Number of event channel definitions*/
  cs_ccp_a2l_source_t *source; /**< DAQ list definitions*/
  int no_source;          /**< Number of DAQ lists*/ 
}cs_ccp_a2l_t;

#ifdef __cplusplus
extern "C" {
#endif

  /** Open A2L-File
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_ccp_a2l_open(
		      cs_ccp_a2l_t *data,  /**< Working storage structure*/
		      char *filename,      /**< Filename to open*/
		      char *project,       /**< Project to open*/
		      char *module,        /**< Module to open*/
		      char *interface,     /**< Interface to open*/
		      char *option         /**< Optional MOD_PAR section*/
		      );
  
  
  /** Close A2L-File and reset working structure
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_ccp_a2l_close(
		       cs_ccp_a2l_t *data  /**< Working storage structure*/
		       );
  
  /** Print out data structure
   */
  void cs_ccp_a2l_print(
			cs_ccp_a2l_t *data /**< Working storage structure*/
			); 

  /** Print out data structure
   */
  void cs_ccp_a2l_print_raster(
			       cs_ccp_a2l_event_t *data /**< Event data structure*/
			       );
    
  /** Print out data structure
   */
  void cs_ccp_a2l_print_source(
			       cs_ccp_a2l_source_t *data /**< DAQ list data structure*/
			       );

  /**Get DAQ definition, identified by DAQ id
     \retval NULL Failure or not found
     \retval !=NULL valid definition pointer
  */
  cs_ccp_a2l_source_t *cs_ccp_a2l_GetDAQ(
					 cs_ccp_a2l_t *data,  /**< Working storage structure*/
					 uint8_t daq          /**< DAQ id*/
					 );

  /** Get DAQ period in s
      \retval 0.0 Not found or failure
  */
  double cs_ccp_a2l_GetDAQPeriod(
				 cs_ccp_a2l_t *data,          /**< Working storage structure*/
				 uint8_t daq                  /**< DAQ id*/
				 );

  /** Get DAQ Event idx if defined by A2L file
      \retval -1 No Event id found
      \retval >=0 Event idx
  */
  int cs_ccp_a2l_GetDAQEvent(
			     cs_ccp_a2l_t *data,                /**< Working storage structure*/
			     uint8_t daq                        /**< DAQ id*/
			     );

#ifdef __cplusplus
}
#endif

#endif
