/*  CCP CAN Stack for SocketCAN
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_ccp_api.h
 * @author Andreas Stewering-Bone
 * @date 11 June 2015
 * @brief File containing the CCP high level API.
 *
 * This file contain the CCP high level API.
 */

#ifndef CS_CCP_API_H
#define CS_CCP_API_H

#include "cs_ccp.h"
#include "cs_ccp_a2l.h"

/** \def CCP_API_MAXARE
    Maximum areay size for up and download in Bytes
*/
#define CCP_API_MAXAREA 0xFFFF

#ifdef __cplusplus
extern "C" {
#endif

/** Load basic data from A2L file
    \retval 0 No failure
    \retval 1 failure
*/
  int cs_ccp_api_load(
		      cs_ccp_t *data,                       /**< Working storage structure*/
		      char * filename,                      /**< A2L filename (with path)*/
		      char* project,                        /**< Project section name*/
		      char* module,                         /**< Module section name*/
		      char* option,                         /**< Optional MOD_PAR section name, maybe NULL*/
		      uint32_t stim                         /**< Configure message seperation time in ms*/
		      );

/** Open physical interface
    \retval 0 No failure
    \retval 1 failure
*/
  int cs_ccp_api_open(
		      cs_ccp_t *data,                       /**< Working storage structure*/
		      char *iface                           /**< System and type dependent interface name, ex for "can0"*/
		      );

  /** Close physical interface
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_ccp_api_close(
		       cs_ccp_t *data                    /**< Working storage structure*/
		       );
  /**< Get Slave id from data structure
     \retval slaveid
  */
  char *cs_ccp_api_getslaveid(
			      cs_ccp_t *data            /**< Working storage structure*/
			      );

  /** Search for Signal and note it for DAQ transmission 
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_ccp_api_addSignalDAQ(
			      cs_ccp_t *data,      /**< Working storage structure*/
			      char* signalname,        /**< Signal name to search for*/
			      char *ev_name,           /**< Event channel name to register the signal*/
			      int prescaler,           /**< DAQ prescaler*/
			      double *exvalue,         /**< Optional: External value reference, 
							  to store last received value,
							  maybe NULL
						       */
			      double *exts,            /**< Optional: External timestamp reference,
							  to store last received timestamp value,
							  maybe NULL
						       */
			      void (*extcallback)(double value, double ts, void *priv), /**< Optional: Receive callback function for new value
											   Arguments: value -> signal value
											   ts    -> signal timestamp
											   priv  -> private data structure,
											   maybe NULL
											*/
			      void *priv               /**< Optional: Private data structure storage pointer, 
							  for use in callback function call,
							  maybe NULL
						       */
			      );

  /** Connect to slave and optional get status
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_ccp_api_connect(
			 cs_ccp_t *data           /**< Working storage structure*/
			 );
  
  /** Configure DAQ transmission, this is done via dynamic configureation
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_ccp_api_configureDAQ(
			      cs_ccp_t *data         /**< Working storage structure*/
			      );
  
  /** Do a synch start of all configured DAQ lists
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_ccp_api_startDAQ(
			  cs_ccp_t *data             /**< Working storage structure*/
			  );
  
  /** Do a stop of all configured DAQ lists
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_ccp_api_stopDAQ(
			 cs_ccp_t *data             /**< Working storage structure*/
			 );
  
  /** Do message processing, has to be called every cylce
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_ccp_api_process(
			 cs_ccp_t *data             /**< Working storage structure*/
			 );

  /** Set timeout factor for DAQ ODT recv messages.
      An timeout happens when the last ODT recv timestamp  plus
      (DAQ period multiplied by fac_timeout) is lesser than actual time
      \retval 0 No timeout
      \retval 1 There was an failure or timeout
  */
  int cs_ccp_api_setDAQTimeout(
			       cs_ccp_t *data,                       /**< Working storage structure*/
			       double fac                                /**< Timeout factor, has to be >=1.0, < 1.0 the check is deactivated*/
			       );
    
  /** Check for DAQ ODT receive timeout
      \retval 0 No timeout
      \retval 1 There was an failure or timeout
  */
  int cs_ccp_api_hasDAQTimeout(
			       cs_ccp_t *data                        /**< Working storage structure*/
			       );
  /** Download memory segment > 5 Byte to slave
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_api_dnload_area(
			     cs_ccp_t *data,    /**< Working storage structure*/
			     uint32_t address,  /**< Start address*/
			     uint8_t extension, /**< Start address extension*/
			     uint8_t *rdata,    /**< Data to transfer*/
			     int rlen           /**< Data length*/
			     );

  /** Upload memory segment > 5 Byte from slave
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_api_upload_area(
			     cs_ccp_t *data,    /**< Working storage structure*/
			     uint32_t address,  /**< Start address*/
			     uint8_t extension, /**< Start address extension*/
			     uint8_t *rdata,    /**< Data storage buffer*/
			     int rlen           /**< Data storage length*/
			     );

  /** Download numeric Value or vector to Slave
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_api_dnload_value(
			      cs_ccp_t *data,                        /**< Working storage structure*/
			      char *name,                            /**< Characteristic name*/
			      double *value,                         /**< Value pointer*/
			      int numelem                            /**< Num elements*/
			      );

  /** Upload numeric value or vector from slave
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_api_upload_value(
			      cs_ccp_t *data,                        /**< Working storage structure*/
			      char *name,                            /**< Characteristic name*/
			      double *value,                         /**< Value pointer*/
			      int *numelem                           /**< Num elements in storage buffer,
									>= number of values to receive*/
			      );

  /** Build checksum for given address and blocksize
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_ccp_api_buildchksum(
			     cs_ccp_t *data,                         /**< Working storage structure*/
			     uint32_t address,                       /**< Memory address*/
			     uint8_t extension,                      /**< Memory extension*/
			     int blocksize,                          /**< Blocsize in Bytes*/
			     uint32_t *checksum                      /**< Checksum feedback*/
			     );
    
#ifdef __cplusplus
}
#endif

#endif
