/*  dbc.h --  encode and decode CAN messages signals
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @file cs_dbc.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing the DBC file access API (Read only).
 *
 * File containing the DBC file access API (Read only).
*/
#ifndef CS_CAN_DBC_H
#define CS_CAN_DBC_H

#include <inttypes.h>

#include "cs_signal.h"
#include "cs_message.h"
#include "cs_debug.h"

/** \struct cs_dbc_node
    DBC Storage node structure, predecleartion
*/
struct cs_dbc_node;

/** \struct cs_dbc_node_t
    DBC Storage node structure
*/
typedef struct cs_dbc_node cs_dbc_node_t;

/** \struct cs_dbc
    DBC Storage structure, predecleartion
*/
struct cs_dbc;

/** \struct cs_dbc_t
    DBC Storage structure
*/
typedef struct cs_dbc cs_dbc_t;

#ifdef __cplusplus
extern "C" {
#endif
  /** Set debug level
      \retval Active Debuglevel
  */  
  cs_debug_level_t cs_dbc_debuglevel(
				     cs_debug_level_t level  /**< Debug level*/
				     );

  /** Set debugging callback routine
      Default: Function registered which prints debugging to STDOUT
  */
  void cs_dbc_debug_setcallback(
				void *ctx,                             /**< Callback private data*/
				void (*callb)(void *ctx,
					      cs_debug_sub_t sub,
					      cs_debug_level_t dl,
					      const char *func,
					      const int line,
					      const char *fmt,
					      va_list ap) /**< Callback routine*/
				);

  /** Init and allocate DBC storage structure
      \retval 0 Failure
      \retval !=0 Pointer to storage structure
  */
  cs_dbc_t *cs_dbc_init();

  /** Free DBC storage structure
   */
  void cs_dbc_free(
		   cs_dbc_t *sk          /**< DBC storage structure*/
		   );

  /** Read DBC file
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_dbc_read(
		  cs_dbc_t *data,        /**< DBC storage structure*/
		  char *filename         /**< DBC file path and name*/
		  );
  
  /** Get number of messages
      \retval -1 Failure
      \retval >=0 Number of messages
  */
  int cs_dbc_getnummessages(
			    cs_dbc_t *dbc         /**< DBC Storage structure*/
			    );
  
  /** Get message by CAN-Id
      Please note, message structure is freed when cs_dbc_free is called,
      do a message copy
      \retval 0 Failure
      \retval != 0 Message structure
  */
  cs_message_t * cs_dbc_getmessage(
				   cs_dbc_t *dbc,            /**< DBC Storage structure*/
				   uint32_t id               /**< CAN-Id*/
				   );

  /** Get signal by Signal name
      Please note, signal structure is freed when cs_dbc_free is called,
      do a signal copy
      \retval 0 Failure
      \retval != 0 Signal pointer
  */
  cs_signal_t *cs_dbc_getsignal(
				cs_dbc_t *dbc,             /**< DBC Storage structure*/
				char *slname              /**< Signal name*/
				);

  /** Get message by index in list
      Please note, message structure is freed when cs_dbc_free is called,
      do a message copy
      \retval 0 Failure
      \retval !=0 Pointer to signal
  */
  cs_message_t* cs_dbc_getmessage_iter(
				       cs_dbc_t *dbc,  /**< DBC Storage structure*/
				       int iter        /**< List index*/
				       );
  
  /** Get number of defined bus nodes
      \retval -1 Failure
      \retval >=0 Number of bus nodes
  */
  int cs_dbc_getnonodes(
			cs_dbc_t *data    /**< DBC structure*/
			);

  /** Get bus node by name
      Please note, node structure is freed when cs_dbc_free is called
      \retval 0 Failure
      \retval !=0 Node structure
  */
  cs_dbc_node_t* cs_dbc_getnode_byname(
				       cs_dbc_t *data,    /**< DBC structure*/
				       char *name         /**< Node name*/
				       );
  
  /** Get bus unit by index in list
      Please note, node structure is freed when cs_dbc_free is called
      \retval 0 Failure
      \retval !=0 Node structure
  */
  cs_dbc_node_t *cs_dbc_getnode_iter(
				     cs_dbc_t *data,    /**< DBC structure*/
				     int index          /**< Node index*/
				     );
  
  /** Get number of requested messages by node for a given bus node
      \retval -1 Failure
      \retval >=0 Number of requested messages by this bus node
  */
  int cs_dbc_node_no_requested_messages(
					cs_dbc_node_t *data    /**< DBC node structure*/
					);

  /** Get requested message structure from node by index
      Please note the message structure is freed by cs_dbc_free, 
      so please do a copy of the structure -> see cs_message.h
      \retval NULL Failure
      \retval != NULL message structure
  */
  cs_message_t *cs_dbc_node_get_requested_message_iter(
						       cs_dbc_node_t *data,    /**< DBC node structure*/
						       int index               /**< Message index*/
						       );
  
  /** Get number of generated messages by node for a given bus node
      \retval -1 Failure
      \retval >=0 Number of generated messages by this bus node
  */
  int cs_dbc_node_no_generated_messages(
				       cs_dbc_node_t *data    /**< DBC node structure*/
				       );

  /** Get generated message structure from node by index
      Please note the message structure is freed by cs_dbc_free, 
      so please do a copy of the structure -> see cs_message.h
      \retval NULL Failure
      \retval != NULL message structure
  */
  cs_message_t *cs_dbc_node_get_generated_message_iter(
						       cs_dbc_node_t *data,    /**< DBC node structure*/
						       int index               /**< Message index*/
						       );

  /** Get bus node name
      Please note the char buffer is freed after cs_dbc_free is called
      \retval 0 Failure
      \retval != 0 Name
  */
  char *cs_dbc_node_get_name(
			     cs_dbc_node_t *data
			     );
  
#ifdef __cplusplus
}
#endif

#endif
