/*
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @file cs_isotp.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing an ISOTP userspace Layer.
 *
 * File containing an ISOTP userspace Layer.
 * Limitations:
 * - Only half duplex
 */
#ifndef CS_ISOTP_H
#define CS_ISOTP_H

#include <inttypes.h>
#include <stdarg.h>

#include "cs_socketcan.h"
#include "cs_debug.h"

#define ISOTP_SF_ID 0x0
/** \def ISOTP_FF_ID
    ISOTP ID First frame
*/
#define ISOTP_FF_ID 0x10
/** \def ISOTP_CF_ID
    ISOTP ID Conductive frame
*/
#define ISOTP_CF_ID 0x20
/** \def ISOTP_FC_ID
    ISOTP ID Flow control frame
*/
#define ISOTP_FC_ID 0x30

/** \def ISOTP_ERROR_NOERROR
    No error
*/
#define ISOTP_ERROR_NOERROR 0x00

/** \def ISOTP_ERROR_SOCKETCAN_INIT
    Init failure in raw SocketCAN layer
*/
#define ISOTP_ERROR_SOCKETCAN_INIT 0x01

/** \def ISOTP_ERROR_SOCKETCAN
    General SocketCAN failure
*/
#define ISOTP_ERROR_SOCKETCAN 0x02

/** \def ISOTP_ERROR_RECV
    General recv failure
*/
#define ISOTP_ERROR_RECV 0x03

/** \def ISOTP_ERROR_SEND
    General send failure
*/
#define ISOTP_ERROR_SEND 0x04

/** \def ISOTP_ERROR_UNSUPPORTED
    Not supported
*/
#define ISOTP_ERROR_UNSUPPORTED 0x05

/** \def ISOTP_ERROR_UNKNOWN
    Unknown FRAME
*/
#define ISOTP_ERROR_UNKNOWN 0x06

/** \def ISOTP_ERROR_SF
    Invalid SF Frame
*/
#define ISOTP_ERROR_SF 0x07

/** \def ISOTP_ERROR_FC
    Invalid FC Frame
*/
#define ISOTP_ERROR_FC 0x08

/** \def ISOTP_ERROR_UNEXPECTED
    Receive unexpected Frame
*/
#define ISOTP_ERROR_UNEXPECTED 0x09

/** \def ISOTP_ERROR_FF
    Invalid FF Frame
*/
#define ISOTP_ERROR_FF 0x0A

/** \def ISOTP_ERROR_DATA
    Data buffer failure
*/
#define ISOTP_ERROR_DATA 0x0B

/** \def ISOTP_ERROR_CF
    Invalid CF frame
*/
#define ISOTP_ERROR_CF 0x0C

/** \def ISOTP_ERROR_PARAMETER
    Invalid Parameter
*/
#define ISOTP_ERROR_PARAMETER 0x0D

/** \def ISOTP_ERROR_SEQUENCE
    Invalid Sequence number
*/
#define ISOTP_ERROR_SEQUENCE 0x0E

/** \def ISOTP_ERROR_TIMEOUT
    Recv timeout
*/
#define ISOTP_ERROR_TIMEOUT 0x0F

/** \def ISOTP_ERROR_TXFULL
    TX path is full timeout
*/
#define ISOTP_ERROR_TXFULL 0x10

/** \def ISOTP_ERROR_OPTION
    Invalid Option
*/
#define ISOTP_ERROR_OPTION 0x11

/** \def ISOTP_ERROR_RXPADDING
    RXPADDING failure
*/
#define ISOTP_ERROR_RXPADDING 0x12

/** \def ISOTP_ERROR_RXLENGTH
    RX Length failure
*/
#define ISOTP_ERROR_RXLENGTH 0x13

/** \def ISOTP_ERROR_OVERFLOW
    Data buffer overflow on receiver side
*/
#define ISOTP_ERROR_OVERFLOW 0x14

/** \def ISOTP_OPTION_NONE
    No option set
*/
#define ISOTP_OPTION_NONE 0x00

/** \def ISOTP_OPTION_TXSTIM_FORCE
    Ignore TXSTIM given by target,
    overwrite by configured value
*/
#define ISOTP_OPTION_TXSTIM_FORCE 0x01

/** \def ISOTP_OPTION_TXPADDING
    Always fill send messages up to 8 Bytes
*/
#define ISOTP_OPTION_TXPADDING 0x02

/** \def ISOTP_OPTION_RXPADDING
    Check RX frame length for full length
*/
#define  ISOTP_OPTION_RXPADDING 0x04

/** \def ISOTP_OPTION_RXLENGTH
    Check RX frame for exact length
*/
#define ISOTP_OPTION_RXLENGTH 0x08

/** \def ISOTP_OPTION_RXDATA
    Check RX frame for padding data
*/
#define ISOTP_OPTION_RXDATA 0x10

/** \def ISOTP_OPTION_RXTESTING
    Add internal testing sequences on receiver side
    Example: Add Wait cycles,...
*/
#define ISOTP_OPTION_RXTESTING 0x20

/** \def ISOTP_OPTION_DEFAULT
    Default option
*/
#define ISOTP_OPTION_DEFAULT (ISOTP_OPTION_NONE)

/** \def ISOTP_PADDING_DEFAULT
    Default padding byte
*/
#define ISOTP_PADDING_DEFAULT 0x00



/** \enum cs_isotp_debug_level_t
    ISOTP debug level
*/
typedef enum{
  ISOTP_DEBUG_LEVEL_READ = 0x00,      /**< Only read debug level*/
  ISOTP_DEBUG_LEVEL_NONE = 0x01,  /**< No debug output*/
  ISOTP_DEBUG_LEVEL_DEBUG = 0x02,     /**< Debugging output*/
  ISOTP_DEBUG_LEVEL_VERBOSE = 0x04   /**< Verbose output*/
}cs_isotp_debug_level_t;

/** \enum cs_isotp_ll_mode_t
    Link Layer running mode
*/
typedef enum{
  ISOTP_LL_MODE_CAN = 0x00,      /**< Only CAN <= 2.0*/
  ISOTP_LL_MODE_CANFD= 0x01,     /**< CAN-FD*/
  ISOTP_LL_MODE_CANFD_BRS= 0x02, /**< CAN-FD with second bitrate*/
}cs_isotp_ll_mode_t;

/** \enum cs_isotp_ll_tx_dl_t
    Link Layer transmit data length
*/
typedef enum{
  ISOTP_LL_TX_DL_8 = 8,      /**< 8 Byte (CAN 2.0 compatible), Default value*/
  ISOTP_LL_TX_DL_12 = 12,    /**< 12 Byte */
  ISOTP_LL_TX_DL_16 = 16,    /**< 16 Byte */
  ISOTP_LL_TX_DL_20 = 20,    /**< 20 Byte */
  ISOTP_LL_TX_DL_24 = 24,    /**< 24 Byte */
  ISOTP_LL_TX_DL_32 = 32,    /**< 32 Byte */
  ISOTP_LL_TX_DL_48 = 48,    /**< 48 Byte */
  ISOTP_LL_TX_DL_64 = 64,    /**< 64 Byte */
}cs_isotp_ll_tx_dl_t;



/** \struct cs_isotp
    IsoTP Storage structure, predecleartion
*/
struct cs_isotp;
/** \struct cs_isotp_t
    IsoTP Storage structure
*/
typedef struct cs_isotp cs_isotp_t; 

#ifdef __cplusplus
extern "C" {
#endif


  /** Set debug level
      \retval Active Debuglevel
  */  
  cs_debug_level_t cs_isotp_debuglevel(
				       cs_debug_level_t level  /**< Debug level*/
				       );

  /** Set debugging callback routine
      Default: Function registered which prints debugging to STDOUT
  */
  void cs_isotp_debug_setcallback(
				  void *ctx,                             /**< Callback private data*/
				  void (*callb)(void *ctx,
						cs_debug_sub_t sub,
						cs_debug_level_t dl,
						const char *func,
						const int line,
						const char *fmt,
						va_list ap) /**< Callback routine*/
				  );
  
  /** Get last error code
      \retval last error code
  */
  int cs_isotp_getlasterror(
			    cs_isotp_t *data  /**< Storage structure*/
			    );


  /** Get error id string
      \retval error string
  */
  char* cs_isotp_geterroridstring(
				      uint8_t lerror /**< Error id*/
				    );
  
  /** Get last error string
      \retval error string
  */
  char* cs_isotp_getlasterrorstring(
				    cs_isotp_t *datalen /**< Storage structure*/
				    );

  /** Reset data structure
      \retval 0 No failure
      \retval 1 Failure
  */  
  int cs_isotp_reset(
		     cs_isotp_t *data  /**< Storage structure*/
		     );

  /* Get low level SocketCAN handle
     \retval NULL Failure
     \retval SocketCAN structure
  */
  cs_socketcan_t *cs_isotp_get_socketcan(
					 cs_isotp_t *data     /**< Storage structure*/
					 );
  
  /** Init IsoTP Layer
      \retval Handle Success
      \retval Null Failure
  */
  cs_isotp_t *cs_isotp_init(
			    uint32_t sa,          /**< Source CAN id*/
			    uint32_t ta           /**< Target CAN id*/
			    );

  /** Free data structures
   */
  void cs_isotp_free(
		     cs_isotp_t *data            /**< Storage structure*/
		     );

  /** Set transfer information callback function pointer
   */
  void cs_isotp_transfer_setcallback(
				     cs_isotp_t *data,                   /**< Storage structure*/
				     void *ctx,                          /**< Callback function context pointer*/
				     void (transfercb)(void *ctx,        /**< Coontex data*/
						       uint8_t mode,     /**< Mode SF,FF,CF,FC frame, see ISOTP_XX_ID definition*/
						       uint32_t length,  /**< Data segment length in bytes */
						       uint32_t counter, /**< Actual frame counter*/
						       uint32_t actbytes /**< Actal send or received bytes*/
						       )/**< Function callback pointer*/
				     );
  
  /** Set option flags
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_isotp_setoptions(
			  cs_isotp_t *data, /**< Storage structure*/ 
			  uint8_t option    /**< Options bitmask mask, look for ISOTP_OPTION_XXXXX */
			  );

  /** Set link layer options
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_isotp_setoptions_link_layer(
				     cs_isotp_t *data,             /**< Storage structure*/
				     cs_isotp_ll_mode_t ll_mode,   /**< CAN running mode*/
				     cs_isotp_ll_tx_dl_t tx_dl     /**< CAN layer transmit data length*/
				     );

  /** Get maximum ISOTP frame length in bytes
      \retval -1 Failure
      \retval >=0 max frame length in bytes
  */
  int cs_isotp_getmaxbytes(
			   cs_isotp_t *data                /**< Storage structure*/
			   );
  
  /** Set padding byte
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_isotp_setpadbyte(
			  cs_isotp_t *data, /**< Storage structure*/ 
			  uint8_t byte      /**< Padding byte*/
			  );

  /** Set STIM timing
      For TX path if value is >= 0, ignore received value and force configured timing
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_isotp_settiming(
			 cs_isotp_t *data,  /**< Storage structure*/
			 int txstim,        /**< if >= 0 force TX stim, ignore value in FC frame 0.000001s*/
			 int rxstim         /**< if >= 0 set RX STIM, value in 0.000001s*/
			 );

  /** Reset timing to default values
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_isotp_reset_timing(
			    cs_isotp_t *data  /**< Storage structure*/
			    );

  
  /** Set RX block size
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_isotp_setrxblocksize(
			      cs_isotp_t *data,  /**< Storage structure*/
			      uint8_t bs         /**< if > 0 set RX block size*/
			      );

  /** Open Library isotp layer
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_isotp_open(
		    cs_isotp_t *data,        /**< Storage structure*/
		    char *iface,             /**< Interface name*/
		    char *server,            /**< Server hostname SocketCANd, if NULL take native socket*/
		    int port                 /**< Server port */          
		    );

  /** Close Library isotp layer
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_isotp_close(
		     cs_isotp_t *data    /**< Storage structure*/
		     );

  /** Receive ISOTP Frame, function is blocking
      \retval 0 success
  */
  int cs_isotp_recv(
		    cs_isotp_t *data,       /**< Storage structure*/
		    uint8_t *buffer,        /**< Receive buffer*/
		    uint32_t *len,          /**< Max len of buffer, will be overwritten by number of received bytes*/
		    uint32_t timeout        /**< Receive timeout in ms*/
		    );

  /** Send ISOTP frame, function is blocking
      \retval 0 success
  */
  int cs_isotp_send(
		    cs_isotp_t *data,      /**< Storage structure*/
		    uint8_t *buffer,       /**< Send buffer*/
		    uint32_t len           /**< Buffer length*/
		    );

#ifdef __cplusplus
}
#endif

#endif
