/*  
    Copyright (C) 2017 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_lin.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2017
 * @brief File containing lin bus header.
 *
 * File containing lin bus header.
*/
#ifndef CS_LIN_H
#define CS_LIN_H

#include <inttypes.h>

#include "cs_lists.h"
#include "cs_debug.h"

/* \def CS_LIN_CHARBUF
   LIN charbuf size
*/
#define CS_LIN_CHARBUF 50

/** /def CS_LIN_MAXDLC
    Maximum allowed frame length
*/
#define CS_LIN_MAXDLC 8

/* \def CS_LIN_SIGNAL_MAX_BYTE
   Maximal length of a signal byte array
*/
#define CS_LIN_SIGNAL_MAX_BYTE 8

/* \def CS_LIN_MAX_ID
   Maximum allowed standard frame id
*/
#define CS_LIN_MAX_ID 0x3F

/* \def CS_LIN_TMAX_FAKTOR
   Faktor for maximum time calculation
   Nominal x faktor = maximum
*/
#define CS_LIN_TMAX_FAKTOR 1.4
/* \def CS_LIN_BITS_HEADER
   Number of Bits in LIN Header
*/
#define CS_LIN_BITS_HEADER 34
/** \def CS_LIN_ERROR_NOERROR
    No Error
*/
#define CS_LIN_ERROR_NOERROR 0x00
/** \def CS_LIN_ERROR_SLLIN
    SLLIN specific error
*/
#define CS_LIN_ERROR_SLLIN 0x01
/** \def CS_LIN_ERROR_SOCKETCAN
    SocketCAN specific error
*/
#define CS_LIN_ERROR_SOCKETCAN 0x02
/** \def CS_LIN_ERROR_INTERNAL
    Internal data error
*/
#define CS_LIN_ERROR_INTERNAL 0x03
/** \def CS_LIN_ERROR_PARAMETER
    Parameter error
*/
#define CS_LIN_ERROR_PARAMETER 0x04
/** \def CS_LIN_ERROR_LDF
    LDF specific error
*/
#define CS_LIN_ERROR_LDF 0x05
/** \def CS_LIN_ERROR_SCHED
    Scheduling specific error
*/
#define CS_LIN_ERROR_SCHED 0x06
/** \def CS_LIN_ERROR_CONTENT
    Parity or checksum specific error
*/
#define CS_LIN_ERROR_CONTENT 0x07

/** \def CS_LIN_GET_ID
    Remove Parity bits from id
*/
#define CS_LIN_GET_ID(X) (X & CS_LIN_MAX_ID)

/** \struct cs_lin
    LIN Storage structure, predecleartion
*/
struct cs_lin;
/** \struct cs_lin_t
    LIN Storage structure
*/
typedef struct cs_lin cs_lin_t;

/** \struct cs_lin_ldf_doc
    LIN LDF parser Storage structure, predecleartion
*/
struct cs_lin_ldf_doc;
/** \struct cs_lin_ldf_doc_t
    LIN LDF parser Storage structure
*/
typedef struct cs_lin_ldf_doc cs_lin_ldf_doc_t;


/*Local predeclerations*/
typedef struct cs_lin_schedule_table cs_lin_schedule_table_t;
typedef struct cs_lin_frame cs_lin_frame_t;
typedef struct cs_lin_node cs_lin_node_t; 


/** \var cs_lin_raw_t
    Basic lin raw value
*/
typedef uint16_t cs_lin_raw_t;


/** \enum cs_lin_mode_t 
    LIN working mode
 */
typedef enum{
  LIN_MODE_SLLIN = 0x00,        /**< Usage SocketCAN SLLIN interface*/
  LIN_MODE_SOCKETCAN,           /**< Emulate LIN bus, only SocketCAN usage*/
}cs_lin_mode_t;

/** \enum Signal encoding type
 */
typedef enum{
  LIN_ENC_DYNAMIC = 0x00,       /**< Dynamic encoded, logical or physical*/
  LIN_ENC_LOGICAL,              /**< Logical encoded*/
  LIN_ENC_PHYSICAL,             /**< Physical scaled signal*/
  LIN_ENC_BCD,                  /**< BCD encoded signal*/
  LIN_ENC_ASCII,                /**< Ascii encoded signal*/
}cs_lin_enc_type_t;

/** \enum cs_lin_cksum_t
    Checksum type
*/
typedef enum{
  LIN_CKS_UNKNOWN = 0x00,      /**< Not configures*/
  LIN_CKS_CLASSIC,             /**< Classic Checksum, <2.0*/
  LIN_CKS_ENHANCED,            /**< Enhanced, >=2.0*/
}cs_lin_cksum_t;

/** \struct cs_lin_enc_logical_t 
    Logical signal encoding instance
 */
typedef struct{
  cs_lin_raw_t value;          /**< Raw value*/
  char text[CS_LIN_CHARBUF];   /**< Text representation*/
}cs_lin_enc_logical_t;

/** \struct cs_lin_enc_physical_t
    Physical signal encoding instance
*/
typedef struct{
  cs_lin_raw_t min_raw;       /**< Minimal raw value*/
  cs_lin_raw_t max_raw;       /**< Maximal raw value*/
  double min_phys;            /**< Minimal physical value*/
  double max_phys;            /**< Maximal physical value*/
  double scale;               /**< Scale value*/
  double offset;              /**< Offset value*/
  char unit[CS_LIN_CHARBUF];  /**< Unit*/
}cs_lin_enc_physical_t;

/** \struct cs_lin_enc_t
    Signal encoding instance
    Set of rules to de- or encode one or more signal
*/
typedef struct{
  char name[CS_LIN_CHARBUF]; /**< Signal encoding type name*/
  cs_lists_t l_logical;      /**< List of logical encoding instances*/
  cs_lists_t l_physical;     /**< List of physical encoding instances*/
  cs_lin_enc_type_t base;    /**< Base encoding type, if dynamical encoding is based on one or more logical or physical instances,
			      Logical and physical encoding is range dependend, BCD and Ascii not, so if BCD or ascii is choosen,
			     we have not to look up a range depended conversion*/
}cs_lin_enc_t;

/** \struct cs_lin_enc_rep_instance_t
    Signal encoding representation instance
    Sort Signal to a signal encoding instance
    Used only for parsing a LDF file
*/
typedef struct{
  char name[CS_LIN_CHARBUF];           /**< Signal encoding instance name*/
  cs_lin_enc_t *enc;                   /**< Encoding definition pointer*/
  cs_lists_t l_signalname;             /**< List of signal names*/
}cs_lin_enc_representation_t;

/** \struct cs_lin_signal_def_t
    Signal definition, can be used in multiple frames
*/
typedef struct{
  char name[CS_LIN_CHARBUF];                 /**< Signal name*/
  cs_lin_raw_t init[CS_LIN_SIGNAL_MAX_BYTE]; /**< Signal init buffer*/
  int initlen;                               /**< Signal init buffer length*/
  int len;                                   /**< Signal length in bits*/
  cs_lin_enc_t *enc;                         /**< Signal encoding pointer*/
  char publisher[CS_LIN_CHARBUF];            /**< Signal publisher*/
  cs_lists_t l_subscribername;               /**< List of subscriber node names*/
}cs_lin_signal_def_t;

/** \enum cs_lin_signal_state_t
    Signal instance state
*/
typedef enum{
  LIN_SIGNAL_STATE_IDLE = 0x00,             /**< Idle, nothing happend*/
  LIN_SIGNAL_STATE_UPDATED,                 /**< Signal was updated via LIN recv or application update*/
} cs_lin_signal_state_t;
  
/** \struct cs_lin_signal_t
    Signal instance, existist in one frame
*/
typedef struct{
  char name[CS_LIN_CHARBUF];                /**< Signal name, only for identification*/
  cs_lin_signal_def_t *signal_def;          /**< Signal definition pointer*/
  cs_lin_raw_t conv[CS_LIN_SIGNAL_MAX_BYTE];/**< Signal converted buffer*/
  uint8_t raw[CS_LIN_MAXDLC];               /**< Signal raw buffer, converted signal values @ correct position in frame*/
  int offset;                               /**< Signal bit offset in frame*/
  cs_lin_signal_state_t state;              /**< Signal state*/
  cs_lin_frame_t *frame;                    /**< Signal parent frame pointer*/
}cs_lin_signal_t;

/** \enum cs_lin_ldf_frame_direction_t
    Frame direction, master view
*/
typedef enum{
  LIN_DIR_UNKNOWN = 0x00,                       /**< Not configured*/
  LIN_DIR_RECEIVE,                              /**< Master receive*/
  LIN_DIR_SEND,                                 /**< Master send*/
}cs_lin_ldf_frame_direction_t;

/** \enum cs_lin_ldf_frame_state_t
    Frame state
*/
typedef enum{
  LIN_STATE_IDLE = 0x00,                        /**< Frame state is idle*/
  LIN_STATE_UPDATED,                            /**< One of the frame signals was updated by appilcation*/
  LIN_STATE_SEND,                               /**< Frame was send*/
  LIN_STATE_RECEIVED,                           /**< Frame was received*/
  LIN_STATE_TIMEOUT,                            /**< Frame has receive timeout*/
  LIN_STATE_ERROR,                              /**< Frame has genereral error*/
}cs_lin_ldf_frame_state_t;

/** \struct cs_lin_frame
    Frame instance
*/
struct cs_lin_frame{
  char name[CS_LIN_CHARBUF];                /**< Frame name*/
  uint32_t id;                              /**< Frame id*/
  int dlc;                                  /**< Frame length*/
  char publisher[CS_LIN_CHARBUF];           /**< Frame publisher node name*/
  cs_lin_node_t *pubnode;                   /**< Publisher node pointer*/  
  cs_lists_t l_signal;                      /**< Signal instance list*/
  cs_lin_cksum_t t_cksum;                   /**< Checksum type*/
  cs_lists_t l_framename;                   /**< List of frame names, used for sporadic or event triggered frames*/
  cs_lists_t l_frame;                       /**< List of frame pointer, used for sporadic or event triggered frames*/
  char ev_tablename[CS_LIN_CHARBUF];        /**< Collision resolving table name*/
  cs_lin_schedule_table_t *ev_table;        /**< Collision resolving scheduler table for event frames*/
  cs_lin_ldf_frame_direction_t dir;         /**< Frame message direction, master view*/
  cs_lin_ldf_frame_state_t state;           /**< Frame state*/
  double recv_ts;                           /**< Receive ts in s*/
  double t_frame_max;                       /**< Maxímum frame handling time in s*/
  uint8_t raw[CS_LIN_MAXDLC];               /**< Raw data buffer, send and receive*/
};
typedef struct cs_lin_frame cs_lin_frame_t;

/** \enum cs_lin_schedule_entry_type_t
    Scheduler entry type
*/
typedef enum{
  MASTERREQ = 0x00,                         /**< Master request*/
  SLAVERESP,                                /**< Slave response*/
  ASSIGNNAD,                                /**< Assign NAD request*/
  CONDITIONALCHANGENAD,                     /**< Conbditional change NAD request*/
  DATADUMP,                                 /**< Data dump request*/
  SAVECONFIGURATION,                        /**< Save configuration request*/
  ASSIGNFRAMEIDRANGE,                       /**< Assign frame PID request*/
  FREEFORMAT,                               /**< Free format request*/
  ASSIGNFRAMEID,                            /**< Assign frame id*/
  TYPEMAX,                                  /**< End mark, for predefined entry types*/
  FRAME,                                    /**< Standard cyclic frame*/
  SPORADICFRAME,                            /**< Sporadic frame*/
  EVENTFRAME,                               /**< Event frame*/
}cs_lin_schedule_entry_type_t;

/** \struct cs_lin_schedule_entry_t
    Scheduler table entry
*/
typedef struct{
  cs_lin_schedule_entry_type_t entry_type; /**< Entry type*/
  char name[CS_LIN_CHARBUF];               /**< Entry name, note can be a reservered name or a frame name*/
  double delay;                            /**< Frame delay in s*/
  cs_lin_frame_t *frame;                   /**< Frame pointer if type == FRAME*/
  double t_frame_slot;                     /**< Time frame slot in s*/
  /* Other types of scheduler entries are not supported until now*/
}cs_lin_schedule_entry_t;

/** \struct cs_lin_schedule_table_t
    Scheduler table
*/
struct cs_lin_schedule_table {
  char name[CS_LIN_CHARBUF];                /**< Table name*/
  cs_lists_t l_entry;                       /**< List of scheduler entries*/
};
typedef struct cs_lin_schedule_table cs_lin_schedule_table_t; 

/** \struct cs_lin_nodeattribute_t
    Node attribute definition
*/
typedef struct{
  char name[CS_LIN_CHARBUF];                /**< Node name*/
  double version_protocol;                  /**< Protocol version, eg. 2.0*/
  uint32_t configured_nad;                  /**< Configured NAD id*/
  uint32_t initial_nad;                     /**< Initial NAD id, optional*/
  uint16_t product_id;                      /**< Product id, only version >=2.0*/
  uint16_t vendor_id;                       /**< Vendor id, only version >=2.0*/
  uint8_t variant_id;                       /**< Variant id, optional, only version >=2.0*/
  char response_errorname[CS_LIN_CHARBUF];  /**< Response error signal name, only version >=2.0*/
  cs_lin_signal_t *response_error;          /**< Response error signal pointer, only version >=2.0*/  
  /* further entries exist, but not implemented*/
}cs_lin_nodeattribute_t;

/** \struct cs_lin_node
    Node definition structure
*/
struct cs_lin_node{
  char name[CS_LIN_CHARBUF];                /**< Node name*/
  int ismaster;                             /**< If equals 1, node is master*/
  double timebase;                          /**< Time base in s, only for master node*/
  double jitter;                            /**< Jitter in s, only for master*/
  cs_lin_nodeattribute_t *attr;             /**< Node attribute pointer*/
  cs_lists_t l_frames_pub;                  /**< List of frames handled this node*/
};
typedef struct cs_lin_node cs_lin_node_t;

/** \struct cs_lin_ldf_t
    Lin LDF file definition structure
*/
typedef struct{
  double protocol_version;               /**< Protocol version, eg. 2.1*/
  double language_version;               /**< Language version, eg. 2.1*/
  uint32_t speed;                        /**< Baud rate in bits/s*/
  double t_bit;                          /**< Bittime in s*/
  double t_header_max;                   /**< Maximal Header time in s*/
  char channel[CS_LIN_CHARBUF];          /**< Channel name*/
  cs_lists_t l_signalencoding;           /**< List of signal encoding instances*/
  cs_lists_t l_signalrepresentations;    /**< List of signals representations*/
  cs_lists_t l_signals;                  /**< List of signals definitions*/
  cs_lists_t l_frames;                   /**< List of frame definitions*/
  cs_lists_t l_schedulingtables;         /**< List of scheduling tables*/
  cs_lists_t l_nodes;                    /**< List of node definitions*/
  cs_lists_t l_nodeattributes;           /**< List of node attributes*/
  cs_lin_node_t *master;                 /**< Master node pointer*/
}cs_lin_ldf_t;


#ifdef __cplusplus
extern "C" {
#endif

  /** Set debug level
      \retval Active Debuglevel
  */  
  cs_debug_level_t cs_lin_debuglevel(
				     cs_debug_level_t level  /**< Debug level*/
				     );
  
  /** Set debugging callback routine
      Default: Function registered which prints debugging to STDOUT
  */
  void cs_lin_debug_setcallback(
				void *ctx,                             /**< Callback private data*/
				void (*callb)(void *ctx,
					      cs_debug_sub_t sub,
					      cs_debug_level_t dl,
					      const char *func,
					      const int line,
					      const char *fmt,
					      va_list ap) /**< Callback routine*/
				);

    /** Get actual error code 
      See CS_LIN_ERROR_XXXXX
      \retval Error code
  */
  int cs_lin_getlasterror(
			  cs_lin_t *sk    /**< Socket struct pointer*/
			  );
  /** Get actual error code human readable
      \retvar char pointer
  */
  char* cs_lin_getlasterrorstring(
				  cs_lin_t *sk  /**< Socket struct pointer*/
				  );
  
  /** Init structure
      \retval storage structure
      \retval 0 in failure case
  */
  cs_lin_t *cs_lin_init();

  /** Free structure
   */
  void cs_lin_free(
		   cs_lin_t *sk           /**< Socket struct pointer*/
		   );


  /** Get actual system time
      \retval Time in s
  */
  double cs_lin_get_time();
  
  /** Set interface specific callback functions
   */
  void cs_lin_set_callback(
			   cs_lin_t *sk,         /**< Socket struct pointer*/
			   void *ctx,            /**< Context pointer for callbacks*/
			   uint8_t send_finished, /**< 0=> No send finished call needed, 1=> send finished call needed*/
			   void (*cs_lin_cb_prepfilt)(void *ctx, uint32_t *id, int len),  /**< Prepare recv filter*/
			   void (*cs_lin_cb_send)(void *ctx, uint32_t id, uint8_t *data, int len, int enhanced),  /**< Master / Slave send callback*/
			   void (*cs_lin_cb_recv)(void *ctx, uint32_t id, int enhanced),  /**< Master recv request callback*/
			   void (*cs_lin_cb_prepcache)(
						       void *ctx,        /**< Context pointer*/
						       uint32_t id,      /**< LIN id*/
						       uint8_t *data,    /**< Data buffer*/
						       int len,          /**< Len data buffer*/
						       int enhanced      /**< 0=> Classic checksum, 1=>Enhanced checksum*/
						       )  /**< Slave prepare callback*/
			   );


  
  /** Configure Lin stack for slave usage
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_init_slave(
			cs_lin_t *sk,           /**< Socket struct pointer*/
			cs_lin_ldf_t *ldf,      /**< LDF based configuration*/
			const char **nodes,     /**< Char pointer array, with node names*/
			int no_nodes            /**< Number of node names*/
			);

  /** Reset prepared slave frame content to initial state
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_reset_slave(
			 cs_lin_t *sk           /**< Socket struct pointer*/
			 );


  /** Init listen only slave
      Receives all defined frames
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_init_listener(
			   cs_lin_t *sk,           /**< Socket struct pointer*/
			   cs_lin_ldf_t *ldf       /**< LDF based configuration*/
			   );

  /** Reset prepared slave listener frame content to initial state
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_reset_listener(
			    cs_lin_t *sk           /**< Socket struct pointer*/
			    );
  
  /** Initial Lin stack for master usage
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_init_master(
			 cs_lin_t *sk,           /**< Socket struct pointer*/
			 cs_lin_ldf_t *ldf,      /**< LDF based configuration*/
			 char *schedtable        /**< Main scheduler table*/
			 );
  
  /** Reset prepared master frame content to initial state
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_reset_master(
			  cs_lin_t *sk           /**< Socket struct pointer*/
			  );


  /** Interface sending is completeted
      \retval 0 No failure
      \retval 1 Failure
   */
  int cs_lin_iface_finished(
			    cs_lin_t *sk           /**< Socket struct pointer*/
			    );

  /** Calculate id with parity bits of id
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_id_parity(
		       uint32_t id,                /**< LIN Id*/
		       uint32_t *pid               /**< LIN id with parity bits*/
		       );

  /** Calculate the frame Checksum
      \retval Checksum
   */
  uint8_t cs_lin_checksum(
			  uint8_t id,           /**< LIN Id with parity bits*/
			  uint8_t *data,        /**< Data buffer*/
			  int len,              /**< Data len*/
			  int enhanced_fl       /**< 0=> Classic CRC, 1=> Enhaced CRC*/
			  );
  
  /** Process cycle start.
      \retval 0 No failure
      \retval 1 Failure
      \retval 2 Timeout
  */
  int cs_lin_process_start(
			   cs_lin_t *sk,         /**< Socket struct pointer*/
			   double ts             /**< Actual timestamp*/
			   );
  
  /** Process incomming message
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_process_msg(
			 cs_lin_t *sk,       /**< Socket struct pointer*/
			 uint32_t id,        /**< Message id received*/
			 uint8_t *data,      /**< Data pointer*/
			 int len             /**< Data length*/
			 );

  /** Process raw incomming message (with parity bits and checksum)
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_process_msg_raw(
			     cs_lin_t *sk,       /**< Socket struct pointer*/
			     uint8_t id,         /**< Message id received, with parity bits*/
			     uint8_t *data,      /**< Data pointer*/
			     int len,            /**< Data length*/
			     uint8_t checksum    /**< Chesum received*/
			     );

  /** Switch master scheduler table
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_master_switch_table(
				 cs_lin_t *sk,         /**< Socket struct pointer*/
				 char *schedtable        /**< Scheduler table*/
				 );

  /** Final call for a process cycle. If we are master and there is a message to send the stack calls the master send callback.
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_process_end(
			 cs_lin_t *sk        /**< Socket struct pointer*/
			 );
  
  /** Search frame identified by id
      Note it is a runtime call, only frames which are handled by this instance (master/slave)
      can be found.
      \retval NULL Failure or not found
  */
  cs_lin_frame_t *cs_lin_get_frame_by_id(
					 cs_lin_t *sk,           /**< Socket struct pointer*/
					 uint32_t id             /**< Frame id*/
					 );

  /** Get frame state
      \retval -1                           General failure
      \retval LIN_STATE_IDLE               Frame state is idle
      \retval LIN_STATE_UPDATED            One of the frame signals was updated by appilcation
      \retval LIN_STATE_SEND               Frame was send
      \retval LIN_STATE_RECEIVED           Frame was received
      \retval LIN_STATE_TIMEOUT            Frame has receive timeout
      \retval LIN_STATE_ERROR              Frame has genereral error
   */
  int cs_lin_get_frame_state(
			     cs_lin_frame_t * frame  /**< Frame data structure*/
			     );
  
  /** Get signal from frame identified by name
      \retval NULL Failure or not found
      \retval signal Signal pointer
  */
  cs_lin_signal_t *cs_lin_get_frame_signal(
					   cs_lin_frame_t * frame,  /**< Frame data structure*/
					   const char *name         /**< Signal name*/
					   );

  /** Reset frame state by application
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_reset_frame_state(
			       cs_lin_frame_t * frame  /**< Frame data structure*/
			       );

  
  /** Search frame signal identified by frame id and signal name
      Note it is a runtime call, only frame signals which are handled by this instance (master/slave)
      can be found.
      \retval NULL Failure or not found
      \retval signal Signal pointer
  */
  cs_lin_signal_t *cs_lin_get_signal(
				     cs_lin_t *sk,           /**< Socket struct pointer*/
				     uint32_t id,            /**< Frame id*/
				     const char *name        /**< Signal name*/
				     );


  /** Update physical encoded signal with new values
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_update_signal_physical(
				    cs_lin_t *sk,           /**< Socket struct pointer*/
				    cs_lin_signal_t * sig,  /**< Signal pointer*/
				    double *value,          /**< Signal value array*/
				    int len                 /**< Value error len*/
				    );

  /** Get physical encoded signals out of the lin stack
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_get_signal_physical(
				 cs_lin_t *sk,           /**< Socket struct pointer*/
				 cs_lin_signal_t * sig,  /**< Signal pointer*/
				 double *value,          /**< Signal value array*/
				 int len                 /**< Value array len*/
				 );

  /** Update logical encoded signal with a new value
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_update_signal_logical(
				   cs_lin_t *sk,           /**< Socket struct pointer*/
				   cs_lin_signal_t * sig,  /**< Signal pointer*/
				   const char *value,      /**< Signal value*/
				   int offset              /**< Offset in value array*/
				   );

  /** Get logical encoded signal out of the lin stack
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_get_signal_logical(
				cs_lin_t *sk,           /**< Socket struct pointer*/
				cs_lin_signal_t * sig,  /**< Signal pointer*/
				char *value,            /**< Signal value array*/
				int len,                /**< Value max buffer len*/
				int offset              /**< Offset in signal array*/
				);

  /** Update raw signal with new values
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_update_signal_raw(
			       cs_lin_t *sk,           /**< Socket struct pointer*/
			       cs_lin_signal_t * sig,  /**< Signal pointer*/
			       cs_lin_raw_t *value,    /**< Signal value*/
			       int len                 /**< Value array length*/
			       );
  
  /** Get Raw signal out of lin stack
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_get_signal_raw(
			    cs_lin_t *sk,           /**< Socket struct pointer*/
			    cs_lin_signal_t * sig,  /**< Signal pointer*/
			    cs_lin_raw_t *value,    /**< Signal value array*/
			    int len                 /**< Value array len*/
			    );
  
  /** Get enhanced or classic flag out of frame
      \retval 0 classic frame
      \retval 1 enhanced frame
  */
  int cs_lin_frame_is_enhanced(
			       cs_lin_frame_t * frame
			       );
  
  /********************************************************************/
  /********************************************************************/
  /** LDF specific functions*/
  /********************************************************************/
  /********************************************************************/
  
  /** Init structure
      \retval LDF storage structure
      \retval NULL Failure
  */
  cs_lin_ldf_t* cs_lin_ldf_init();

  /** Free structure
   */
  void cs_lin_ldf_free(
		       cs_lin_ldf_t *ldf       /**< Storage structure*/
		       );

  /** Set general LDF information
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_ldf_set_general(
			     cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
			     double prot_version,       /**< Protocol version*/
			     double lang_version,       /**< Language version*/
			     uint32_t speed             /**< Baud rate in bits/s*/
			     );
  
  /** Allocate an encoding instance. Note do not free this pointer!
      \retval instace if encoding does not exist it will be inernally allocated.
      \retval NULL Failure
  */
  cs_lin_enc_t* cs_lin_ldf_get_encoding(
				       cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
				       const char *name,          /**< Encoding name*/
				       cs_lin_enc_type_t basetype /**< Encoding base type*/ 
				       );


  /** Add physical encoding directive to encoding definition
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_ldf_encoding_add_physical(
				       cs_lin_enc_t* enc,      /**< Encoding structure*/
				       cs_lin_raw_t min,       /**< Minimal raw value*/
				       cs_lin_raw_t max,       /**< Maximal raw value*/
				       double scale,           /**< Scale value*/
				       double offset,          /**< Offset value*/
				       const char *unit        /**< Unit*/
				       );
  
  /** Add logical encoding directive to encoding definition
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_ldf_encoding_add_logical(
				      cs_lin_enc_t* enc,      /**< Encoding structure*/
				      cs_lin_raw_t value,     /**< Raw value*/
				      const char *text        /**< Text*/
				      );


  /** Allocate an signal encoding representation. Note do not free this pointer!
      \retval instace if representation does not exist it will be inernally allocated.
      \retval NULL Failure
  */
  cs_lin_enc_representation_t* cs_lin_ldf_get_signalrepresentation(
								   cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
								   const char *name          /**< Signal representation name*/
								   );

  /** Add signal to signal representation
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_ldf_signalrepresentation_add_signal(
						 cs_lin_enc_representation_t* rep,      /**< Encoding structure*/
						 const char *name                       /**< Signal name*/
						 );


  /** Allocate an signal definition. Note do not free this pointer!
      \retval instace if representation does not exist it will be internally allocated.
      \retval NULL Failure
  */
  cs_lin_signal_def_t* cs_lin_ldf_get_signaldefinition(
						       cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
						       const char *name,          /**< Signal representation name*/
						       cs_lin_raw_t *init,        /**< Signal initial values*/
						       int initlen,               /**< Init length in byte*/
						       int len,                   /**< Signal length in bits*/
						       const char *publisher      /**< Signal publisher*/
						       );
  

  /** Add subscriber to signal definition
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_ldf_signaldefinition_add_subscriber(
						 cs_lin_signal_def_t* sig,      /**< Signal definition structure*/
						 const char *subscriber         /**< Subscriber name*/
						 );

  /** Allocate a frame definition. Note do not free this pointer!
      \retval instace if representation does not exist it will be internally allocated.
      \retval NULL Failure
  */
  cs_lin_frame_t* cs_lin_ldf_get_frame(
				       cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
				       const char *name,          /**< Frame name*/
				       uint32_t id,               /**< Frame id*/
				       int dlc,                   /**< Frame DLC in bytes*/
				       const char *publisher      /**< Frame publisher node name*/
				       );

  /** Add signal to frame definition
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_ldf_frame_add_signal(
				  cs_lin_frame_t* frame,         /**< Frame definition structure*/
				  const char *name,              /**< Signal name*/
				  int offset                     /**< Signal bit offset in frame*/
				  );

  /** Allocate a scheduler table definition. Note do not free this pointer!
      \retval instace if representation does not exist it will be internally allocated.
      \retval NULL Failure
  */
  cs_lin_schedule_table_t* cs_lin_ldf_get_schedulertable(
							 cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
							 const char *name,          /**< Scheduler table name*/
							 int flag                   /**< 0=> Only serach no creation, 1=> search and creation*/
							 );

  /** Add Frame entry to scheduler table.
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_ldf_schedulertable_add_entry(
					  cs_lin_schedule_table_t *table,/**< Scheduler table definition structure*/
					  const char *name,              /**< Frame name*/
					  double delay                   /**< Delay in s*/
					  );

  /** Add node definition
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_ldf_add_node(
			  cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
			  int ismaster,              /**< 0=Slave, 1=Master*/
			  const char *name,          /**< Node name*/
			  double timebase,           /**< Time base in s, only for master node*/
			  double jitter              /**< Jitter in s, only for master*/
			  );

  /** Search node by name, Note do not free this pointer
      \retval instace  Node instance.
      \retval NULL Failure or not found
  */
  cs_lin_node_t* cs_lin_ldf_get_node(
				     cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
				     const char *name          /**< Node name*/
				     );

  /** Search master node, Note do not free this pointer
      \retval instace  Node instance.
      \retval NULL Failure or not found
  */
  cs_lin_node_t* cs_lin_ldf_get_master(
				       cs_lin_ldf_t *ldf         /**< LDF storage structure*/
				       );
  
  
  /** Allocate a node attribute definition. Note do not free this pointer!
      This functions defines only properties protocol version < 2.0
      \retval instace if representation does not exist it will be internally allocated.
      \retval NULL Failure
  */
  cs_lin_nodeattribute_t* cs_lin_ldf_get_nodeattribute(
						       cs_lin_ldf_t *ldf,         /**< LDF storage structure*/
						       const char *name,          /**< Node name*/
						       double version_protocol,   /**< Protocol version, eg. 2.0*/
						       uint32_t configured_nad,   /**< Configured NAD id*/
						       uint32_t initial_nad       /**< Initial NAD id*/
						       );

  /** Apply given LDF configuration and fill internal structures
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_ldf_apply_config(
			      cs_lin_ldf_t *ldf         /**< LDF storage structure*/
			      );


  /********************************************************************/
  /********************************************************************/
  /** LDF parser specific functions*/
  /********************************************************************/
  /********************************************************************/

  /** Initiliaze the LDF parser structure
      \retval NULL Failure
      \retval !=NULL No failure, Parser storage structure
  */
  cs_lin_ldf_doc_t* cs_lin_ldf_doc_init();
  
  
  /** Open and parse the LDF file
      \retval 0 No Failure
      \retval 1 Failure
  */
  int cs_lin_ldf_doc_open(
			  cs_lin_ldf_doc_t *data,      /**< LDF parser storage structure*/
			  char *filename               /**< LDF file name*/
			  );
  
  /** Close and release LDF file
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_ldf_doc_close(
			   cs_lin_ldf_doc_t *data        /**< LDF parser storage structure*/
			   );
  
  /** Release LDF storage structure
   */
  void cs_lin_ldf_doc_free(
			   cs_lin_ldf_doc_t *data
			   );
  
#ifdef __cplusplus
}
#endif

#endif
