/*  
    Copyright (C) 2017 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_lin_bit.h
 * @author Andreas Stewering-Bone
 * @date 9 March 2018
 * @brief File containing lin bus bit interface header.
 *
 * File containing lin bus bit interface header.
*/
#ifndef CS_LIN_BIT_H
#define CS_LIN_BIT_H

#include <inttypes.h>

#include "cs_debug.h"


/** \def CS_LIN_SYNC_DATA
    Synch data byte
*/
#define CS_LIN_SYNC_DATA 0x55
/** \def CS_LIN_PHYS_BITS
    Bits per byte
*/
#define CS_LIN_PHYS_BITS 8

/** \def CS_LIN_PHYS_TIME_BREAK
    Number of Bit times for break
*/
#define CS_LIN_PHYS_TIME_BREAK 13
/** \def CS_LIN_PHYS_TIME_BREAK_SCAN_MIN
    Minimal number of Bit times for break scan
*/
#define CS_LIN_PHYS_TIME_BREAK_SCAN_MIN 11
/** \def CS_LIN_PHYS_TIME_BREAK_SCAN_MAX
    Maximal number of Bit times for break scan
*/
#define CS_LIN_PHYS_TIME_BREAK_SCAN_MAX 15
/** \def CS_LIN_PHYS_TIME_BREAKDELIMITER
    Number of Bit times for break delimiter
*/
#define CS_LIN_PHYS_TIME_BREAKDELIMITER 1
/** \def CS_LIN_PHYS_TIME_BIT_TOL
    Bit time tolerance F_TOL_RES_SLAVE in %
*/
#define CS_LIN_PHYS_TIME_BIT_TOL 0.05
/** \def CS_LIN_PHYS_TIME_INTERBYTE
    Number of Bit times for inter byte space
*/
#define CS_LIN_PHYS_TIME_INTERBYTE 1
/** \def CS_LIN_PHYS_TIME_START
    Number of Bit times for start bit
*/
#define CS_LIN_PHYS_TIME_START 1
/** \def CS_LIN_PHYS_TIME_STOP
    Number of Bit times for stop bit
*/
#define CS_LIN_PHYS_TIME_STOP 1
/** \def CS_LIN_PHYS_TIME_DATA
    Number of Bit times for data bit
*/
#define CS_LIN_PHYS_TIME_DATA 1
/** \def CS_LIN_PHYS_TIME_BYTE
    Number of Bit times for data byte
*/
#define CS_LIN_PHYS_TIME_BYTE (CS_LIN_PHYS_TIME_START + 8 * CS_LIN_PHYS_TIME_DATA + CS_LIN_PHYS_TIME_STOP)

/** \struct cs_lin_phys_bit
    LIN Bit conversion Storage structure, predecleartion
*/
struct cs_lin_phys_bit;
/** \struct cs_lin_phys_bit_t
    LIN Bit conversion Storage structure
*/
typedef struct cs_lin_phys_bit cs_lin_phys_bit_t; 

#ifdef __cplusplus
extern "C" {
#endif

  /** Set debug level
      \retval Active Debuglevel
  */  
  cs_debug_level_t cs_lin_bit_debuglevel(
					 cs_debug_level_t level  /**< Debug level*/
					 );
  
  /** Set debugging callback routine
      Default: Function registered which prints debugging to STDOUT
  */
  void cs_lin_bit_debug_setcallback(
				    void *ctx,                             /**< Callback private data*/
				    void (*callb)(void *ctx,
						  cs_debug_sub_t sub,
						  cs_debug_level_t dl,
						  const char *func,
						  const int line,
						  const char *fmt,
						  va_list ap) /**< Callback routine*/
				    );
  
  /** Init LIN Bit structure
      Precalculate break and sync field
      \retval != NULL Data structure
      \retval NULL Failure
  */
  cs_lin_phys_bit_t* cs_lin_phys_bit_init(
					  uint32_t lin_baud,     /**< LIN-Baud rate in baud/s*/
					  uint32_t if_baud       /**< Interface Baud rate in baud/s*/
					  );

  /** Free structure
      \retval 1 Failure
      \retval 0 No Failure
  */
  int cs_lin_phys_bit_free(
			   cs_lin_phys_bit_t* bt         /**< Data structure*/
			   );
  
  /** Get the minimum interface data buffer size to handle a LIN frame with maximum length
    \retval >=0 Number of bytes
    \retval -1 Failure
 */ 
  int cs_lin_phys_bit_get_buffer_size(
				      cs_lin_phys_bit_t* bt         /**< Data structure*/
				      );

  /** Prepare a send interface send buffer
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_phys_bit_send(
			   cs_lin_phys_bit_t* bt,         /**< Data structure*/
			   uint8_t *bits,                 /**< Interface bit field*/
			   int *bitlen,                   /**< Bitbuffer len, will be overwritten by use of buffer*/
			   uint8_t pid,                   /**< Protected identifier*/
			   uint8_t *data,                 /**< LIN data*/
			   int datalen,                   /**< Lin data length*/
			   uint8_t crc                    /**< Check sum*/
			   );

  /** Prepare a send request interface send buffer
      \retval 0 No failure
      \retval 1 Failure
  */
  int cs_lin_phys_bit_send_req(
			       cs_lin_phys_bit_t* bt,         /**< Data structure*/
			       uint8_t *bits,                 /**< Interface bit field*/
			       int *bitlen,                   /**< Bitbuffer len in byte, will be overwritten by use of buffer*/
			       uint8_t pid                   /**< Protected identifier*/
			       );

  /** Scan bitfield for LIN Frames
      \retval -1 Failure
      \retval 0 Idle, nothing detected
      \retval 1 Frame content detected, but with unspecific state
      \retval 2 Frame header detected, without data
      \retval 3 Frame end detected, you can call the function with the same bitfield again
      \retval 4 Detect bit layer failure
  */
  int cs_lin_phys_lin_scan(
			   cs_lin_phys_bit_t* bt,         /**< Data structure*/
			   uint8_t *bits,                 /**< Input bitfield*/
			   int bitlen,                    /**< Bitfield length in byte*/
			   int *offset,                   /**< Startoffset of data segment, if we need a second call*/
			   uint8_t *scan_pid,             /**< Scan PID content*/
			   uint8_t *scan_data,            /**< Scan data field, assumes length CS_LIN_SIGNAL_MAX_BYTE*/
			   int *scan_length,              /**< Scan data length, min CS_LIN_SIGNAL_MAX_BYTE long*/
			   uint8_t *scan_crc               /**< Scan CRC field*/
			   );
  
#ifdef __cplusplus
}
#endif

#endif
