/*  
    Simple linked list implementation
    Copyright (C) 2015 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_lists.h
 * @author Andreas Stewering-Bone
 * @date 4 Nov 2015
 * @brief Simple linked list implementation.
 *
 * Simple linked list implementation.
 */

#ifndef CS_LISTS_H
#define CS_LISTS_H

#include <inttypes.h>

/** /struct cs_lists_entry
    List entry element
*/
struct cs_lists_entry{
  void *data;
  struct cs_lists_entry  *next;
  struct cs_lists_entry *prev;
};

/** /struct cs_lists_t
    List header
*/
typedef struct{
  void* (*copyfc)(void *ext);
  void (*delfc)(void *intern);
  struct cs_lists_entry *first;
  struct cs_lists_entry *end;
  uint64_t list_size;
}cs_lists_t;

#ifdef __cplusplus
extern "C" {
#endif

  /** Init list structure
      \retval 0 no failure
      \retval 1 failure
  */
  int cs_lists_init(
		    cs_lists_t *list,            /**< List structure*/
		    void* (*copyf)(void *ext), /**< If not NULL, when a new entry inserted, data is copied to an internal allocated space*/
		    void (*delf)(void *intern)   /**< If not NULL, if an element is deleted, the internal data is freed*/
		    );

  /** Insert a list element
      \retval 0 no failure
      \retval 1 failure
  */ 
  int cs_lists_insert(
		      cs_lists_t *list,            /**< List structure*/
		      void *data                   /**< Data pointer*/
		      );

  /** Delete a list element, identidied by data pointer
      \retval 0 no failure
      \retval 1 failure
  */ 
  int cs_lists_delete(
		      cs_lists_t *list,            /**< List structure*/
		      void *data                   /**< Data pointer*/
		      );


  /** Get number of list elements
      \retval number of elements
   */
  int cs_lists_size(
		    cs_lists_t *list            /**< List structure*/
		    );

  /** Get ith element data
    \retval NULL in failure case
  */
  void *cs_lists_get_iter(
			  cs_lists_t *list,     /**< List structure*/
			  int iter              /**< Element position*/
			  );

  /** Delete element
      \retval 0 no failure
  */
  int cs_lists_delete_iter(
			   cs_lists_t *list,     /**< List structure*/
			   int iter              /**< Element position*/
			   );

  /** Destroy list
      \retval 0 no failure
  */
  int cs_lists_destroy(
		       cs_lists_t *list          /**< List structure*/
		       );

  /** Reset list
      Remove all data items from list
      \retval 0 no failure
  */
  int cs_lists_reset(
		     cs_lists_t *list          /**< List structure*/
		     );

  /** Get first list element
      \retval first element
      \retval NULL in failure case
  */
  struct cs_lists_entry * cs_lists_get_first(
					     cs_lists_t *list /**< List structure*/
					     );
  /** Get next element
      \retval next element
      \retval NULL in failure case
  */
  struct cs_lists_entry * cs_lists_get_next(
					    struct cs_lists_entry *entry /**< List entry structure*/
					    );
  /** Get last element (end marker of list, no valid data)
      \retval last element
      \retval NULL in failure case
  */
  struct cs_lists_entry * cs_lists_get_last(
					    cs_lists_t *list /**< List entry structure*/
					    );

  /** Get element data pointer
      \retval data pointer
      \retval NULL in failure case
  */
  void* cs_lists_get_data(
			  struct cs_lists_entry *entry /**< List entry structure*/
			  );

#ifdef __cplusplus
}
#endif

#endif
