/*  message.h --  encode and decode CAN messages signals
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_message.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing an CAN message handling functions.
 *
 * File containing an file and syslog interface.
 */
#ifndef CS_CAN_MESSAGE_H
#define CS_CAN_MESSAGE_H

#include <inttypes.h>
#include <sys/time.h>

#include "cs_defines.h"
#include "cs_lists.h"
#include "cs_signal.h"
#include "cs_debug.h"

/** \enum cs_message_dir_t
    Flag for message direction
*/
typedef enum{
  MESSAGE_DIR_NONE,         /**< Not configured*/
  MESSAGE_DIR_RECEIVE,      /**< Receive message*/
  MESSAGE_DIR_SEND,         /**< Send message*/
  MESSAGE_DIR_SKIP          /**< Skip message*/
}cs_message_dir_t;

/** \enum cs_message_type_t
    Message type flag
*/
typedef enum{
  MESSAGE_TYPE_ASSEMBLED=0,  /**< Message assembled by signals*/
  MESSAGE_TYPE_DIAGNOSTIC,   /**< Diagnostic (UDS) message*/
  MESSAGE_TYPE_RAW,          /**< RAW Buffer*/
  MESSAGE_TYPE_DEFAULT       /**< Default is equal to RAW handling*/
}cs_message_type_t;

/** \enum cs_message_send_type_t
    Message send type flag
*/
typedef enum{
  MESSAGE_STYPE_CYCLIC=0,          /**< Cyclic send messages*/
  MESSAGE_STYPE_TRIGGERED,         /**< Triggered messages*/
  MESSAGE_STYPE_CYCLICENABLED,     /**< Cyclic messaged if subsystem is enabled*/
  MESSAGE_STYPE_CYCLICTRIGGERED,   /**< Cyclic messaged if subsystem is triggered*/
  MESSAGE_STYPE_CYCLICENABLEDTRIGGERED, /**< Cyclic messaged if subsystem is enabled and triggered*/
  MESSAGE_STYPE_NOTACTIVE               /**< Not active*/
}cs_message_send_type_t;

/** \enum cs_message_ptype_t
    Message type, physical layer
    CAN 2.0 vs CAN-FD
*/
typedef enum{
  MESSAGE_PTYPE_CAN = 0x00,      /**< CAN <= 2.0*/
  MESSAGE_PTYPE_CANFD = 0x01,    /**< CAN-FD*/
}cs_message_ptype_t;
  
/** \struct cs_message_t
    Message definition structure
*/
typedef struct cs_message cs_message_t;

#ifdef __cplusplus
extern "C" {
#endif

  /** Set debug level
      \retval Active Debuglevel
  */  
  cs_debug_level_t cs_message_debuglevel(
					 cs_debug_level_t level  /**< Debug level*/
					 );

  /** Set debugging callback routine
      Default: Function registered which prints debugging to STDOUT
  */
  void cs_message_debug_setcallback(
				    void *ctx,                             /**< Callback private data*/
				    void (*callb)(void *ctx,
						  cs_debug_sub_t sub,
						  cs_debug_level_t dl,
						  const char *func,
						  const int line,
						  const char *fmt,
						  va_list ap) /**< Callback routine*/
				    );

  /** Allocate and reset message structure
      \retval NULL Failure
  */
  cs_message_t* cs_message_init();

  /** Reset and free messages and depended signal data
   */
  void cs_message_free(
		       cs_message_t *m                  /**< Message structure*/
		       );
  
  /** Get number of signals
      \retval -1 Failure
      \retval >= 0 Number of signals
  */
  int cs_message_get_no_signals(
				cs_message_t *message  /**< Message structure*/
				);

  /** Set message id
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_message_set_id(
			cs_message_t *message,  /**< Message structure*/
			uint32_t id             /**< Message id*/
			);
  
  /** Get ID of message
      \retval 0 Failure
      \retval >0 ID
  */
  uint32_t cs_message_get_id(
			     cs_message_t *message  /**< Message structure*/
			     );

  /** Set message DLC
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_message_set_dlc(
			 cs_message_t *message,  /**< Message structure*/
			 uint8_t dlc             /**< Message DLC in Byte*/
			 );
  
  /** Get DLC of message
      \retval -1 Failure
      \retval >=0 DLC
  */
  int cs_message_get_dlc(
			 cs_message_t *message  /**< Message structure*/
			 );

  /** Set message name
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_message_set_name(
			  cs_message_t *message,  /**< Message structure*/
			  char * name             /**< Message name*/
			  );
  
  /** Get name of message
      Please note, the string is freed when cs_message_free is called
      \retval 0 Failure
      \retval !=0 Name
  */
  char *cs_message_get_name(
			    cs_message_t *message  /**< Message structure*/
			    );

  /** Set message producer node name
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_message_set_producer(
			      cs_message_t *message,  /**< Message structure*/
			      char * name             /**< Producer node name*/
			      );
  
  /** Get producer of message
      Please note, the string is freed when cs_message_free is called
      \retval 0 Failure
      \retval !=0 Name
  */
  char *cs_message_get_producer(
				cs_message_t *message  /**< Message structure*/
				);
  
  /** Get signal pointer from message
      \retval NULL failure
  */
  cs_signal_t* cs_message_get_signal_iter(
					  cs_message_t *message,  /**< Message structure*/
					  int iter                /**< Position*/
					  );

  /**< Get signal pointer from message, identified by signal name
     \retval 0 Failure
     \retval != 0 Signal pointer
  */
  cs_signal_t *cs_message_get_signal_byname(
					    cs_message_t *message,  /**< Message structure*/
					    char *slname            /**< Signal name*/
					    );

  /** Reset message structure
         \retval 0 success
	 \retval 1 Failure
  */  
  int cs_message_reset(
		       cs_message_t *message  /**< Message structure*/
		       );
  
  /** Build up message manually, firstly basic values
      Summary of _set functions
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_set(
		     cs_message_t *message, /**< Message structure*/
		     uint32_t id,           /**< CAN id*/
		     uint8_t dlc,           /**< DLC*/
		     char* name,            /**< Optional: message name*/
		     char* producer,        /**< Optional: producer name*/
		     uint32_t interval      /**< Send interval in 0.000001s*/
		     );


  /** Copy message structure, Note all internal data is copied. Only the basic configuration (DBC) is copied.
      All runtime dependent informations are not copied.
      \retval NULL Failure
  */
  cs_message_t *cs_message_cp(
			      cs_message_t *ext
			      );
  
  /** Configure send type of message
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_set_sendtype(
			      cs_message_t *message,   /**< Message structure*/
			      cs_message_send_type_t send_type /**< Send type*/
			      );


  /** Get message send type
      Note no failure handling
      \retval send type
  */
  cs_message_send_type_t cs_message_get_sendtype(
						 cs_message_t *message   /**< Message structure*/
						 );

  /** Set message physical layer format
      \retval 1 Failure
      \retval 0 Success
  */
  int cs_message_set_ptype(
			   cs_message_t *message,   /**< Message structure*/
			   cs_message_ptype_t ptype /**< Physical layer type for frame, CAN 2.0 vs CAN-FD*/
			   );

  /** Get message physical layer format
      Note no failre handling
      \retval layer format
  */
  cs_message_ptype_t cs_message_get_ptype(
					  cs_message_t *message   /**< Message structure*/
					  );
  
  /** Configure send timing parameter interval of message
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_set_interval(
			      cs_message_t *message,   /**< Message structure*/
			      uint32_t interval        /**< Send interval in 0.000001s*/
			      );

  /** Get message send interval in 0.000001s units
      TODO no failure handling
      \retval Send interval in 0.000001s
  */
  uint32_t cs_message_get_interval(
				   cs_message_t *message   /**< Message structure*/
				   );

  
  /** Configure send timing parameter interval offset of message
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_set_offset(
			    cs_message_t *message,   /**< Message structure*/
			    uint32_t offset          /**< Send cyclic offset in 0.000001s*/
			    );

  /** Get message send interval offset in 0.000001s units
      TODO no failure handling
      \retval Send interval offset in 0.000001s
  */
  uint32_t cs_message_get_offset(
				 cs_message_t *message   /**< Message structure*/
				 );
  
  /** Configure send timing parameter start offset of message
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_set_startoffset(
				 cs_message_t *message,   /**< Message structure*/
				 uint32_t offset          /**< Send cyclic offset in 0.000001s*/
				 );

  /** Get message send interval start offset in 0.000001s units
      TODO no failure handling
      \retval Send interval start offset in 0.000001s
  */
  uint32_t cs_message_get_startoffset(
				      cs_message_t *message   /**< Message structure*/
				      );
  

  /** Configure message receive timeout
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_set_timeout(
			     cs_message_t *message,  /**< Message structure*/
			     uint32_t timeout        /**< Receive timeout in 0.000001s*/
			     );

  /** Configure message receive timeout feedback flag
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_set_timeoutflag(
				 cs_message_t *message,  /**< Message structure*/
				 uint8_t *flag           /**< Pointer for timeout feedback 
							    0=> No timeout
							    1=> timeout*/
				 );

  /** Return message timeout value in 0.000001s resolution
      TODO no failure handling
      \retval timeout
  */
  uint32_t cs_message_get_timeout(
				  cs_message_t *message   /**< Message structure*/
				  );
  
  /** Configure message manually, add signal to message
      Change Version 2.0 to 2.1: Signal structure is internally copied
      \retval 0 success
  */
  int cs_message_add_signal(
			    cs_message_t *message, /**< Message structure*/
			    cs_signal_t *signal    /**< Signal structure*/
			    );


  /** Set Alive counter function callback for send messages,
      Note: Callback function has to store internal value by static values,
      identified by signal name
      \retval 0 success
  */
  int cs_message_set_signalcounter_cb(
				      cs_message_t *message,                 /**< Message structure*/
				      char *slname,                          /**< Signal name*/ 
				      void (*calc)(void* ctx,double *value, int reset), /**< Callback function*/
				      void *ctx                             /**< Callback context pointer*/
				      );

  /** Set Alive sounter check function callback for received messages, 
      Note: Callback function has to store internal value by static values,
      identified by signal name
      \retval 0 success
  */
  int cs_message_set_signalcheck_cb(
				    cs_message_t *message,                     /**< Message structure*/
				    char *slname,                              /**< Signal name*/
				    uint8_t (*check)(void *ctx,double *value, int reset),/**< Callback function*/
				    void *ctx,                                 /**< Call context pointer*/
				    uint8_t *flag                              /**< Failure flag,
										  if counter values do not equal*/
				    );

  /** Set signal default value for send messages,
      Value defined by number,
      values is preencoded to reduce load
      \retval 0 success
  */
  int cs_message_set_signal_default(
				    cs_message_t *message, /**< Message structure*/
				    char *slname,          /**< Signal name*/
				    double default_value   /**< Static default value*/
				    );

  /** Set signal default value for send messages,
      Value defined by label,
      values is preencoded to reduce load
      \retval 0 success
  */
  int cs_message_set_signal_defaultlabel(
					 cs_message_t *message, /**< Message structure*/
					 char *slname,          /**< Signal name*/
					 char *label            /**< Label*/
					 );

  /** Set signal mux type
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_message_set_signal_mux_type(
				     cs_message_t *message, /**< Message structure*/
				     char *slname,          /**< Signal name*/
				     cs_signal_mux_t mtype  /**< Signal mux type*/
				     );
  
  /** Set signal feedback pointer
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_set_signal_rvalue(
				   cs_message_t *message, /**< Message structure*/
				   char *slname,          /**< Signal name*/
				   double *rvalue,        /**< Signal feedback pointer*/
				   char *label,           /**< Signal label feedback*/
				   uint8_t *fault         /**< Signal fault feedback*/
				   );

  /** Set message default value for send messages,
      Value defined by raw buffer,
      \retval 0 success
  */
  int cs_message_set_defaultraw(
				cs_message_t *message,  /**< Message structure*/
				uint8_t *buffer,        /**< Default buffer*/
				int len                 /**< Buffer length*/
				);

  /** Get the content of the default_raw buffer
      Note buffer has to be minimum DLC length
      \retval 1 Failure
      \retval 0 Success
  */
  int cs_message_get_defaultraw(
				cs_message_t *message,  /**< Message structure*/
				uint8_t *buffer         /**< Default buffer*/
				);

  /** Set message value for send messages,
      Value defined by external raw buffer pointer,
      \retval 0 success
  */
  int cs_message_set_rawvalue(
			      cs_message_t *message, /**< Message structure*/
			      uint8_t *buffer,       /**< External message raw buffer*/
			      int len                /**< Buffer length*/
			      );

  /** Set signal crc callback,
      is called after signal calculation for sending,
      is called after signal decoding for receiving
      \retval 0 success
  */
  int cs_message_set_crc_cb(
			    cs_message_t *message, /**< Message structure*/
			    char *slname,                        /**< Signal name*/
			    int (*crc)(void *message, uint8_t *buffer, int len, double *value), /**< CRC Callback function*/
			    void *ctx,                           /**< Context pointer*/
			    uint8_t *crcflag                     /**< Feedback flag for receiving CRC check,
								    1=> calculated value does not equal received value*/
			    );

  /** Set message transport direction
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_set_direction(
			       cs_message_t *message,      /**< Message structure*/
			       cs_message_dir_t direction  /**< Transport direction*/
			       );

  /** Get message direction
      Note no failure handling
      \retval Message direction
  */
  cs_message_dir_t cs_message_get_direction(
					    cs_message_t *message      /**< Message structure*/
					    );

  /** Set message transport type
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_set_type(
			  cs_message_t *message,      /**< Message structure*/
			  cs_message_type_t mtype     /**< Transport type*/
			  );

  /** Get message direction
      Note no failure handling
      \retval Message direction
  */
  cs_message_type_t cs_message_get_type(
					cs_message_t *message      /**< Message structure*/
					);
  

  /** Set transport direction in  message array for all messages with defined producer
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_config_set_direction_byproducer(
						 cs_message_t *messages,     /**< Message structure array*/
						 int lt_messages,            /**< Array size*/
						 char *producer,             /**< Producer name*/
						 cs_message_dir_t direction  /**< Transport direction*/
						 );

  /** Set send and receive timestamp feedback
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_set_timestamp(
			       cs_message_t *message,  /**< Message structure array*/
			       struct timeval *ts      /**< Feedack timestamp*/
			       );

  /** Set send and receive timestamp double feedback
      Receive time in s
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_set_timestamp_d(
				 cs_message_t *message,  /**< Message structure array*/
				 double *tvh             /**< Feedack timestamp*/
				 );

  /** Set raw message buffer feedback value
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_set_rawbuffer(
			       cs_message_t *message,  /**< Message structure*/
			       uint8_t *buffer         /**< Feedback buffer, min size CS_DATA*/
			       );

  /** Get signal raw value pointer (encoded value), used in CRC callback functions
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_runtime_get_signalrawvalue(
					   cs_message_t *message, /**< Message structure*/
					   char *slname,          /**< Signal name*/
					   int64_t** rawvalue     /**< Signal raw value pointer*/
					   );

  /** Set signal raw value value, used for overwrite signal encoded value
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_runtime_set_signalraw(
				      cs_message_t *message, /**< Message structure*/
				      char *slname,          /**< Signal name*/
				      int64_t value          /**< Signal raw value*/
				      );

  /** Set signal encoding back to default behavior, used after cs_message_runtime_setsignalraw
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_runtime_set_signalstd(
				      cs_message_t *message,  /**< Message structure*/
				      char *slname            /**< Signal name*/
				      );

  /** After configure all messages this function checks the configuration,
      for a sending message, if all signal or raw value pointer are placed,
      so message content can be calced correctly
      Note this has to be called before message en- or decoding is called
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_check(
		       cs_message_t *message    /**< Message structure*/
		       );

  /** Decode received rawmessage to signal content
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_runtime_decode(
				cs_message_t *message, /**< Message structure*/
				uint8_t *buffer,       /**< Received raw buffer*/
				int len                /**< Raw buffer length*/
				);

  /** Encode signal content and place it in a message raw buffer
      \retval 0 success
      \retval 1 Failure
  */
  int cs_message_runtime_encode(
				cs_message_t *message, /**< Message structure*/
				uint8_t *buffer,       /**< Send raw buffer*/
				int len,               /**< Raw buffer length*/
				int init               /**< Flag for signal preencoding
							  >0 all signels set to static default values
							  will be preencoded*/
				);

#ifdef __cplusplus
}
#endif

#endif
