/*  
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_parser.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing the low level parsing API.
 *
 * File containing the low level parsing API.
*/
#ifndef CS_PARSER_H
#define CS_PARSER_H

#include <pcre.h>
#include <stdio.h>

/** \def CS_REGEX_BUFFER
    Maximal regex buffer length
*/
#define CS_REGEX_BUFFER 1024

/** \def CS_REGEX_RESULTSIZE
    Minimal increment for regex result vector
*/
#define CS_REGEX_RESULTSIZE 3

#define CS_PARSER_CHARBUF 255

/** \struct cs_parser_result_t
    Storage structure for regex result
*/
typedef struct{
  char *searchin;    /**< Where to search */
  int searchlen;     /**< search string length*/
  char *result;      /**< result srorage pointer*/
  int resultlen;     /**< result string len*/
  int resultoffset;  /**< Offset in search string*/
  char *next;        /**< Pointer to first char behind result string*/
  int nextlen;       /**< number of chars behind result string in searchin*/
}cs_parser_result_t;

/** \struct cs_parser_regex_t
    Storage structure for abstract regex usage
*/
typedef struct {
  cs_parser_result_t *search; /**< Serach area information*/
  pcre *re;          /**< Compiled regex storage pointer*/
  int noresult;      /**< Number of expected results*/
  int *ovector;      /**< result pointer array*/
}cs_parser_regex_t;

#ifdef __cplusplus
extern "C" {
#endif

  /** Open file and map it
      \retval 0 success
  */
  /**************************************************************************/
  int cs_parser_open(
		     char*filename,   /**< File name*/
		     FILE **fs,       /**< Pointer to file descriptor*/
		     long *filelen,   /**< File length*/
		     void **map       /**< Mapping pointer*/
		     );
  /**************************************************************************/
  
  /** Unmap file andclose filehandle
    \retval 0 success
  */
  /**************************************************************************/
  int cs_parser_close(
		      FILE * fs,    /**< File descriptor*/
		      void *map,    /**< Mapping pointer*/
		      long filelen  /**< File length*/
		      );
  /**************************************************************************/
  
  /** Reset result structure
   */
  /**************************************************************************/
  void cs_parser_resetresult(
			     cs_parser_result_t *result /**< Result pointer*/
			     );
  /**************************************************************************/
  
  /** Calculate next values for search loops
      \retval 0 success
  */
  /**************************************************************************/
  int cs_parser_calcresult(
			   cs_parser_result_t *result /**< Result pointer*/
			   );
  /**************************************************************************/

  /** Step to after retult at index
   */
  /**************************************************************************/
  int cs_parser_stepnext(
			 cs_parser_result_t *result, /**< Result pointer*/
			 int *ovector,               /**< Offset pointer vector*/
			 int vsize,                  /**< Size of vector*/
			 int index                   /**< Index of result*/
			 );
  /**************************************************************************/
  
  /** Printout regex result
   */
  /**************************************************************************/
  void cs_parser_printresult(
			     cs_parser_result_t *result  /**< Result pointer*/
			     );
  /**************************************************************************/

  /** Printout regex search area
   */
  /**************************************************************************/
  void cs_parser_printsearch(
			     cs_parser_result_t *result  /**< Result pointer*/
			     );
  /**************************************************************************/
  
  /** Precompile regex
      \retval Null if there is an failue
      \retval Pointer to regex structure
  */
  /**************************************************************************/
  pcre *cs_parser_precompile(
			     const char *pattern /**< Regex string*/
			     );
  /**************************************************************************/
  
  /**< get named substring from regex result structure
     \retval 0 success
  */
  /**************************************************************************/
  int cs_parser_getsubstring(
			     pcre *re,        /**< PCRE regex structure*/
			     char *searchin,  /**< Rawbuffer the regex was used for*/
			     int *ovector,    /**< PCRE regex result vector*/
			     int osize,       /**< PCRE regex result vector size*/
			     char *tag,       /**< Name of Tag you are searching for*/
			     char *substring  /**< Feedback tagcontent buffer, should be minimum
						 CS_PARSER_CHARBUF size*/
			     );

  /**************************************************************************/
  
  /**< get named substring from regex result structure, internal malloc of resultbuffer
     \retval 0 success
  */
  /**************************************************************************/
  int cs_parser_getsubstring_malloc(
				    pcre *re,        /**< PCRE regex structure*/
				    char *searchin,  /**< Rawbuffer the regex was used for*/
				    int *ovector,    /**< PCRE regex result vector*/
				    int osize,       /**< PCRE regex result vector size*/
				    char *tag,       /**< Name of Tag you are searching for*/
				    char **substring /**< Feedback tagcontent buffer, has to be released after usage*/
				    );

  /**************************************************************************/

  /**< get named substring from regex result structure, convert to int type
     \retval 0 success
  */
  /**************************************************************************/
  int cs_parser_getsubstring_int(
				 pcre *re,        /**< PCRE regex structure*/
				 char *searchin,  /**< Rawbuffer the regex was used for*/
				 int *ovector,    /**< PCRE regex result vector*/
				 int osize,       /**< PCRE regex result vector size*/
				 char *tag,       /**< Name of Tag you are searching for*/
				 int *value,      /**< Int return value*/
				 int dvalue       /**< Default return value*/
			     );
  /**************************************************************************/

  /**< get named substring from regex result structure, convert from hex to int type
     \retval 0 success
  */
  /**************************************************************************/
  int cs_parser_getsubstring_hex(
				 pcre *re,        /**< PCRE regex structure*/
				 char *searchin,  /**< Rawbuffer the regex was used for*/
				 int *ovector,    /**< PCRE regex result vector*/
				 int osize,       /**< PCRE regex result vector size*/
				 char *tag,       /**< Name of Tag you are searching for*/
				 int *value,      /**< Int return value*/
				 int dvalue       /**< Default return value*/
			     );
  /**************************************************************************/
    
  /**< get named substring from regex result structure, convert to double type
     \retval 0 success
  */
  /**************************************************************************/
  int cs_parser_getsubstring_double(
				    pcre *re,        /**< PCRE regex structure*/
				    char *searchin,  /**< Rawbuffer the regex was used for*/
				    int *ovector,    /**< PCRE regex result vector*/
				    int osize,       /**< PCRE regex result vector size*/
				    char *tag,       /**< Name of Tag you are searching for*/
				    double *value,   /**< Double return value*/
 				    double dvalue    /**< Default return value*/
				    );
  /**************************************************************************/

  /** Serach for argument line
      \retval 1 Failure
  */
  /**************************************************************************/
  int cs_parser_search_stringline(
				  cs_parser_result_t *result,   /**< Area to search in*/
				  char *name                    /**< Primary argument name*/
				  );
  /**************************************************************************/
    
  /** Prepare searchstring for regex, escape specific chars
      \retval 1 Failure
  */
  /**************************************************************************/
  int cs_parser_preparestring(
			      char *raw,                        /**< Input string*/
			      char *clean,                      /**< Output buffer*/
			      int BUFLEN                        /**< Output buffer size*/
			      );
  /**************************************************************************/

  /** Labeled regex based parsing, Prepare and execute regex
      \retval NULL Failure
      \retval !=NULL result storage structure
  */
      
  /**************************************************************************/
  cs_parser_regex_t* cs_parser_regex(
				     cs_parser_result_t *result,   /**< Area to search in*/
				     char *regex,                  /**< Regex with labels*/
				     int noresult                  /**< Number of expected result sets*/
				     );
  /**************************************************************************/

  /** Release parser structure
   */
  /**************************************************************************/
  void cs_parser_regex_free(
			    cs_parser_regex_t* st              /**< Parser storage structure*/
			    );
  /**************************************************************************/

  /** Step to the next searchable area
      \return 1 There is space for searching
      \return 0 Nothing to search
  */
  /**************************************************************************/
  int cs_parser_regex_next(
			   cs_parser_regex_t* st,             /**< Parser storage structure*/
			   cs_parser_result_t *next           /**< Area to next search in*/
			   );
  /**************************************************************************/

  /** Get labeled substring, identified by label
      \retval 0 No failure
      \retval 1 Failure
  */
  /**************************************************************************/
  int cs_parser_regex_getStr(
			     cs_parser_regex_t* st,            /**< Parser storage structure*/
			     char *label,                      /**< Label of substring*/
			     char **value                      /**< Result storage pointer, note memory has to be free after usage*/
			     );
  /**************************************************************************/

  /** Get labeled int value
      \retval 0 No failure
      \retval 1 Failure
  */
  /**************************************************************************/
  int cs_parser_regex_getInt(
			     cs_parser_regex_t* st,            /**< Parser storage structure*/
			     char *label,                      /**< Label of value*/
			     int *value,                       /**< Value pointer*/
			     int defaultvalue                  /**< Default value, if label is not found*/
			     );
  /**************************************************************************/

  /** Get labeled hex value
      \retval 0 No failure
      \retval 1 Failure
  */
  /**************************************************************************/
  int cs_parser_regex_getHex(
			     cs_parser_regex_t* st,            /**< Parser storage structure*/
			     char *label,                      /**< Label of value*/
			     int *value,                       /**< Value pointer*/
			     int defaultvalue                  /**< Default value, if label is not found*/
			     );
  /**************************************************************************/

  /** Get labeled double value
      \retval 0 No failure
      \retval 1 Failure
  */
  /**************************************************************************/
  int cs_parser_regex_getDouble(
				cs_parser_regex_t* st,            /**< Parser storage structure*/
				char *label,                      /**< Label of value*/
				double *value,                    /**< Value pointer*/
				double defaultvalue               /**< Default value, if label is not found*/
				);
  /**************************************************************************/
  

#ifdef __cplusplus
}
#endif

#endif
