/*
    Copyright (C) 2019 IgH, Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
/**
 * @file cs_process.h
 * @author Andreas Stewering-Bone
 * @date 08 Jan 2019
 * @brief File containing restbus simulation process handling functions.
 *
 * File containing restbus simulation process handling functions.
 */
#ifndef CS_PROCESS_H
#define CS_PROCESS_H

#include <inttypes.h>
#include <time.h>
#include <stdio.h>
#include <string.h>

#include "cs_debug.h"


/** \struct cs_process_t
    Process definition structure
*/
typedef struct {
  double sampletime;
  struct timespec start_time;
  struct timespec start_time_old;
  struct timespec end_time;
  struct timespec exec_time;
  struct timespec cycle_time;
  struct timespec sleep_time;
  struct timespec wakeup_time;
  struct timespec external_time;
  int overruns;
  int overruns_allowed;
  double dbltime;
  double cycle_time_d;
  double exec_time_d;
}cs_process_t;

#ifdef __cplusplus
extern "C" {
#endif


  /** Set debug level
      \retval Active Debuglevel
  */  
  cs_debug_level_t cs_process_debuglevel(
					 cs_debug_level_t level  /**< Debug level*/
					 );
  
  /** Set debugging callback routine
      Default: Function registered which prints debugging to STDOUT
  */
  void cs_process_debug_setcallback(
				    void *ctx,                             /**< Callback private data*/
				    void (*callb)(void *ctx,
						  cs_debug_sub_t sub,
						  cs_debug_level_t dl,
						  const char *func,
						  const int line,
						  const char *fmt,
						  va_list ap) /**< Callback routine*/
				    );
    
  /** Process relative sleep for given timeout
   */
  void cs_process_sleep(
			uint32_t sleeptime /**< Sleep time in ms*/
			);

  
  /** Allocate process structure
      \retval NULL Failure
      \retval != Process structure
  */
  cs_process_t *cs_process_init();
  
  /** Free process structure
   */
  void cs_process_free(
		      cs_process_t *process /**< Process structure*/
		      );

  /** Reset process data structure
      \retval 1 Failure
      \retval 0 Success
  */
  int cs_process_reset(
		       cs_process_t *process /**< Process structure*/
		       );

  /** Prepare call process
      Precalc sleep times
      Set process realtime priority
      Set memory lock
      \retval 1 Failure
      \retval 0 Success
  */
  int cs_process_prepare(
			 cs_process_t * process,  /**< Process structure*/
			 int priority,            /**< Realtime priority, see sched.h*/
			 double sampletime,       /**< Sample time step in s*/
			 int overruns_allowed     /**< Number of allowed overruns, 
						     if value is >= 0 you will get an failure
						     if an oerrun accurs, a value of -1 disables
						     the functionality*/
			 );
 
  /** Sleep process until scheduling time is reached
   */
  void cs_process_cycle_start(
			      cs_process_t * process  /**< Process structure*/
			      );
  /** Check realtime timing of call cycle
      \retval 0 Success, timing is ok
      \retval 1 General failure
      \retval 2 An overrun is accured
      \retval 3 An overrun is accured and we have reached 
      the maximal allowed numbers of overruns
  */
  int cs_process_cycle_end(
			   cs_process_t * process  /**< Process structure*/
			   );

  
#ifdef __cplusplus
}
#endif
  
#endif
