/*  
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @file cs_schedule.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing restbus simulation message scheduling functions.
 *
 * File containing restbus simulation message scheduling functions.
 */
#ifndef CS_SCHEDULE_H
#define CS_SCHEDULE_H

#include <inttypes.h>
#include "cs_message.h"
#include "cs_socketcan.h"
#include "cs_debug.h"

/** \enum cs_schedule_msgstate_t
    Scheduler message state
*/
typedef enum{
  SCHEDULE_STATE_IDLE = 0, /**< Message is idle, no recv, no sending*/
  SCHEDULE_STATE_RECIEVED, /**< Message was received*/
  SCHEDULE_STATE_SEND      /**< Message was send*/
}cs_schedule_msgstate_t;

/** \enum cs_schedule_msgaction_t
    Message action command
*/
typedef enum{
  SCHEDULE_MSGACTION_PAUSED = 0, /**< Pause message (set idle)*/
  SCHEDULE_MSGACTION_RUN          /**< Activate message (set receive or send)*/
}cs_schedule_msgaction_t;

/** \enum cs_scheduler_mode_t
    Scheduler call mode
*/
typedef enum{
  SCHEDULE_MODE_IDLE = 0, /**< Scheduler inactice*/ 
  SCHEDULE_MODE_SEND,     /**< Scheduler send signals*/
  SCHEDULE_MODE_RECIEVE   /**< Scheduler receive signals*/
} cs_scheduler_mode_t;


/** \struct cs_schedule_t
    Scheduler definition structure
*/
typedef struct cs_schedule cs_schedule_t;

#ifdef __cplusplus
extern "C" {
#endif

  /** Set debug level
      \retval Active Debuglevel
  */  
  cs_debug_level_t cs_schedule_debuglevel(
					  cs_debug_level_t level  /**< Debug level*/
					  );

  /** Set debugging callback routine
      Default: Function registered which prints debugging to STDOUT
  */
  void cs_schedule_debug_setcallback(
				     void *ctx,                             /**< Callback private data*/
				     void (*callb)(void *ctx,
						   cs_debug_sub_t sub,
						   cs_debug_level_t dl,
						   const char *func,
						   const int line,
						   const char *fmt,
						   va_list ap) /**< Callback routine*/
				     );
  
  /** Set message action for a send message (pause, run)
      \retval 0 success
  */
  int cs_schedule_set_msgaction(
				cs_schedule_t *scheduler,      /**< Scheduler structure*/
				uint32_t id,                   /**< CAN id*/
				cs_schedule_msgaction_t action /**< Message action*/
				);

  /** Set message action for all messages in send list (pause, run)
      \retval 0 success
  */
  int cs_schedule_set_msgaction_all(
				cs_schedule_t *scheduler,      /**< Scheduler structure*/
				cs_schedule_msgaction_t action /**< Message action*/
				);

  /** Get actual message state, (received, timeout, idle) for a received message
      \retval Message state
  */
  cs_schedule_msgstate_t cs_schedule_get_msgstate(
						  cs_schedule_t *scheduler, /**< Scheduler structure*/
						  uint32_t id               /**< CAN id*/
						  );

  /** Init scheduler structure
      \retval 0 Failure
      \retval !=0 Success
  */
  cs_schedule_t *cs_schedule_init();


  /** Free scheduler structure
   */
  void cs_schedule_free(
			cs_schedule_t *scheduler /**< Scheduler structure*/
			);


  /** Reset scheduler
      Reset message states and counter
      \retval 1 Failure
      \retval 0 Success
  */
  int cs_schedule_reset(
			cs_schedule_t *scheduler /**< Scheduler structure*/
			);

  /** Check if message exists in scheduler lists
      \retval 2 Message already exists in lists
      \retval 1 Failure
      \retval 0 Message does not exists in lists
  */
  int cs_schedule_exist_message(
				cs_schedule_t *scheduler, /**< Scheduler structure*/
				uint32_t id               /**< Message id*/
				);

  /** Add message to scheduler
      \retval 2 Message already exists in lists
      \retval 1 Failure
      \retval 0 Success
  */
  int cs_schedule_add_message(
			      cs_schedule_t *scheduler, /**< Scheduler structure*/
			      cs_message_t* message     /**< Message structure*/			      
			      );


  /** Set basic cycle time
      \retval 1 Failure
      \retval 0 Success
  */
  int cs_schedule_set_cycletime(
				cs_schedule_t *scheduler, /**< Scheduler structure*/
				uint32_t cycle            /**< Scheduler call cycle in 0.000001s*/
				);

  /** Set maximum allowed send messages per scheduler cylce
      \retval 1 Failure
      \retval 0 Success
  */
  int cs_schedule_set_maxpcycle(
				cs_schedule_t *scheduler, /**< Scheduler structure*/
				int  max_messpcycle       /**< Maximum number of messages send per one scheduler cycle, 
							     default 1 message per cycle*/
				);


  /** Open CAN interface
      \retval 1 Failure
      \retval 0 Success
  */
  int cs_schedule_open(
		       cs_schedule_t *scheduler, /**< Scheduler structure*/
		       char *iface,              /**< CAN interface name*/
		       cs_socketcan_mode_t mode  /**< SocketCAN mode*/
		       );


  /** Close CAN interface
      \retval 1 Failure
      \retval 0 Success
  */
  int cs_schedule_close(
			cs_schedule_t *scheduler /**< Scheduler structure*/
			);
  

  /** Check message configuration
      \retval 0 success
  */
  int cs_schedule_checkmessages(
				cs_schedule_t *scheduler /**< Scheduler structure*/
				);

  /** Scheduler call in realtime context,
      Send or receive messages
      \retval 0 success
  */
  int cs_schedule_schedule(
			   cs_schedule_t *scheduler,  /**< Scheduler structure*/
			   cs_scheduler_mode_t mode   /**< Scheduler call mode*/
			   );

  /** Get general send failure counter of physical layer
      Note, no failure handling
      \retval -1 Failure
      \retval >= 0 Send failure counter
  */
  int cs_schedule_get_send_errorcounter(
					cs_schedule_t *scheduler /**< Scheduler structure*/
					);
  
  /** Get TX-Buffer full  send failure counter of physical layer
      Indication for a configuration missmatch between busrate and message sending interval
      Please check your message settings
      \retval -1 Failure
      \retval >= 0 Send TX-Buffer failure counter
  */
  int cs_schedule_get_send_txfullerrorcounter(
					      cs_schedule_t *scheduler /**< Scheduler structure*/
					      );

  /** Get number of messages to be send in this cycle
      \retval -1 Failure
      \retval Number of messages to send
  */
  int cs_schedule_get_send_msgcounter(
				      cs_schedule_t *scheduler /**< Scheduler structure*/
				      );

  /** Get number of messages received in this cycle
      \retval -1 Failure
      \retval Number of messages received
  */
  int cs_schedule_get_recv_msgcounter(
				      cs_schedule_t *scheduler /**< Scheduler structure*/
				      );

  /** Get number of messages to be send but shifted to next schedule cycle
      Indication for a configuration missmatch between busrate, 
      maximum messages to be send in one cycle and messages sending interval.
      Please check your message settings
      \retval -1 Failure
      \retval >= 0 Shifted send message counter
  */
  int cs_schedule_get_send_msgshiftedcounter(
					     cs_schedule_t *scheduler /**< Scheduler structure*/
					     );
  
#ifdef __cplusplus
}
#endif

#endif
