/* 
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @file cs_signal.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing restbus simulation signal handling functions.
 *
 * File containing restbus simulation signal handling functions.
 */
#ifndef CS_CAN_SIGNAL_H
#define CS_CAN_SIGNAL_H

#include <inttypes.h>
#include <stdarg.h>

#include "cs_debug.h"

/** \enum cs_mux_t
    Signal category
*/
typedef enum{
  SIGNAL_MUX_SIGNAL = 0,    /**< Normal encoded signal*/
  SIGNAL_MUX_RAW,           /**< Signal defined by raw buffer*/
  SIGNAL_MUX_STATIC,        /**< Static signal, example start value defined in DBC, will be preencoded*/
  SIGNAL_MUX_CRC,           /**< Signal is a CRC checksum and has a callback function*/
  SIGNAL_MUX_COUNTER,       /**< Signal is a alive counter and has a callback function*/
  SIGNAL_MUX_MULTIPLEXOR,   /**< Signal is a multiplexor signal*/
  SIGNAL_MUX_MULTIPLEXED,   /**< Signal is a multiplexed signal*/
}cs_signal_mux_t;

/** \enum cs_type_t
    Signal type
*/
typedef enum{
  SIGNAL_TYPE_SIGNED = 0, /**< Signed value*/
  SIGNAL_TYPE_UNSIGNED,   /**< Unsigned value*/
  SIGNAL_TYPE_SINGLE,     /**< Single floating point value*/
  SIGNAL_TYPE_DOUBLE      /**< Double floating point value*/
}cs_signal_type_t;

/** \enum cs_byteorder_t
    Signal byteorder
*/
typedef enum{
  SIGNAL_BT_BIG = 0, /**< Big Endian*/
  SIGNAL_BT_LITTLE   /**< Little Endian*/
}cs_signal_byteorder_t;

/** \struct cs_signal_t
    Signal definition structure
*/
typedef struct cs_signal cs_signal_t;

#ifdef __cplusplus
extern "C" {
#endif

  /** Set debug level
      \retval Active Debuglevel
  */  
  cs_debug_level_t cs_signal_debuglevel(
					cs_debug_level_t level  /**< Debug level*/
					);

  /** Set debugging callback routine
      Default: Function registered which prints debugging to STDOUT
  */
  void cs_signal_debug_setcallback(
				   void *ctx,                             /**< Callback private data*/
				   void (*callb)(void *ctx,
						 cs_debug_sub_t sub,
						 cs_debug_level_t dl,
						 const char *func,
						 const int line,
						 const char *fmt,
						 va_list ap) /**< Callback routine*/
				   );


  /** Allocate signal structure
      \retval NULL Failure
      \retval != Signal structure
  */
  cs_signal_t *cs_signal_init();
  
  /** Free signal structure
   */
  void cs_signal_free(
		      cs_signal_t *signal /**< Signal structure*/
		      );
  
  /** Get signal name
      \retval NULL Failure
      \reval != NULL name
  */
  char *cs_signal_get_name(
			   cs_signal_t *signal /**< Signal structure*/
			   );

  /** Set signal name
      \retval 1 Failure
      \retval 0 No Failure
  */
  int cs_signal_set_name(
			 cs_signal_t *signal,  /**< Signal structure*/
			 char * name           /**< Signal name*/
			 );
  
  /** Get signal unit
      \retval NULL Failure
      \reval != NULL name
  */
  char *cs_signal_get_unit(
			   cs_signal_t *signal /**< Signal structure*/
			   );

  /** Set signal unit
      \retval 1 Failure
      \retval 0 No Failure
  */
  int cs_signal_set_unit(
			 cs_signal_t *signal,  /**< Signal structure*/
			 char * unit           /**< Signal unit*/
			 );
  
  /** Get signal note
      \retval NULL Failure
      \reval != NULL name
  */
  char *cs_signal_get_note(
			   cs_signal_t *signal /**< Signal structure*/
			   );

    /** Set signal note
      \retval 1 Failure
      \retval 0 No Failure
  */
  int cs_signal_set_note(
			 cs_signal_t *signal,  /**< Signal structure*/
			 char * note           /**< Signal note*/
			 );
  
  /** Get signal bit length in message
      TODO,no failure handling
  */
  uint16_t cs_signal_get_bitlength(
				   cs_signal_t *signal /**< Signal structure*/
				   );

  /** Set signal bit length
      \retval 1 Failure
      \retval 0 No Failure
  */
  int cs_signal_set_bitlength(
			      cs_signal_t *signal,  /**< Signal structure*/
			      uint16_t bitlength    /**< Signal bitlength*/
			      );
  
  /** Get signal bit offset in message
      TODO,no failure handling
  */
  uint16_t cs_signal_get_bitoffset(
				   cs_signal_t *signal /**< Signal structure*/
				   );

  /** Set signal bit offset in message
      \retval 1 Failure
      \retval 0 No Failure
  */
  int cs_signal_set_bitoffset(
			      cs_signal_t *signal,  /**< Signal structure*/
			      uint16_t bitoffset    /**< Signal bit offset*/
			      );
  
  /** Get signal scaling scale
      TODO,no failure handling
  */
  double cs_signal_get_scale(
			     cs_signal_t *signal /**< Signal structure*/
			     );

  /** Set signal scale
      \retval 1 Failure
      \retval 0 No Failure
  */
  int cs_signal_set_scale(
			  cs_signal_t *signal,  /**< Signal structure*/
			  double scale          /**< Signal scale*/
			  );
  
  /** Get signal scaling offset
      TODO,no failure handling
  */
  double cs_signal_get_offset(
			      cs_signal_t *signal /**< Signal structure*/
			      );

  /** Set signal offset
      \retval 1 Failure
      \retval 0 No Failure
  */
  int cs_signal_set_offset(
			   cs_signal_t *signal,  /**< Signal structure*/
			   double offset         /**< Signal offset*/
			   );
  
  /** Get signal min value (scaled values)
      TODO,no failure handling
  */
  double cs_signal_get_min(
			   cs_signal_t *signal /**< Signal structure*/
			   );

  /** Set signal min and max value
      \retval 1 Failure
      \retval 0 No Failure
  */
  int cs_signal_set_minmax(
			   cs_signal_t *signal,  /**< Signal structure*/
			   double min,           /**< Signal min value*/
			   double max            /**< Signal max value*/
			   );
  
  /** Get signal max value (scaled values)
      TODO,no failure handling
  */
  double cs_signal_get_max(
			   cs_signal_t *signal /**< Signal structure*/
			   );

  /** Set signal min and max value
      \retval 1 Failure
      \retval 0 No Failure
  */
  int cs_signal_set_minmax(
			   cs_signal_t *signal,  /**< Signal structure*/
			   double min,           /**< Signal min value*/
			   double max            /**< Signal max value*/
			   );
  
  /** Get signal type
      TODO,no failure handling
  */
  cs_signal_type_t cs_signal_get_signal_type(
				      cs_signal_t *signal /**< Signal structure*/
				      );

  /** Set signal type
      \retval 1 Failure
      \retval 0 No Failure
  */
  int cs_signal_set_signal_type(
				cs_signal_t *signal,   /**< Signal structure*/
				cs_signal_type_t stype /**< Signal type*/
				);
  
  /** Get signal byteorder
      TODO,no failure handling
  */
  cs_signal_byteorder_t cs_signal_get_byteorder(
						cs_signal_t *signal /**< Signal structure*/
						);

  /** Set signal byteorder
      \retval 1 Failure
      \retval 0 No Failure
  */
  int cs_signal_set_byteorder(
			      cs_signal_t *signal,   /**< Signal structure*/
			      cs_signal_byteorder_t bt /**< Signal byteorder*/
			      );
  
  /** Get signal mux type
      TODO,no failure handling
  */
  cs_signal_mux_t cs_signal_get_mux_type(
					 cs_signal_t *signal /**< Signal structure*/
					 );

  /** Set signal mux type
      \retval 1 Failure
      \retval 0 No Failure
  */
  int cs_signal_set_mux_type(
			     cs_signal_t *signal,   /**< Signal structure*/
			     cs_signal_mux_t mux    /**< Signal mux type*/
			     );

  /** Set signal start value
      Note it is a interger, because start values are stored in integer format in DBC-File
      Value is scaled internally to double startalue
      during encoding
      \retval 1 Failure
      \retval 0 Success
  */
  int cs_signal_set_startrawvalue(
				  cs_signal_t *signal,   /**< Signal structure*/
				  int64_t rawstartvalue  /**< Raw Start value*/
				  );

  /** Set signal start value
      \retval 1 Failure
      \retval 0 Success
  */
  int cs_signal_set_startvalue(
			       cs_signal_t *signal,   /**< Signal structure*/
			       double startvalue      /**< Raw Start value*/
			       );

  /** Get signal start raw value
      \retval Start raw value
  */
  double cs_signal_get_startvalue(
				  cs_signal_t *signal   /**< Signal structure*/
				  );
  
  /** Set basic values of signal
      \retval 1 Failure
      \retval 0 No Failure
  */  
  int cs_signal_set(
		    cs_signal_t *signal,   /**< Signal structure*/
		    char *name,            /**< Signal name*/
		    char *unit,            /**< Signal unit*/
		    uint16_t bitoffset,    /**< Bit offset in message*/
		    uint16_t bitlength,    /**< Bit length*/
		    cs_signal_byteorder_t bt, /**< Signal byte order*/
		    cs_signal_mux_t mux,  /**< Signal max type*/
		    cs_signal_type_t stype, /**< Signal signedness type*/
		    double scale,          /**< Signal scale*/
		    double offset,         /**< Signal offset*/
		    double min,            /**< Signal min value*/
		    double max             /**< Signal max value*/
		    );
  
  /** Reset signal data structure
      \retval 0 success
  */
  
  int cs_signal_reset(
		      cs_signal_t *signal  /**< Signal structure*/
		      );
  

  /** Copy signal structure without references
      \retval signal pointer
      \retval NULL failure
  */
  
  cs_signal_t *cs_signal_cp(
			    cs_signal_t *ext  /**< Signal structure*/
			    );
  

  /** 
      Check signal data and
      Calculate internal packaging data and signal definition checking
      Note this has to be called before signal en- or decoding is called
      If message length is given, signal length will be checked against message length
      \retval 0 Success
      \retval 1 Failure
   */
  
  int cs_signal_check(
		      cs_signal_t *signal, /**< Signal structure*/
		      uint8_t mlength      /**< Message length in bytes*/
		      );
   
  /** Set signal feedback value pointer
   */
  
  int cs_signal_set_rvalue(cs_signal_t *signal,  /**< Signal structure*/
			   double *value,        /**< Value pointer*/
			   char *label,          /**< Label pointer*/
			   uint8_t *fault        /**< Fault pointer*/
			   );
  

  /** Add label definition to signal
      \retval 0 success
  */
  
  int cs_signal_add_label(
			  cs_signal_t *signal, /**< Signal structure*/
			  char *label,         /**< Label*/
			  double value         /**< Numeric value*/
			  );
  

  /** Add signal requesting node definition to signal
      \retval 0 success
  */
  
  int cs_signal_add_node(
			 cs_signal_t *signal,  /**< Signal structure*/
			 char *name            /**< Node name*/
			 );
  

  /** Decode signal
      \retval 0 success
  */
  
  int cs_signal_runtime_decode(
			       cs_signal_t *signal,  /**< Signal structure*/
			       double *rvalue,       /**< Decoded value*/
			       uint8_t *fault,       /**< Fault flag, maybe NULL*/
			       uint8_t *buffer,      /**< Message raw buffer*/
			       int len               /**< Raw buffer length*/
			       );
  

  /** Do static preencoding of default values
      \retval 0 success
  */
  
  int cs_signal_runtime_staticencode(
				     cs_signal_t *signal,  /**< Signal structure*/
				     uint8_t *buffer,      /**< Message raw buffer*/
				     int len               /**< Raw buffer length*/
				     );
  

  /** Encode signal
      \retval 0 success
  */
  
  int cs_signal_runtime_encode(
			       cs_signal_t *signal,   /**< Signal structure*/
			       double value,          /**< Signal to encode*/
			       uint8_t *buffer,       /**< Message raw buffer*/
			       int len,               /**< Raw buffer length*/ 
			       int init               /**< >0 => preencode internal values*/
			       );
  

  /** Decode signal to label
      \retval 0 success
  */
  
  int cs_signal_runtime_decodelabel(
				    cs_signal_t *signal, /**< Signal structure*/
				    double value,       /**< Value to decode*/
				    uint8_t *fault,      /**< Feedback flag for failure*/
				    char *label          /**< Decoded label*/
				    );


  /** Decode label to value
      \retval 0 success
  */
  
  int cs_signal_runtime_getlabelvalue(
				      cs_signal_t *signal,   /**< Signal structure*/
				      double *value,         /**< Decoded value*/
				      char *label            /**< Label to decode*/
				      );
  


  /** Get number of defined labels
      \retval < 0 Failure
      \retval >= 0 Number of defined labels
  */
  
  int cs_signal_getnolabel(
			   cs_signal_t *signal    /**< Signal structure*/
			   );
  

  /** Get defined label by index
      \retval = 0 Success
      \retval 1 Failure
  */
  
  int cs_signal_getlabel_iter(
			      cs_signal_t *signal,    /**< Signal structure*/
			      int index,              /**< Label index*/
			      double *value,          /**< Value feedback pointer*/
			      char *label,            /**< Label buffer*/
			      int bufsize             /**< Label buffer size*/
			      );
  

  /** Get number of signal requesting nodes
      \retval < 0 Failure
      \retval >= 0 Number of defined nodes
  */
  
  int cs_signal_getnonodes(
			   cs_signal_t *signal    /**< Signal structure*/
			   );
  

  /** Get defined node by index
      \retval = 0 Success
      \retval 1 Failure
  */
  
  int cs_signal_getnode_iter(
			     cs_signal_t *signal,    /**< Signal structure*/
			     int index,              /**< Node index*/
			     char *node,             /**< Node name buffer*/
			     int bufsize             /**< Label buffer size*/
			     );
  
  
#ifdef __cplusplus
}
#endif

#endif
