/*  Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
#ifndef CS_SOCKETCAN_H
#define CS_SOCKETCAN_H

/**
 * @file cs_socketcan.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing SOCKETCAN abstraction layer.
 *
 * File containing SOCKETCAN abstraction layer.
 */

#include <inttypes.h>
#include <stdarg.h>
#include <time.h>
#include <sys/time.h>

#include "cs_debug.h"

/*Note keep values in sync with CS_OK and CS_FAIL*/
/** \def CS_SOCKETCAN_OK
    Everything is OK
*/
#define CS_SOCKETCAN_OK 0x00
/** \def CS_SOCKETCAN_FAIL
    A Failure happens
*/
#define CS_SOCKETCAN_FAIL 0x01
/** \def CS_SOCKETCAN_TXFULL
    TX-QUEUE is FULL
*/
#define CS_SOCKETCAN_TXFULL 0x02
/** \def CS_SOCKETCAN_ERROR
    Receive Error Frame
*/
#define CS_SOCKETCAN_ERROR 0x03
/** \def CS_SOCKETCAN_RTR
    Receive RTR Frame
*/
#define CS_SOCKETCAN_RTR 0x04
/** \def CS_SOCKETCAN_TIMEOUT
    A Recv timeout happens
*/
#define CS_SOCKETCAN_TIMEOUT 0x05

/** \def CS_SOCKETCAN_ERROR_NOERROR
    SocketCAN No Error
*/
#define CS_SOCKETCAN_ERROR_NOERROR 0x00
/** \def CS_SOCKETCAN_ERROR_SOCKET
    SocketCAN Error Socket error
*/
#define CS_SOCKETCAN_ERROR_SOCKET 0x01
/** \def CS_SOCKETCAN_ERROR_LOSTARB
    SocketCAN Error Lost arbitration
*/
#define CS_SOCKETCAN_ERROR_LOSTARB 0x02
/** \def CS_SOCKETCAN_ERROR_TXTIMEOUT
    SocketCAN Error TX timeout
*/
#define CS_SOCKETCAN_ERROR_TXTIMEOUT 0x03
/** \def CS_SOCKETCAN_ERROR_CTRL_UNKNOWN
    SocketCAN Controller Unspecific Error
*/
#define CS_SOCKETCAN_ERROR_CTRL_UNKNOWN 0x04
/** \def CS_SOCKETCAN_ERROR_CTRL_RXOVERFLOW
    SocketCAN Controller RX Overflow
*/
#define CS_SOCKETCAN_ERROR_CTRL_RXOVERFLOW 0x05
/** \def CS_SOCKETCAN_ERROR_CTRL_TXOVERFLOW
    SocketCAN Controller TX Overflow
*/
#define CS_SOCKETCAN_ERROR_CTRL_TXOVERFLOW 0x06
/** \def CS_SOCKETCAN_ERROR_CTRL_RXERROR
    SocketCAN Controller RX Error
*/
#define CS_SOCKETCAN_ERROR_CTRL_RXERROR 0x07
/** \def CS_SOCKETCAN_ERROR_CTRL_TXERROR
    SocketCAN Controller TX Error
*/
#define CS_SOCKETCAN_ERROR_CTRL_TXERROR 0x08
/** \def CS_SOCKETCAN_ERROR_TRX
    SocketCAN Transmitter failure
*/
#define CS_SOCKETCAN_ERROR_TRX 0x09
/** \def CS_SOCKETCAN_ERROR_ACK
    SocketCAN Acknoledge failure
*/
#define CS_SOCKETCAN_ERROR_ACK 0x0A
/** \def CS_SOCKETCAN_ERROR_PROT
    SocketCAN Protocol failure
*/
#define CS_SOCKETCAN_ERROR_PROT 0x0B
/** \def CS_SOCKETCAN_ERROR_BUSERROR
    SocketCAN Bus error
*/
#define CS_SOCKETCAN_ERROR_BUSERROR 0x0C
/** \def CS_SOCKETCAN_ERROR_BUSOFF
    SocketCAN Bus off
*/
#define CS_SOCKETCAN_ERROR_BUSOFF 0x0D
/** \def CS_SOCKETCAN_ERROR_PARAMETER
    User API Parameter failure
*/
#define CS_SOCKETCAN_ERROR_PARAMETER 0x0E
/** \def CS_SOCKETCAN_ERROR_MALLOC
    Internal malloc failure
*/
#define CS_SOCKETCAN_ERROR_MALLOC 0x0F
/** \def CS_SOCKETCAN_ERROR_FRAME
    Invalid CAN frame
*/
#define CS_SOCKETCAN_ERROR_FRAME 0x10
/** \def CS_SOCKETCAN_ERROR_MODE
    Invalid CAN mode
*/
#define CS_SOCKETCAN_ERROR_MODE 0x20
/** \def CS_SOCKETCAN_ERROR_BUFFER
    Invalid user buffer size
*/
#define CS_SOCKETCAN_ERROR_BUFFER 0x30
/** \def CS_SOCKETCAN_ERROR_ERRNO
    Failure value is accessible by errno
*/
#define CS_SOCKETCAN_ERROR_ERRNO 0x31

/** \enum cs_socketcan_mode_t
    SocketCAN running mode
*/
typedef enum{
  SOCKETCAN_MODE_CAN = 0x00,      /**< Only CAN <= 2.0*/
  SOCKETCAN_MODE_FLEX = 0x01,     /**< Try to use CAN-FD, fallback CAN2.0*/
  SOCKETCAN_MODE_CANFD = 0x02,    /**< Force using CAN-FD*/
}cs_socketcan_mode_t;


/** \enum cs_socketcan_send_mode_t 
    Send mode
    Also if CAN-FD is enabled, the sender can decide for every frame to send 
    in one of the specific modes.
*/ 
typedef enum{
  SOCKETCAN_SEND_CAN = 0x00,       /**< Only CAN <= 2.0*/
  SOCKETCAN_SEND_CANFD = 0x01,     /**< Force using CAN-FD*/
  SOCKETCAN_SEND_CANFD_BRS = 0x02, /**< Force using CAN-FD with second bitrate*/
}cs_socketcan_send_mode_t;




/** \struct cs_socketcan
    SocketCAN Storage structure, predecleartion
*/
struct cs_socketcan;
/** \struct cs_socketcan_t
    SocketCAN Storage structure
*/
typedef struct cs_socketcan cs_socketcan_t; 

#ifdef __cplusplus
extern "C" {
#endif

  /** Set debug level
      \retval Active Debuglevel
  */  
  cs_debug_level_t cs_socketcan_debuglevel(
					   cs_debug_level_t level  /**< Debug level*/
					   );

  /** Set debugging callback routine
      Default: Function registered which prints debugging to STDOUT
  */
  void cs_socketcan_debug_setcallback(
				      void *ctx,                             /**< Callback private data*/
				      void (*callb)(void *ctx,
						    cs_debug_sub_t sub,
						    cs_debug_level_t dl,
						    const char *func,
						    const int line,
						    const char *fmt,
						    va_list ap) /**< Callback routine*/
				      );
  

  /** Reset error counter
   */ 
  void cs_socketcan_errorcnt_reset(
				   cs_socketcan_t *sk  /**< Socket struct pointer*/
				   );

  /** Get actual error code 
      See CS_SOCKETCAN_ERROR_XXXXX
      \retval Error code
  */
  int cs_socketcan_getlasterror(
				cs_socketcan_t *sk    /**< Socket struct pointer*/
				);
  /** Get actual error code human readable
      \retvar char pointer
  */
  char* cs_socketcan_getlasterrorstring(
					cs_socketcan_t *sk  /**< Socket struct pointer*/
				      );
  
  /** Init structure
      \retval storage structure
      \retval 0 in failure case
  */
  cs_socketcan_t *cs_socketcan_init();

  /** Free structure
   */
  void cs_socketcan_free(
			 cs_socketcan_t *sk           /**< Socket struct pointer*/
			 );

  /** Check compability between libc version and kernel.
      Note best way to solve, Take actual distro, eg. Debian 9
      and a kernel > 4.2
      Date of writing 2018-12
      Background: struct can_frame and struct canfd_frame are not size compatible in old libc versions
      I do not think we need this function, maybe this will removed in the future.
      \retval -1 Failure
      \retval 0 Incompatible, old style libc header, and old kernel
      \retval 1 Incompatible, old style libc header, but new kernel
      \retval 2 Incompatible, new style libc header, but old kernel
      \ŗetval 3 Compatible, new style libc and new kernel
  */
  int cs_socketcan_checkcompat();
  
  /** Is socket  CAN connected
      \retval -1 Failure
      \retval 0 Not connected
      \retval 1 Connected
  */
  int cs_socketcan_isconnected(
			       cs_socketcan_t *sk            /**< Socket struct pointer*/
			       );

  /** Check if native socket device is CAN-FD capable
      \retval -1 Genaral failure
      \retval 0 Not CAN-FD capable
      \retval 1 CAN-FD capable
  */
  int cs_socketcan_check_canfd(
			       char *ifcan                  /**< Interface name*/
			       );

  /** Is CAN-FD enabled
      \retval -1 Failure
      \retval 0 No, disabled or not available
      \retval 1 Yes, enabled and active
  */
  int cs_socketcan_canfd_isenabled(
				   cs_socketcan_t *sk            /**< Socket struct pointer*/
				   );
  /** Maximum allowed DLC
      \retval 0 Failure
      \retval >0 DLC (8 for CAN 2.0 and 64 for CAN-FD)
  */
  int cs_socketcan_maxdlc(
			  cs_socketcan_t *sk            /**< Socket struct pointer*/
			  );
  
  /** Configure local socket
      \retval CS_SOCKETCAN_OK No failure
      \retval CS_SOCKETCAN_FAIL General Failure
  */
  int cs_socketcan_config_native(
				 cs_socketcan_t *sk,           /**< Socket struct pointer*/
				 char *ifcan,                  /**< Interface name*/
				 cs_socketcan_mode_t mode      /**< Enable CAN-FD for native socket,
								0 => Do not use CAN-FD,
							        1 => Use CAN-FD if available,
							        2 => Force use CAN-FD
							       */
				 );

  /** Configure SocketCANd interface
      \retval CS_SOCKETCAN_OK No failure
      \retval CS_SOCKETCAN_FAIL General Failure
  */
  int cs_socketcan_config_socketcand(
				     cs_socketcan_t *sk,       /**< Socket struct pointer*/
				     char *ifcan,              /**< Interface name*/
				     char *hostname,           /**< SocketCANd server hostname*/
				     int port                  /**< SocketCANd server port*/
				     );
  
  /** Open CAN interface
      \retval CS_SOCKETCAN_OK No failure
      \retval CS_SOCKETCAN_FAIL General Failure
  */
  int cs_socketcan_open(
			cs_socketcan_t *sk            /**< Socket struct pointer*/
			);
  
  /** Close CAN interface
      \retval CS_SOCKETCAN_OK No failure
      \retval CS_SOCKETCAN_FAIL General Failure
  */
  int cs_socketcan_close(
			 cs_socketcan_t *sk          /**< Socket struct pointer*/
			 );

  /** Get socket handle pointer.
      Used for higher level api like QT
      \retval -1 Failure
      \retval >= 0 Socket
  */
  int cs_socketcan_getsocket(
			     cs_socketcan_t *sk  /**< Socket struct pointer*/
			     );
  
  /** Set CAN receive filter
      \retval CS_SOCKETCAN_OK No failure
      \retval CS_SOCKETCAN_FAIL General Failure
  */
  int cs_socketcan_set_filter(
			      cs_socketcan_t *sk,      /**< Socket struct pointer*/
			      uint32_t *ids,           /**< Array of CAN ids*/
			      int num_ids              /**< Length of array*/
			      );

  /** CAN send
      \retval CS_SOCKETCAN_OK No failure
      \retval CS_SOCKETCAN_FAIL General Failure
  */
  int cs_socketcan_send(
			cs_socketcan_t *sk,           /**< Socket struct pointer*/
			uint32_t id,                  /**< CAN id*/
			uint8_t *buffer,              /**< Data buffer*/
			int len,                      /**< Data buffer length*/
			cs_socketcan_send_mode_t mode /**< Send mode CAN2.0 or CAN-FD*/
			);

  /** CAN send RTR
      \retval CS_SOCKETCAN_OK No failure
      \retval CS_SOCKETCAN_FAIL General Failure
  */
  int cs_socketcan_send_rtr(
			    cs_socketcan_t *sk,           /**< Socket struct pointer*/
			    uint32_t id                  /**< CAN id*/
			    );

  /** CAN receive non blocking
      \retval CS_SOCKETCAN_OK No failure
      \retval CS_SOCKETCAN_FAIL General Failure
      \retval CS_SOCKETCAN_ERROR Receive Error frame
      \retval CS_SOCKETCAN_RTR Receive RTR frame
      \retval CS_SOCKETCAN_TIMEOUT Receive timeout
  */
  int cs_socketcan_recv_nonblock(
				 cs_socketcan_t *sk,  /**< Socket struct pointer*/
				 uint32_t *id,        /**< CAN id storage pointer*/
				 uint8_t *buffer,     /**< Receive data buffer*/
				 int *len,            /**< Receive buffer length, should be initialized with min 8 Bytes*/
				 struct timeval *tv   /**< Receive timestamp pointer*/
				 );

  /** CAN receive blocking
      \retval CS_SOCKETCAN_OK No failure
      \retval CS_SOCKETCAN_FAIL General Failure
      \retval CS_SOCKETCAN_ERROR Receive Error frame
      \retval CS_SOCKETCAN_RTR Receive RTR frame
      \retval CS_SOCKETCAN_TIMEOUT Receive timeout
  */
  int cs_socketcan_recv_block(
			      cs_socketcan_t *sk,  /**< Socket struct pointer*/
			      uint32_t *id,        /**< CAN id storage pointer*/
			      uint8_t *buffer,     /**< Receive data buffer*/
			      int *len,            /**< Receive buffer length, should be initialized with min 8 Bytes*/
			      struct timeval *tv,  /**< Receive timestamp pointer*/
			      uint32_t timeout     /**< Timeout in ms*/
			      );

#ifdef __cplusplus
}
#endif
#endif
