/*  uds.h --  UDS (ISO 14229-1) Stack
    Copyright (C) 2013 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/**
 * @file cs_uds.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing the UDS basic low level functions.
 *
 * This file contain the UDS basic low level functions.
*/
#ifndef CS_UDS_H
#define CS_UDS_H

#include <inttypes.h>
#include <time.h>
#include <sys/time.h>

#include "cs_defines.h"
#include "cs_isotp.h"
#include "cs_uds_defines.h"
#include "cs_debug.h"

/** \struct cs_uds_service
    UDS Storage structure, predecleartion
*/
struct cs_uds_service;
/** \struct cs_uds_service_t
    UDS Storage structure
*/
typedef struct cs_uds_service cs_uds_service_t; 

/** \struct cs_uds_time_server
    Server storage for timing information
*/
struct cs_uds_time_server{
  uint16_t P2CANServer;       /**< Recv timeout in ms*/
  uint16_t P2CANServerE;      /**< Extended recv timeout in ms*/
};

/** \struct cs_uds_time_client
    Client storage for timing information
*/
struct cs_uds_time_client{
  uint16_t pcandelta;         /**< Delta*/
  uint16_t P2CANClient;       /**< P2 Timeout recv client in ms*/
  uint16_t P2CANClientE;      /**< P2 Timeout extended recv client in ms*/
  uint16_t P3CANClientPhys;   /**< P3 Timeout in ms*/
  uint16_t P3CANClientFunc;   /**< P3 Timeoutin ms*/
};


/** \enum
    Flag if response is requested
*/
typedef enum{
  RESPREQ=0,  /**< Response is requested*/
  RESPSUP     /**< Response is suppressed*/
} resp_t;

/** \struct cs_uds_diagsession_t
    Feedback storage structure for diagnostic session control
*/
typedef struct{
  int no;                                            /**< Number of response*/
  uint32_t sa[CS_UDS_MAXFUNC];                       /**< Array of recv source adresses*/
  struct cs_uds_time_server tserver[CS_UDS_MAXFUNC]; /**< Response timing information*/
}cs_uds_diagsession_t;


/** \struct cs_uds_dtc_reqdata_t
    DTC request structure,
    will be simplified in future
*/
typedef struct{
  uint8_t status_mask;        /**< Request status mask*/
  uint8_t severeity_mask;     /**< Request severeity mask*/
  uint32_t dtc_mask;          /**< Request DTC mask*/
  uint8_t extdata_recno;      /**< Extended data record number*/
  uint8_t storeddata_recno;   /**< Stored data record number*/
  uint8_t snapdata_recno;     /**< Snapshot data record number*/
  uint8_t memory_select;      /**< Memory segment selected*/
  uint8_t group_id;           /**< Group id*/
} cs_uds_dtc_reqdata_t;

/** \struct cs_uds_dtc_extdatarec_t
    Response data structure for DTC
    extended data handling
*/
typedef struct{
  uint8_t id;            /**< Extended ID*/
  uint8_t data[CS_UDS_MAXBYTE]; /**< Extended data buffer*/
  uint16_t lt;           /**< Extended data buffer length*/
} cs_uds_dtc_extdatarec_t;

/** \struct cs_uds_dtc_statusdtcrec_t
    Response DTC status record structure
*/
typedef struct{
  uint32_t dtc;                 /**< DTC Id*/
  uint8_t status;               /**< Status flag*/
  uint8_t severity;             /**< Severity flag*/
  uint8_t functionalunit;       /**< Functional unit*/
  uint8_t faultcounter;         /**< Fault counter*/
  uint8_t snaprecno;            /**< Number of Snapshot records*/
  uint8_t snaprecnoid;          /**< Snapshot record Id*/
} cs_uds_dtc_statusdtcrec_t;

/** \struct cs_uds_dtc_respdata_t
    DTC response data structure
*/
typedef struct{
  uint8_t avail_mask;          /**< Available mask*/
  uint8_t severity_mask;       /**< Severity mask*/
  uint8_t functionalgroup;     /**< Functional group*/
  uint8_t format;              /**< Format Id*/
  uint16_t count;              /**< Count, TODO description*/
  uint16_t lt_statusdtcrec;    /**< Status record length*/
  cs_uds_dtc_statusdtcrec_t statusdtcrec[CS_UDS_MAXDTC]; /** Status record data*/
  uint16_t lt_didrec;          /**< DID record length in byte*/
  uint16_t no_didrec;          /**< DID record length in DID entries*/
  uint8_t didrec[CS_UDS_MAXBYTE];  /**< DID record data*/
  uint16_t no_extdata;         /**< Number of extended data entries*/
  cs_uds_dtc_extdatarec_t extdata; /**< Extended data record*/
} cs_uds_dtc_respdata_t;

/** \struct cs_uds_client_flash_t
    Helper structure to get an nonblocking
    flashing handling
*/
typedef struct{
  cs_uds_service_t *udsdata;   /**< UDS stack data*/
  uint8_t SI;               /**< Set to CS_UDS_SI_RU or CS_UDS_SI_RD to 
			       show data flow direction*/ 
  uint8_t *transferdata;   /**< Pointer to transfer data*/
  uint64_t transferlen;    /**< Transfer data length*/
  uint8_t seqcounter;      /**< Sequence counter*/
  uint64_t maxblocklen;    /**< Max allowed block len*/
  uint64_t transfered;     /**< Already transfered bytes*/
}cs_uds_client_flash_t;


/** \struct c_uds_server_cb_t
    Server Callback function table
    Return value of each function is a CS_UDS_ERROR_* Value (cs_uds_defines.h),
    defined in the UDS-Standard
    If no failure happen, return CS_UDS_ERROR_PR
*/
typedef struct{
  /** Server callback diagsession*/
  int (*server_cb_diagsession)(void *srv_ctx,      /**< Server context pointer*/
			       uint8_t subfunc     /**< Subfunction called by client,  CS_UDS_SI_DSC_* (cs_uds_defines.h)*/
			       );
  /** Server callback ecu reset*/
  int (*server_cb_ecureset)(void *srv_ctx,         /**< Server context pointer*/
			    uint8_t subfunc,       /**< Subfunction called by client,  CS_UDS_SI_ER_* (cs_uds_defines.h)*/
			    uint8_t *powerDownTime /**< Return value, power down time*/
			    );
  /** Server callback Security access, get seed*/
  int (*server_cb_securityaccess_seed)(void *srv_ctx,      /**< Server context pointer*/
				       uint8_t subfunc,    /**< Subfunction and requesteted seed level*/
				       uint8_t *seed,      /**< Seed storage buffer*/
				       int *seedlen        /**< Seed buffer length pointer, initialize on max allowed buffer len*/
				       );
  /** Server callback Security access, set key*/
  int (*server_cb_securityaccess_key)(void *srv_ctx,      /**< Server context pointer*/
				      uint8_t subfunc,    /**< Subfunction and key level*/
				      uint8_t *key,       /**< Key buffer*/
				      int keylen          /**< Key buffer length*/
				      );
  /** Server callback Control DTC setting*/
  int (*server_cb_controldtcsetting)(void *srv_ctx,       /**< Server context pointer*/
				     uint8_t subfunc      /**< Subfunction called, CS_UDS_SI_CDTCS_* (cs_uds_defines.h)*/
				     );
  /** Server callback Link control*/
  int (*server_cb_linkcontrol)(void *srv_ctx,             /**< Server context pointer*/
			       uint8_t subfunc,           /**< Subfunction called CS_UDS_SI_LC_* (cs_uds_defines.h)*/
			       uint32_t data              /**< Optional data, subfunction depended*/
			       );
  /** Server callback Tester present*/
  int (*server_cb_testerpresent)(void *srv_ctx           /**< Server context pointer*/
				 );
  /** Server callback Read data by identifier*/
  int (*server_cb_readdatabyidentifier)(void *srv_ctx,    /**< Server context pointer*/
					uint16_t did,     /**< DID requested*/
					uint8_t *data,    /**< Raw data buffer*/
					int *datalen      /**< Raw data buffer length pointer, initialized on max allowd value*/ 
					);
  /** Server callback Write data by identifier*/
  int (*server_cb_writedatabyidentifier)(void *srv_ctx,   /**< Server context pointer*/
					 uint16_t did,    /**< DID to write*/
					 uint8_t *data,   /**< Raw data buffer*/
					 int datalen      /**< Raw data buffer length*/
					 );
  /** Server callback Routine control*/
  int (*server_cb_routinecontrol)(void *srv_ctx,         /**< Server context pointer*/
				  uint8_t subfunction,   /**< Subfunction called CS_UDS_SI_RC_* (cs_uds_defines.h)*/
				  uint16_t routine,      /**< Routine to call*/
				  uint8_t *reqbuffer,    /**< Request buffer data*/
				  int reqlen,            /**< Request buffer length*/
				  uint8_t *respbuffer,   /**< Reponse buffer*/
				  int *resplen,          /**< Response bufferlength pointer, initialized on max allowed value*/
				  uint8_t *routineinfo,  /**< Optional Routine info data*/
				  uint8_t *send_routineinfo/**< Flag point send info data, if value != 0, sen routine info*/
				  );
  /** Server callback Clear DTC information*/
  int (*server_cb_cleardtcinformation)(void *srv_ctx,     /**< Server context pointer*/
				       uint32_t mask      /**< DTC mask*/
				       );
  /** Server callback Read DTC information, get number of DTC by mask*/
  int (*server_cb_readdtcinformation_getnumberofdtcbymask)(void *srv_ctx,       /**< Server context pointer*/
							   uint8_t subfunction,   /**< Subfunction called CS_UDS_SI_RDTCI_* (cs_uds_defines.h)*/
							   uint8_t dtcstatusmask,/**< DTC status mask*/
							   uint8_t dtcseveritymask,/**< DTC severity mask*/
							   uint8_t *dtcavailmask,/**< DTC status available mask*/
							   uint8_t *format,     /**< Response format CS_UDS_DTC_FORMAT_* (cs_uds_defines.h)*/
							   uint16_t *count      /**< DTC count*/
							   );
    /** Server callback Read DTC information, get DTC by mask*/
  int (*server_cb_readdtcinformation_getdtcbymask)(void *srv_ctx,       /**< Server context pointer*/
						   uint8_t subfunction,   /**< Subfunction called CS_UDS_SI_RDTCI_* (cs_uds_defines.h)*/
						   uint8_t dtcstatusmask,/**< DTC status mask*/
						   uint8_t *dtcavailmask,/**< DTC status available mask*/
						   uint32_t *dtc,       /**< DTC buffer*/
						   uint8_t *status,     /**< Status buffer*/
						   int *count           /**< Length of buffer dtc and status, initialized on max length*/
						   );
    /** Server callback Read DTC information, report snapshot identification*/
    int (*server_cb_readdtcinformation_reportsnapshotidentification)(void *srv_ctx,       /**< Server context pointer*/
								     uint32_t *dtc,       /**< DTC buffer*/
								     uint8_t *status,     /**< Status buffer*/
								     int *count           /**< Length of buffer dtc and status, initialized on max length*/
						     );
    /** Server callback Read DTC information, report DTC by mask record*/
    int (*server_cb_readdtcinformation_reportdtcmaskrecord)(void *srv_ctx,       /**< Server context pointer*/
							    uint8_t subfunction,   /**< Subfunction called CS_UDS_SI_RDTCI_* (cs_uds_defines.h)*/
							    uint8_t dtcstatusmask,/**< DTC status mask*/
							    uint8_t dtcseveritymask,/**< DTC severity mask*/
							    uint32_t dtcmask,    /**< DTC mask*/
							    uint8_t *dtcavailmask,/**< DTC status available mask*/
							    uint8_t *serverity,  /**< Severity buffer*/
							    uint8_t *func,       /**< Functional buffer*/
							    uint32_t *dtc,       /**< DTC buffer*/
							    uint8_t *status,     /**< Status buffer*/
							    int *count           /**< Length of buffer, initialized on max length*/
							    );
    /** Server callback Upload/Download request*/
    int (*server_cb_uploaddownload)(void *srv_ctx,          /**< Server context pointer*/
				    uint8_t si,             /**< Service request Upload or Download CS_UDS_SI_RD or CS_UDS_SI_RU (cs_uds_defines.h)*/
				    uint8_t compression,    /**< Compression requested, default 0x00*/
				    uint8_t encyption,      /**< Encryption requested, default 0x00*/
				    uint64_t address,       /**< Target address*/
				    uint64_t len,           /**< Segment length in bytes*/
				    uint64_t *blocksize     /**< Transfer lock size in bytes*/
				    );
    /** Server callback Transfer data*/
    int (*server_cb_transfer)(void *srv_ctx,                /**< Server context pointer*/
			      uint8_t direction,            /**< Working direction CS_UDS_SI_RD or CS_UDS_SI_RU*/
			      uint8_t blockcounter,         /**< Request block counter*/
			      uint8_t *databuffer,          /**< Data buffer*/
			      int *datalen                  /**< Data length*/
			      );
    /** Server callback Transfer exit*/
    int (*server_cb_transferexit)(void *srv_ctx,            /**< Server context pointer*/
				  uint8_t *reqdata,         /**< Request buffer*/
				  int reqlen,               /**< Request buffer length*/
				  uint8_t *respdata,        /**< Response buffer*/
				  int *resplen              /**< Response buffer length pointer, initialid on max length*/
				  );		
}cs_uds_server_cb_t;

#ifdef __cplusplus
extern "C" {
#endif

  /** Set debug level
      \retval Active Debuglevel
  */  
  cs_debug_level_t cs_uds_debuglevel(
				     cs_debug_level_t level  /**< Debug level*/
				     );
  
  /** Set debugging callback routine
      Default: Function registered which prints debugging to STDOUT
  */
  void cs_uds_debug_setcallback(
				void *ctx,                             /**< Callback private data*/
				void (*callb)(void *ctx,
					      cs_debug_sub_t sub,
					      cs_debug_level_t dl,
					      const char *func,
					      const int line,
					      const char *fmt,
					      va_list ap) /**< Callback routine*/
				);
  
  /** Get verion magic of UDS stack
      \retval uint16 high byte is major version , low byte is minor version
  */
  uint16_t cs_uds_getversion();

  /** Reset error state
   */
  void cs_uds_reseterror(
			 cs_uds_service_t *data      /**< UDS data paket structure*/
			 );

  /** Get active error flag
      \retval Errorflag
  */
  uint16_t cs_uds_getlerror(
			    cs_uds_service_t *data   /**< UDS data service structure*/
			    );

  /** Translate error id flag to string
      \retval NULL Failure
      \retval Error string
  */
  char * cs_uds_strlerrorid(
			    uint16_t lerror         /**< UDS Error flag*/
			    );
  
  /** Get active error flag description
      \retval error string
  */
  char *cs_uds_strlerror(
			 cs_uds_service_t *data     /**< UDS data service structure*/
			 );

  /** Reset UDS timing behavior to default values
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_channel_reset_timing(
				  cs_uds_service_t *data           /**< UDS data service structure*/
				  );

  
  /** Set UDS timing behavior
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_channel_settiming(
			       cs_uds_service_t *data,           /**< UDS data service structure*/
			       struct cs_uds_time_server tserver,/**< Timeing data structure*/
			       uint16_t pcandelta                /**< Optional Detla in ms*/
			       );

  /** Get actual timing data
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_channel_gettiming(
			       cs_uds_service_t *data,            /**< UDS data service structure*/
			       struct cs_uds_time_server *tserver /**< Timeing data storage pointer*/
			       );
  
  /** Receive non blocking. This is used to handle messages send by master in case of RESPSUP.
      No failure code handling inside the function.
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_receive_nonblock(
			      cs_uds_service_t *data,             /**< UDS data service structure*/
			      uint32_t timeout                    /**< Timeout in ms*/
			      );

  /** Init and allocate data structure
      \retval 0 Failure
      \retval != 0 Success
  */
  cs_uds_service_t *cs_uds_init();
  
  /** Free data structure
   */

  void cs_uds_free(
		   cs_uds_service_t *data             /**< UDS data service structure*/
		   );

  /** Initiate CAN/IsoTP stack for communication
      Default setting is 
      - CAN 2.0 /  8 Byte
      - IsoTP max frame size is 4095 Bytes 
      - IsoTP RX-STIM 50 ms
      - IsoTP TX-BS 20
      - IsoTP TX-Padding / 0 Byte
      For further settings call cs_uds_get_isotp_handle and call
      cs_isotp_ realatred functions.
      Please note this has to be done before call cs_uds_can_open
      \retval 0 Sucess
      \retval 1 Failure
  */
  int cs_uds_can_init(
		       cs_uds_service_t *data,          /**< UDS data service structure*/
		       char *iface,                     /**< SocketCAN interface name*/
		       uint32_t sa,                     /**< CAN Source adress*/
		       uint32_t ta                      /**< CAN target adress*/
		       );

  /** Get IsoTP-Layer socket handle
      Please note, Options can be set between calls can_init and can_open
      \retval 0 Failure
      \retval != IsoTP Layer handle
  */
  cs_isotp_t * cs_uds_get_isotp_handle(
				       cs_uds_service_t *data          /**< UDS data service structure*/
				       );
  
  /** Open CAN interface
      \retval 0 Sucess
      \retval 1 Failure
  */
  int cs_uds_can_open(
		      cs_uds_service_t *data          /**< UDS data service structure*/
		      );
  

  /** Close CAN interface
      \retval 0 Sucess
      \retval 1 Failure
   */
  int cs_uds_can_close(
		       cs_uds_service_t *data        /**< UDS data service structure*/
		       );
    
  /** Request Server  diagnostic session
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_diagsession(
				cs_uds_service_t *data,       /**< Stack data storage structure*/
				uint8_t subfunc,              /**< Subfunction call, valid values are 
								 predefined values
								 CS_UDS_SI_DSC_DS: Default session
								 CS_UDS_SI_DSC_PRGS: Programming session
								 CS_UDS_SI_DSC_EXTDS: Extended Session
								 CS_UDS_SI_DSC_SSDS: Safety Session
								 CS_UDS_SI_DSC_EOL: EOL Extended Session
								 CS_UDS_SI_DSC_SUP: Supplier Session
							      */
				resp_t resp,                  /**< Is response requested*/
				cs_uds_diagsession_t *rvalue  /**< Return value storage structure pointer*/
				);

  /** Send a tester present
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_testerpresent(
				  cs_uds_service_t *data,   /**< Stack data storage structure*/
				  resp_t resp               /**< Is response requested*/
				  );

  /** Blocking sleep with sending of tester present
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_sleep(
			  cs_uds_service_t *data,           /**< Stack data storage structure*/
			  uint32_t time                     /**< Time in s*/
			  );
  
  /** Send a ECU reset command
      \retval 0 Success
      \retval 1 Failure
  */     
  int cs_uds_client_ecureset(
			     cs_uds_service_t *data,        /**< Stack data storage structure*/
			     uint8_t subfunc,               /**< Subfunction call, valid values are 
								 predefined values
								 CS_UDS_SI_ER_HR: Hard reset
								 CS_UDS_SI_ER_KOFFONR: keyOffOnReset
								 CS_UDS_SI_ER_SR: Soft Reset
								 CS_UDS_SI_ER_ERPSD: Enable Rapid Shutdown
								 CS_UDS_SI_ER_DRPSD: Disable Rapid Shutdown
							    */
			     resp_t resp,                   /**< Is response requested*/
			     uint8_t *powerdown             /**< Feedback powerdown time for Subfunction RapidShutdown,
							       CS_UDS_SI_ER_ERPSD, in other case this value maybe NULL*/
			     );

  /** Sercurity access call 
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_securityaccess(
				   cs_uds_service_t *data,  /**< Stack data storage structure*/
				   uint8_t subfunc,         /**< Subfunction call, valid values are 
								 predefined values 
								 CS_UDS_SI_SA_RSD: Request Seed
								 CS_UDS_SI_SA_SK: Send Key
								 Note to define 
								 the acces level higher than basic level 1 and 2 
								 increment base value by 2
								 Example: Request seed level 3: CS_UDS_SI_SA_RSD + 2 = 3
							         Example: Send Key level 4 : CS_UDS_SI_SA_SK + 2 = 4*/
				   uint8_t *sdbuffer,       /**< Send data buffer, user for send key 
							       CS_UDS_SI_SA_SK, in other case value 
							       can be NULL*/
				   int sdlen,               /**< Send buffer len*/
				   uint8_t *rcvbuffer,      /**< Recv data buffer, has to be preallocated,
							       used for request seed CS_UDS_SI_SA_RSD, 
							       in other case value can be NULL*/
				   int *rcvlen              /**< Recv buffer max len, 
							       will be overwritten by recv data len */
				   );

  /** Control DTC handling
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_controldtcsettings(
				       cs_uds_service_t *data, /**< Stack data storage structure*/ 
				       uint8_t subfunc,        /**< Subfunction call, valid values are 
								  predefined values 
								  CS_UDS_SI_CDTCS_ON: Switch on
								  CS_UDS_SI_CDTCS_OFF: Switch off
							       */
				       resp_t resp             /**< Is response requested*/
				       );

  /** Link control function
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_linkcontrol(
				cs_uds_service_t *data,  /**< Stack data storage structure*/ 
				uint8_t subfunc,         /**< Subfunction call, valid values are 
							    predefined values 
							    CS_UDS_SI_LC_VMTWFP: verifyModeTransitionWithFixedParameter
							    CS_UDS_SI_LC_VMTWSP: verifyModeTransitionWithSpecificParameter
							    CS_UDS_SI_LC_TM: transitionMode
							 */
				resp_t resp,             /**< Is response requested*/
				uint32_t sidata          /**< Mode definition data, following values are predefined
							    CS_UDS_SI_LC_PC9600: PC9600Baud
							    CS_UDS_SI_LC_PC19200: PC19200Baud
							    CS_UDS_SI_LC_PC38400: PC38400Baud
							    CS_UDS_SI_LC_PC57600: PC57600Baud
							    CS_UDS_SI_LC_PC38400: PC38400Baud
							    CS_UDS_SI_LC_CAN125000: CAN125000Baud
							    CS_UDS_SI_LC_CAN250000: CAN250000Baud
							    CS_UDS_SI_LC_CAN500000: CAN500000Baud
							    CS_UDS_SI_LC_CAN1000000: CAN1000000Baud
							    CS_UDS_SI_LC_PROGSU: ProgrammingSetup
							 */
				);

  /** Read Data by Identifier
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_readDatabyIdentifier(
					 cs_uds_service_t *data,   /**< Stack data storage structure*/
					 uint16_t did,             /**< DID identifier*/
					 int *len,                 /**< Recv buffer len
								      -1 for no checking of recv len, 
								      for this rvdata should be minimal
								      CS_UDS_MAXBYTE bytes long
								      >= 0 the function checks the len and fails
								      if recv len does not equal*/
					 void *rvdata              /**< Recv buffer*/
					 );

  /** Write data by Identifier
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_writeDatabyIdentifier(
					  cs_uds_service_t *data,   /**< Stack data storage structure*/
					  uint16_t did,             /**< DID identifier*/
					  uint16_t len,             /**< Send buffer len*/
					  void *senddata            /**< Send buffer*/
					  );

  /** Routine control
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_routinecontrol(
				   cs_uds_service_t *data,         /**< Stack data storage structure*/
				   uint8_t subfunc,                /**< Subfunction call, valid values are 
								      predefined values 
								      CS_UDS_SI_RC_STR: Start
								      CS_UDS_SI_RC_STPR: Stop
								      CS_UDS_SI_RC_RRR: Request results
								   */
				   resp_t resp,                    /**< Is response requested*/
				   uint16_t routine,               /**< Routine Id*/
				   uint8_t *param,                 /**< Data array as optional routine
								      parameter, maybe NULL*/
				   int rqlen,                      /**< Length of data array*/
				   uint8_t *routineinfo,           /**< Optional RoutineInfo recv parameter,
								      maybe NULL, needed by special routine calls*/
				   uint8_t *rvalue,                /**< Recv data array, may be NULL*/
				   int *resplen                    /**< Recv buffer length*/
				   );

  /** Clear diagnostic information
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_ClearDiagnosticInformation(
					       cs_uds_service_t *data,  /**< Stack data storage structure*/
					       resp_t resp,             /**< Is response requested*/
					       uint32_t dtcmask         /**< DTC mask*/
					       );
  
  /** Read DTC information, TODO the interface of this function will change in future
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_ReadDTCInformation(
				       cs_uds_service_t *data,         /**< Stack data storage structure*/
				       resp_t resp,                    /**< Is response requested*/
				       uint8_t subfunc,                /**< Subfunction call, valid values are 
									  predefined values 
									  CS_UDS_SI_RDTCI_RNODTCBSM: reportNumberOfDTCByStatusMask
									  CS_UDS_SI_RDTCI_RDTCBSM: reportDTCByStatusMask
									  CS_UDS_SI_RDTCI_RDTCSSI: reportDTCSnapshotIdentification
									  CS_UDS_SI_RDTCI_RDTCSSBDTC: reportDTCSnapshotRecordByDTCNumber
									  CS_UDS_SI_RDTCI_RDTCSDBRN: reportDTCStoredDataByRecordNumber
									  CS_UDS_SI_RDTCI_RDTCEDRBDN: reportDTCExtDataRecordByDTCNumber
									  CS_UDS_SI_RDTCI_RNODTCBSMR: reportNumberOfDTCBySeverityMaskRecord
									  CS_UDS_SI_RDTCI_RDTCBSMR: reportDTCBySeverityMaskRecord
									  CS_UDS_SI_RDTCI_RSIODTC: reportSeverityInformationOfDTC
									  CS_UDS_SI_RDTCI_RSUPDTC: reportSupportedDTC
									  CS_UDS_SI_RDTCI_RFTFDTC: reportFirstTestFailedDTC
									  CS_UDS_SI_RDTCI_RFCDTC: reportFirstConfirmedDTC
									  CS_UDS_SI_RDTCI_RMRTFDTC: reportMostRecentTestFailedDTC
									  CS_UDS_SI_RDTCI_RMRCDTC: reportMostRecentConfirmedDTC
									  CS_UDS_SI_RDTCI_RMMDTCBSM: reportMirrorMemoryDTCByStatusMask
									  CS_UDS_SI_RDTCI_RMMDEDRBDN: reportMirrorMemoryDTCExtDataRecordByDTCNumber
									  CS_UDS_SI_RDTCI_RNOMMDTCBSM: reportNumberOfMirrorMemoryDTCByStatusMask
									  CS_UDS_SI_RDTCI_RNOOEOBDDTCBSM: reportNumberOfEmissionsOBDDTCByStatusMask
									  CS_UDS_SI_RDTCI_ROBDDTCBSM: reportEmissionsOBDDTCByStatusMask
									  CS_UDS_SI_RDTCI_RDTCFDC: reportDTCFaultDetectionCounter
									  CS_UDS_SI_RDTCI_RDTCWPS: reportDTCWithPermanentStatus
									  CS_UDS_SI_RDTCI_RDTCEDBR: reportDTCExtDataRecordByRecordNumber
									  CS_UDS_SI_RDTCI_RUDMDTCBSM: reportUserDefMemoryDTCByStatusMask
									  CS_UDS_SI_RDTCI_RUDMDTCSSBDTC: reportUserDefMemoryDTCSnapshotRecordByDTCNumber
									  CS_UDS_SI_RDTCI_RUDMDTCEDRBDN: reportUserDefMemoryDTCExtDataRecordByDTCNumber
									  CS_UDS_SI_RDTCI_RWWHOBDDTCBMR: reportWWHOBDDTCByMaskRecord
									  CS_UDS_SI_RDTCI_RWWHOBDDTCWPS: reportWWHOBDDTCWithPermanentStatus
								       */
				       cs_uds_dtc_reqdata_t *reqdata,  /**< Request data structure*/
				       cs_uds_dtc_respdata_t *respdata /**< Response data structure*/
				       );

  /** Request data download
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_requestDownload(
				    cs_uds_service_t *data, /**< Stack data storage structure*/
				    uint8_t compression,    /**< Compression flag, 0=> without compression*/
				    uint8_t encryption,     /**< Encryption flag, 0=> without encryption*/
				    uint64_t address,       /**< Memory address*/
				    uint64_t len,           /**< Data segment length*/
				    uint64_t *maxblocklen   /**< Feedback maximum block length*/
				    );

  /** Request data upload
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_requestUpload(
				  cs_uds_service_t *data, /**< Stack data storage structure*/
				  uint8_t compression,    /**< Compression flag, 0=> without compression*/
				  uint8_t encryption,     /**< Encryption flag, 0=> without encryption*/
				  uint64_t address,       /**< Memory address*/
				  uint64_t len,           /**< Data segment length*/
				  uint64_t *maxblocklen   /**< Feedback maximum block length*/
				  );

  /** Transfer data from/to server
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_transferData(
				 cs_uds_service_t *data,   /**< Stack data storage structure*/
				 uint8_t SI,               /**< Set to CS_UDS_SI_RU or CS_UDS_SI_RD to 
							      show data flow direction*/ 
				 uint8_t seqcounter,       /**< Sequence counter*/
				 uint64_t maxblocklen,     /**< Max block length*/
				 uint8_t *transferdata,    /**< Transfer data buffer*/
				 uint64_t *transferlen     /**< Transfer data buffer length, 
							      if data upload is requested, set value to max receive
							      buffer size;*/
				 );

  /** Stop running Transfer
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_transferExit(
				 cs_uds_service_t *data,    /**< Stack data storage structure*/
				 uint8_t *senddata,         /**< Optional Send data*/
				 int sendlen,               /**< Send data length*/
				 uint8_t *recvdata,         /**< Optional: Recv data*/
				 int *recvlen               /**< Recv buffer length*/
				 );

  /** Transfer data from/ to client
      Combination out of 
      requestDownload/requestUpload
      transferData
      transferExit
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_api_transfer(
				 cs_uds_service_t *data, /**< Stack data storage structure*/
				 uint8_t SI,             /**< Set to CS_UDS_SI_RU or CS_UDS_SI_RD to 
							      show data flow direction*/ 
				 uint8_t compression,    /**< Compression flag, 0=> without compression*/
				 uint8_t encryption,     /**< Encryption flag, 0=> without encryption*/
				 uint64_t address,       /**< Memory address*/
				 uint8_t *transferdata,  /**< Data buffer*/
				 uint64_t transferlen,   /**< Transfer len*/
				 uint8_t *senddata,      /**< TransferExit send data*/
				 int sendlen,            /**< Transfer send data length*/
				 uint8_t *recvdata,      /**< TransferExit recv data*/
				 int *recvlen            /**< TransferExit recv data length*/
				 );

  /** Init flash structure fore non blocking flashing
      \retval 0 Success
      \retval 1 Failure
  */
  int cs_uds_client_api_flash_init(
				   cs_uds_client_flash_t *flash,   /**< Flash permanent data structure*/
				   cs_uds_service_t *data,         /**< UDS Stachk pointer*/
				   uint8_t SI,                     /**< Set to CS_UDS_SI_RU or CS_UDS_SI_RD to 
								      show data flow direction*/ 
				   uint8_t compression,            /**< Compression flag, 0=> without compression*/
				   uint8_t encryption,             /**< Encryption flag, 0=> without encryption*/
				   uint64_t address,               /**< Memory address*/
				   uint8_t *transferdata,          /**< Data buffer*/
				   uint64_t transferlen            /**< Transfer len*/
				   );

  /** Transfer data, non blocking flashing, internal handling of sequence counter
      \retval 0 Success or busy
      \retval 1 Failure
      \retval 2 Finished
  */
  int cs_uds_client_api_flash_run(
				  cs_uds_client_flash_t *flash    /**< Flash permanent data structure*/
				  );


  /** Set server context and callback table
   */
  void cs_uds_server_set_context(
				 cs_uds_service_t *data,              /**< Stack data storage structure*/
				 void *srv_ctx,                       /**< Server callpack private data*/
				 cs_uds_server_cb_t *cb               /**< Callback table pointer*/
				 );
  
  /** Server process call
      Receive, precess and call callbacks
  */
  void cs_uds_server_process(
			     cs_uds_service_t *data               /**< Stack data storage structure*/
			     );
  
  
#ifdef __cplusplus
}
#endif

#endif
