/*  
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_vbf.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing the parsing API for Ford VBF-files.
 *
 * File containing the parsing API for Ford VBF-files.
*/
#ifndef CS_VBF_H
#define CS_VBF_H

#include <inttypes.h>
#include <stdio.h>

/** \def VBF_MAX_VERSION
    Max supported version
*/
#define VBF_MAX_VERSION 2.40

/** \def VBF_MAX_BLOCKS
    Maximum number of blocks
*/
#define VBF_MAX_BLOCKS 255

/** \def VBF_ADDRESS_SIZE
    Address definition size in bytes
*/
#define VBF_ADDRESS_SIZE 4

/** \def VBF_LENGTH_SIZE
    Data length definition size in bytes
*/
#define VBF_LENGTH_SIZE 4

/** \def VBF_CHECKSUM_SIZE
    Checksum definition size in data segment in bytes
*/
#define VBF_CHECKSUM_SIZE 2


/** \def VBF_ERROR_NOERROR
    No error
*/
#define VBF_ERROR_NOERROR 0x00

/** \def VBF_ERROR_VERSION
    Incompatible version
*/
#define VBF_ERROR_VERSION 0x01

/** \def VBF_ERROR_FORMAT
    Invalid format
*/
#define VBF_ERROR_FORMAT 0x02

/** \def VBF_ERROR_FILE
    Invalid file
*/
#define VBF_ERROR_FILE 0x03

/** \def VBF_ERROR_UNSUPPORTED
    Unsupported feature
*/
#define VBF_ERROR_UNSUPPORTED 0x04



/** \enum cs_vbf_part_t
    Binary data type
*/
typedef enum{
  VBF_UNKNOWN = 0, /**< Undefined type*/
  VBF_CARCFG,      /**< Car configuration*/
  VBF_CUSTOM,      /**< Custom section*/
  VBF_DATA,        /**< Data segment*/
  VBF_EXE,         /**< Execution part data*/
  VBF_SBL,         /**< Software boot loader*/
  VBF_SIGCFG,      /**< SIGCFG, TODO what is this*/
  VBF_TEST         /**< Testing data*/
}cs_vbf_part_t;

/** \enum cs_vbf_network_t
    Network data type
*/
typedef enum{
  VBF_CAN_UNKNOWN = 0, /**< Transport network is unknown*/
  VBF_CAN_HS,          /**< High speed can*/
  VBF_CAN_MS,          /**< Medium speed can*/
}cs_vbf_network_t;

/** \enum cs_vbf_frame_t
    Transport frame types to use.
*/
typedef enum{
  VBF_FRAME_UNKNOWN = 0,  /**< Unknown frame format*/
  VBF_FRAME_STANDARD,     /**< Standard CAN frame type*/
}cs_vbf_frame_t;

typedef struct{
  uint32_t address;       /**< Start address*/
  uint32_t length;        /**< Data length*/
  uint8_t *data;          /**< Data pointer, to mapping area*/
  uint16_t checksum;      /**< CRC16-CCIT checksum from file*/
} cs_vbf_data_t;

/** \struct cs_vbf_t
    Storage structure for parsing VBF files.
*/
typedef struct{
  FILE *fs;               /**< FILE stream*/
  char *filename;         /**< Filename to parse*/
  long filelen;           /**< File length*/
  void *map;              /**< File map pointer*/
  float vbf_version;      /**< VBF file version*/
  char *vbf_description; /**< Description of VBF file*/
  char *vbf_swpartnumber; /**< Software part number*/
  cs_vbf_part_t vbf_parttype; /**< Binary data type*/
  cs_vbf_network_t vbf_networtype; /**< Network type to use for flashing*/
  uint32_t vbf_ecuaddr;   /**< ECU address to connect to ECU*/
  uint32_t vbf_calladdress; /**< Call Address for Bootloader*/
  cs_vbf_frame_t vbf_frame; /**< Transport frame type to use*/
  int vbf_no_eraseblocks;   /**< Number of defined erase blocks*/
  cs_vbf_data_t vbf_eraseblocks[VBF_MAX_BLOCKS]; /**< Erase block array*/
  uint8_t vbf_format_id;    /**< Format identifier, ?Maybe Compressen enrcyption, see UDS requestDownload?*/
  uint32_t vbf_filechecksum; /**< CRC32 checksum from file, used to validate content*/
  int vbf_no_blocks;   /**< Number of defined blocks*/
  cs_vbf_data_t vbf_blocks[VBF_MAX_BLOCKS]; /**< Block array*/
  uint8_t *binarystart;    /**< Start of binary section*/
  uint8_t *binaryend;      /**< End of binary section*/
  int binarylength;        /**< Binary section length*/
  uint8_t lerror;          /**< Last error code*/
}cs_vbf_t;

/** \enum cs_vbf_debug_level_t
    VBF debug level
*/
typedef enum{
  VBF_DEBUG_LEVEL_READ = 0x00,      /**< Only read debug level*/
  VBF_DEBUG_LEVEL_NONE = 0x01,      /**< No debug output*/
  VBF_DEBUG_LEVEL_DEBUG = 0x02,     /**< Debugging output*/
  VBF_DEBUG_LEVEL_VERBOSE = 0x04    /**< Verbose output*/
}cs_vbf_debug_level_t;

#ifdef __cplusplus
extern "C" {
#endif


 /**************************************************************************/
  int cs_vbf_open(
		  cs_vbf_t *vbf,   /**< VBF storage structure*/
		  char *filename   /**< Absolute path to VBF file*/
		  );
 /**************************************************************************/

  /** Close VBF file
      \retval 0 success
  */
 /**************************************************************************/
  int cs_vbf_close(
		  cs_vbf_t *vbf   /**< VBF storage structure*/
		  );
 /**************************************************************************/

  /** Get number of defined erase block 
      \retval 0 success
  */
 /**************************************************************************/
  int cs_vbf_getnumber_eraseblock(
				  cs_vbf_t *vbf,   /**< VBF storage structure*/
				  int * number     /**< Number of erase blocks*/
				  );
 /**************************************************************************/

  /** Get defined erase block, identified by index 
      \retval 0 success
  */
 /**************************************************************************/
  int cs_vbf_get_eraseblock(
			    cs_vbf_t *vbf,       /**< VBF storage structure*/
			    int index,           /**< Block index*/
			    cs_vbf_data_t *erase /**< VBF data structure*/
			    );
 /**************************************************************************/

  /** Get number of defined data block 
      \retval 0 success
  */
 /**************************************************************************/
  int cs_vbf_getnumber_datablock(
				 cs_vbf_t *vbf,   /**< VBF storage structure*/
				 int * number     /**< Number of data blocks*/
				 );
 /**************************************************************************/
  
  /** Get defined data block, identified by index 
      \retval 0 success
  */
 /**************************************************************************/
  int cs_vbf_get_datablock_byindex(
				   cs_vbf_t *vbf,       /**< VBF storage structure*/
				   int index,           /**< Block index*/
				   cs_vbf_data_t *data  /**< VBF data structure*/
				   );
 /**************************************************************************/

  /** Printf VBF file data to stdout
   */
  /**************************************************************************/
  void cs_vbf_print(
		    cs_vbf_t *vbf  /**< VBF storage structure*/
		    );
  /**************************************************************************/

  
  /** Get last error code
      \retval last error code
  */
  /**************************************************************************/
  int cs_vbf_getlasterror(
			  cs_vbf_t *data  /**< Storage structure*/
			  );
  /**************************************************************************/


  /** Get error id string
      \retval error string
  */
  /**************************************************************************/
  char* cs_vbf_geterroridstring(
				  uint8_t lerror /**< Error id*/
				  );
  /**************************************************************************/
  
  /** Get last error string
      \retval error string
  */
  /**************************************************************************/
  char* cs_vbf_getlasterrorstring(
				  cs_vbf_t *data /**< Storage structure*/
				  );
  /**************************************************************************/

  /** Set debug level
      \retval Active Debuglevel
  */
  /**************************************************************************/
  cs_vbf_debug_level_t cs_vbf_debuglevel(
					 cs_vbf_debug_level_t level  /**< Debug level*/
					 );
  /**************************************************************************/
  
#ifdef __cplusplus
}
#endif


#endif
