/*  XCP CAN Stack for SocketCAN
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_xcp.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing the XCP low level API.
 *
 * This file contain the XCP low level API. It is based on ASAM XCP specification.
 * The implementation is not complete and based on the Spec 1.0 version.
 * Actual focus is running DAQ signal handling.
 */

#ifndef CS_XCP_H
#define CS_XCP_H

#include <inttypes.h>

#include "cs_a2l.h"
#include "cs_debug.h"
#include "cs_xcp_a2l.h"
#include "cs_xcp_codes.h"

/** \def CS_XCP_DATALENGTH
    Internal maximal data segment length for paket data segment or odt entry
*/
#define CS_XCP_DATALENGTH 1024 
/** \def CS_XCP_FRAMELENGTH
    Internal maximal supported framelength, max data segment +64 Byte
*/
#define CS_XCP_FRAMELENGTH (CS_XCP_DATALENGTH+64) 

/** \def CS_XCP_IDENTlENGTH
    Maximal ident field length
*/
#define CS_XCP_IDENTLENGTH 4 

/** \def CS_XCP_ERROR_HANDLINGLENGTH 
    Array handling size
*/
#define CS_XCP_ERROR_HANDLINGLENGTH 100

/** \def CS_XCP_REPEATMAX
    Maximal cmd retries 
*/
#define CS_XCP_REPEATMAX 10
/** \def CS_XCP_CHARBUF
    Internal char buffer length
*/
#define CS_XCP_CHARBUF 255 

/** \enum cs_xcp_ifacetype_t
    Interface type
*/
typedef enum{
    XCP_IFUNDEF=0,/**< Interface undefined*/
    XCP_CAN,      /**< CAN physical layer*/
    XCP_USB,      /**< USB phsical layer*/
    XCP_ETHERNET  /**< Ethernet physical layer*/
}cs_xcp_ifacetype_t;

/** \enum cs_xcp_pkg_type_t
    Package type
*/
typedef enum{
    XCP_PKG_NONE = 0,       /**< Unknown package*/
    XCP_PKG_CMD = 0x01,     /**< Cmd package*/
    XCP_PKG_RESP = 0xFF,    /**< Cmd Response package*/
    XCP_PKG_EV = 0xFD,      /**< Event package*/
    XCP_PKG_ERR = 0xFE,     /**< Error package*/
    XCP_PKG_SERV = 0xFC,    /**< Service request package*/
    XCP_PKG_DAQ = 0xFB,     /**< DAQ package*/
    XCP_PKG_STIM = 0xFA,    /**< STIM package*/
}cs_xcp_pkg_type_t;

/** \enum cs_xcp_pa_t
    Different defined Pre-Actions
*/
typedef enum{
    XCP_PA_NONE = 0x01,          /**< No preaction*/
    XCP_PA_WAIT_T7 = 0x02,       /**< Wait T7*/
    XCP_PA_SYNCH = 0x04,         /**< Send Synch*/
    XCP_PA_SETMTA = 0x08,        /**< Send MTA*/
}cs_xcp_pa_t;

/** \enum cs_xcp_ac_t
    Different defined actions
*/
typedef enum{
    XCP_AC_NONE = 0x01,          /**< No action*/
    XCP_AC_REPEAT2 = 0x02,       /**< Repeat 2 times*/
    XCP_AC_REPEAT = 0x04,        /**< Repeat unlimited*/
    XCP_AC_OK = 0x08             /**< Ok*/
}cs_xcp_ac_t;

/** \enum cs_xcp_indicator_t
    Event and indicator interface
*/
typedef enum{  
    XCP_INDICATOR_NONE = 0x00,                      /**< No indication or event*/
    XCP_INDICATOR_EVENT_RESUME = 0x01,              /**< Event Resume mode recieved*/
    XCP_INDICATOR_EVENT_CLEAR_DAQ = 0x02,           /**< Event Clear DAQ recieved*/
    XCP_INDICATOR_EVENT_STORE_DAQ = 0x04,           /**< Event Store DAQ recieved*/
    XCP_INDICATOR_EVENT_STORE_CAL = 0x08,           /**< Event Store CAL recieved*/
    XCP_INDICATOR_OVERLOAD = 0x10,                  /**< Overload indication recieved*/
    XCP_INDICATOR_EVENT_SESSION_TERMINATED = 0x20,  /**< Event session terminated received*/
    XCP_INDICATOR_EVENT_USER_DEFINED = 0x40,        /**< Event user defined received*/
    XCP_INDICATOR_EVENT_TRANSPORT = 0x80,           /**< Event transport layer received*/
    XCP_INDICATOR_SERV_RESET = 0x100,               /**< Service request reset received*/
    XCP_INDICATOR_SERV_TEXT = 0x200,                /**< Service Text received*/
    XCP_INDICATOR_INVALID = 0x7FFF,                 /**< Invalid*/
    XCP_INDICATOR_MASK_ALL = 0xFFFF,                /**< Mask for all Indicator*/
}cs_xcp_indicator_t;

/** \struct cs_xcp_error_handling_t
    Error handling structure,
    Specific error handling for each cmd
*/
typedef struct{
    uint8_t cmd;            /**< Command*/
    uint8_t subcmd;         /**< Subcommand*/
    uint16_t error;         /**< Error code*/
    cs_xcp_pa_t preaction;  /**< Preaction code*/
    cs_xcp_ac_t action;     /**< Action code*/
}cs_xcp_error_handling_t;

/** \struct cs_xcp_status_t
    Feedback storage structure for GET_STATUS cmd
*/
typedef struct{
    uint8_t store_cal_req;   /**< 1 => REQuest to STORE CALibration data is set*/
    uint8_t store_daq_req;   /**< 1 => REQuest to STORE DAQ list is set*/
    uint8_t clear_daq_req;   /**< 1 => REQuest to CLEAR DAQ configuration is set*/
    uint8_t daq_running;     /**< 1 => Data transfer is running*/
    uint8_t resume;          /**< 1 => Slave is in RESUME mode*/
    uint8_t prot_calpag;     /**< 1 => CALibration/PAGing commands are protected with SEED & Key mechanism*/
    uint8_t prot_daq;        /**< 1 => DAQ list commands are protected with SEED & Key mechanism*/
    uint8_t prot_stim;       /**< 1 => DAQ (Direction STIM) list commands are protected with SEED & Key mechanism*/
    uint8_t prot_pgm;        /**< 1 => ProGraMming commands are protected with SEED & Key mechanism*/
    uint8_t reserved;        /**< Reserved placeholder*/
    uint16_t session_id;     /**< Session id*/
}cs_xcp_status_t;

/** \struct cs_xcp_pkg_t
    Package transfer structure
*/
typedef struct{
    cs_xcp_pkg_type_t pkgtype;              /**< Package type*/
    uint8_t PID;                            /**< Package identification id*/
    uint8_t data[CS_XCP_DATALENGTH];        /**< Package data segment*/
    int dl;                                 /**< Package data segment length*/
    int exp_dl;                             /**< Package data segment expected length*/
    uint8_t serv_req;                       /**< Service request code for SERV packages*/
    struct timeval tv;                      /**< timestamp*/
}cs_xcp_pkg_t;

/** \struct cs_xcp_odt_entry_t
    ODT entry structure
*/
typedef struct{
    uint8_t data[CS_XCP_DATALENGTH];/**< Raw data storage*/
    cs_a2l_measurement_t me; /**< Signal dependent information*/
}cs_xcp_odt_entry_t;

/** \struct cs_xcp_odt_t
    ODT structure
*/
typedef struct{
    int no_entry;              /**< Number of ODT entries in list*/
    cs_xcp_odt_entry_t *entry; /**< Storage pointer*/
    int datasize;              /**< Sum of data bytes*/
    double last;               /**< Timestamp of last received ODT*/
}cs_xcp_odt_t;

/** \struct cs_xcp_daq_t
    DAQ structure
*/
typedef struct{
    int no_odt;                /**< Number of ODTs in DAQ*/
    cs_xcp_odt_t *odt;         /**< Storage pointer*/
    cs_xcp_daq_dir_t dir;      /**< DAQ direction*/
    int prescaler;             /**< Prescaler*/
    int ev;                    /**< Evcent channel number*/
    uint8_t priority;          /**< Priority, 0xff => highest*/
    int timestamp;             /**< 1 => Enable Timestamp*/
    int pid_off;               /**< 0 Transmit with PID field*/
    int resume;                /**< 1 => DAQ is prepared for resume mode*/
    int running;               /**< 1 => DAQ is running*/
    int selected;              /**< 1 => DAQ is selected*/
    uint16_t first_pid;        /**< First PID of DTO*/
    uint32_t canid;            /**< DAQ CAN id*/
    uint16_t pid;              /**< actual PID of DTO*/
    double ts_ref;             /**< Timestamp of first received ODT*/ 
    uint32_t ts_old;           /**< Storage for old received timestamp*/
    uint64_t counter;          /**< Recv or send counter*/
    double period;             /**< Sampling period time in s*/
    double ts_expect;          /**< Expected timestamp for next ODT package*/
}cs_xcp_daq_t;

/** \struct cs_xcp_daq_list_t
    DAQ list structure
*/
typedef struct{
    int no_daq;               /**< Number of DAQ lists*/
    cs_xcp_daq_t *daq;        /**< Storage pointer*/
    double clock_shift;       /**< Clock shift between master and slave*/
    double clock_base;        /**< Clock base timestamp of slave*/
}cs_xcp_daq_list_t;


/** \struct cs_xcp_resources_t
    General feature settings
*/
typedef struct{
    uint8_t CALPAG;           /**< 1 => Calibration/Pagingavailable*/
    uint8_t DAQ;              /**< 1 => DAQ lists available*/
    uint8_t STIM;             /**< 1 => stimulation available*/
    uint8_t PGM;              /**< 1 => Flash programming available*/
}cs_xcp_resources_t;

typedef struct{
  cs_xcp_a2l_t xcpa2l;        /**< Interface description*/ 
  cs_xcp_ifacetype_t iftype;  /**< Physical layer type*/
  cs_xcp_can_t *ifcan;         /**< Storage structure for CAN specific stuff*/
  int ifmax;                  /**< Max data length of physical layer*/
  int lerror;                 /**< Latest stored error code*/
  cs_xcp_resources_t resources; /**< Feature storage structure*/
  struct{
    uint32_t address;        /**< Memory access address*/
    uint8_t extension;       /**< Memory extension*/
  }mta;
  cs_xcp_error_handling_t error_handling[CS_XCP_ERROR_HANDLINGLENGTH]; /**< Static definition of error handling matrix*/
  cs_xcp_daq_list_t daqlist;   /**< DAQ list storage pointer*/
  cs_xcp_odt_entry_t *work_entry; /**< Working entry pointer for confogiration*/
  cs_xcp_indicator_t indicator; /**< Asynchron indicator*/
  uint8_t isConnected;    /**< Storage if API is connected, 0 => not connected, 1 => connected*/
  uint32_t stim;          /**< Configure message seperation time in ms*/
  double daq_fac_recvtimeout; /**< factor for recv timeout check, default 2.0*/
}cs_xcp_t;

#ifdef __cplusplus
extern "C" {
#endif

  /** Set debug level
      \retval Active Debuglevel
  */  
  cs_debug_level_t cs_xcp_debuglevel(
				     cs_debug_level_t level  /**< Debug level*/
				     );

  /** Set debugging callback routine
      Default: Function registered which prints debugging to STDOUT
  */
  void cs_xcp_debug_setcallback(
				void *ctx,                             /**< Callback private data*/
				void (*callb)(void *ctx,
					      cs_debug_sub_t sub,
					      cs_debug_level_t dl,
					      const char *func,
					      const int line,
					      const char *fmt,
					      va_list ap) /**< Callback routine*/
				);
  
  /** Get verion magic of XCP stack
      \retval uint16 high byte is major version , low byte is minor version
  */
  /********************************************************************/
  uint16_t cs_xcp_getversion();
  /********************************************************************/

    /** Convert timeout (ms) into a timespec structure
        \retval Result timespec structure
    */
    struct timespec cs_xcp_ts_timeout2ts(
        uint32_t timeout     /**< Time in 0.001s*/
        );

    /** Get last error code
        \retval 0 No error
        \retval >0 specific error code
    */
    int cs_xcp_get_lerror(
        cs_xcp_t *data         /**< XCP stack structure*/ 
        );

    /** Get last error code string
        \retval Char pointer to error description
    */
    char *cs_xcp_get_lerrorstr(
        cs_xcp_t *data         /**< XCP stack structure*/ 
        );

    /** Get actual indicator
        \retval actual active indication flags
    */
    cs_xcp_indicator_t cs_xcp_get_indicator(
        cs_xcp_t *data         /**< XCP stack structure*/ 
        );
    /** Reset indicator
     * Use XCP_INDICATOR_MASK_ALL for all
     */
    void cs_xcp_reset_indicator(
        cs_xcp_t *data,                 /**< XCP stack structure*/ 
        cs_xcp_indicator_t indicator    /**< Indicator*/
        );

    /** Set protocol layer version
        \retval 0 No failure
        \retval 1 Failure
    */  
    int cs_xcp_data_set_protversion(
        cs_xcp_t *data,   /**< XCP stack structure*/ 
        int version       /**< Protocol version*/
        );

    /**Set timeout data
       \retval 0 No failure
       \retval 1 Failure
    */
    int cs_xcp_data_set_timeout(
        cs_xcp_t *data,        /**< XCP stack structure*/ 
        int *timeout,         /**< Timeout value array*/
        int len               /**< Array len, has to be 7*/
        );


    /** Get resources information
        \retval Resources information
    */
    cs_xcp_resources_t cs_xcp_data_get_resources(
        cs_xcp_t *data /**< XCP stack structure*/ 
        );

    /** Get DAQ list information
        \retval DAQ list information pointer , for failure value is NULL
    */
    cs_xcp_daq_t *cs_xcp_data_get_daq(
        cs_xcp_t *data,  /**< XCP stack structure*/ 
        int daq          /**< DAQ list number*/
        );

    /** Get DAQ receive counter
        \retval DAQ list receive counter , for failure always zero
    */
    uint64_t cs_xcp_data_get_daq_counter(
        cs_xcp_t *data,  /**< XCP stack structure*/ 
        int daq          /**< DAQ list number*/
        );

    /** Preallocate local (not on slave) DAQ structure
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_data_set_no_daq(
        cs_xcp_t *data,         /**< XCP stack structure*/ 
        int no_daq              /**< Number of DAQ list to allocate (>=1)*/          
        );

    /** Preallocate local (not on slave) ODT in a DAQ list structure
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_data_set_no_odt(
        cs_xcp_t *data,          /**< XCP stack structure*/
        int daq,                 /**< DAQ list number*/
        int no_odt               /**< Number of ODT to allocate (>=1)*/   
        );

    /** Preallocate local (not on slave) ODT Entry in a ODT structure
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_data_set_no_odt_entry(
        cs_xcp_t *data,   /**< XCP stack structure*/
        int daq,          /**< DAQ list number*/
        int odt,          /**< ODT number*/
        int no_entry      /**< Number of ODT Entries to allocate (>=1)*/   
        );


    /** Get entry raw value (uint8 buffer)
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_entry_value_raw(
        cs_xcp_t *data,        /**< XCP stack structure*/
        int daq,               /**< DAQ list number*/
        int odt,               /**< ODT number*/
        int entry,             /**< ODT Entry number*/
        uint8_t *udata,        /**< Result raw data pointer, may not be NULL*/
        int *size              /**< Result size, may not be NULL*/
        );

    /** Define a compu method description object, here linear description
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_compu_method_linear_create(
        cs_a2l_compu_method_t* cm,   /**<Compu method storage*/
        char *name,                  /**< Name of cm method*/
        char *label,                 /**< Optional label of method*/
        char *unit,                  /**< Optional unit of value*/
        char *format,                /**< Optional output format of value*/
        double scale,                /**< Scale value*/
        double offset                /**< Offstet value*/
        );

    /** Register callback value and function for ODT entry
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_entry_reg_value(			
        cs_xcp_t *data,        /**< XCP stack structure*/
        int daq,               /**< DAQ list number*/
        int odt,               /**< ODT number*/
        int entry,             /**< ODT Entry number*/
        double *value,         /**< Optional: Feedback value pointer, maybe NULL*/
        cs_a2l_compu_method_t cm, /** Compu method description*/
        cs_a2l_data_type_t dtype, /**< Data conversion type*/
        double *ts,             /**< Otional: Feedback timestamp counter, maybe NULL*/
        void (*extcallback)(double value, double ts, void *priv), /**Optional: Receive callback function for new value
                                                                     Arguments: value -> signal value
                                                                     ts    -> signal timestamp
                                                                     priv  -> private data structure,
                                                                     maybe NULL
                                                                  */
        void *priv              /**< Optional: Private data structure storage pointer, 
                                   for use in callback function call,
                                   maybe NULL
                                */
        );

    /** Convert recv raw buffer to double value
        \retval 0 No failure
        \retval 1 Failure
    */

    int cs_xcp_convert_to_double(			
        cs_xcp_t *data,      /**< XCP stack structure*/
        uint8_t *udata,      /**< Raw data pointer*/
        int size,            /**< Raw data size*/
        cs_a2l_compu_method_t cm, /** Compu method description*/
        cs_a2l_data_type_t dtype, /**< Data conversion type*/
        double *value        /**< Return pointer*/
        );

    /** Check for running DAQ entries if there was an receive timeout.
        An timeout happens when the last ODT recv timestamp  plus
        (DAQ period multiplied by fac_timeout) is lesser than actual time
        \retval 0 No timeout
        \retval 1 There was an failure or timeout
    */
    int cs_xcp_daq_istimeout(			
        cs_xcp_t *data,       /**< XCP stack structure*/
        double fac_timeout    /**< Timeout factor, has to be >=1.0, < 1.0 the check is deactivated*/
        );

    /** Print out resource information to STDOUT
     */
    void cs_xcp_print_resources(
        cs_xcp_resources_t *resources /**< Resources inormation pointer*/
        );

    /** Print out status information to STDOUT
     */
    void cs_xcp_print_status(
        cs_xcp_status_t *status /**< Status information pointer*/
        );

    /** Configure physical layer, here CAN usage, this function is used to do the configuration manually
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_iface_can_configure(
        cs_xcp_t *data,         /**< XCP stack structure*/       
        unsigned int cmd_id,    /**< Master id, to send to*/
        unsigned int resp_id,   /**< Slave id, for response recv*/
        unsigned int bcast_id   /**< Broadcat id*/
        );

    /** Setup physical layer, here CAN usage, this function is used to do the configuration via A2L interface
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_iface_can_setup(
        cs_xcp_t *data,         /**< XCP stack structure*/       
        cs_xcp_can_t *iface      /**< Interface data*/
        );

    /** Open the pyhsical layer, here CAN usage
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_iface_can_open(
        cs_xcp_t *data,    /**< XCP stack structure*/      
        char *iface        /**< SocketCAN interface name*/
        );

    /** CAN Layer specific: Send broadcast, expect resonse from all XCP slaves with CAN IDs
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_get_slave_id(
        cs_xcp_t *data,      /**< XCP stack structure*/ 
        int mode,            /**< Echo mode: 0=> by echo, 1=> inverse echo*/
        uint32_t **respid,   /**< Storage array of received CAN IDs, array is allocated in function, may not be NULL*/
        int *len             /**< Number of response, Pointer, may not be NULL*/
        );

    /** Basic init of XCP stack
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_init(
        cs_xcp_t *data         /**< XCP stack structure*/ 
        );

    /** Close physical layer
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_close(
        cs_xcp_t *data         /**< XCP stack structure*/ 
        );

    /** Receive Data DTO, EV, SERV, ...
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_receive(
        cs_xcp_t *data       /**< XCP stack structure*/  
        );

    /** Connect to slave
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_connect(
        cs_xcp_t *data,      /**< XCP stack structure*/ 
        int mode             /**< Connection mode, 0=> normal, 1=> user defined*/
        );

    /** Disconnect from slave
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_disconnect(
        cs_xcp_t *data      /**< XCP stack structure*/
        );

    /** Get slave status information
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_get_status(
        cs_xcp_t *data,             /**< XCP stack structure*/
        cs_xcp_status_t *status     /**< Status feedback storage structure*/
        );

    /** Send synch frame
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_synch(
        cs_xcp_t *data      /**< XCP stack structure*/
        );

    /** Set slave Memory Transfer address
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_set_mta(
        cs_xcp_t *data,        /**< XCP stack structure*/
        uint32_t address,      /**< Memory address*/
        uint8_t extension      /**< Memory address extension*/
        );

    /** Upload data from slave to master
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_upload(
        cs_xcp_t *data,         /**< XCP stack structure*/
        uint32_t *recv,         /**< recv buffer pointer, min *len bytes long*/
        int *len                /**< recv buffer length*/
        );

    /** Get additional communication information
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_get_commode_info(
        cs_xcp_t *data      /**< XCP stack structure*/
        );

    /** Get slave device identification
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_get_id(
        cs_xcp_t *data,      /**< XCP stack structure*/
        int mode,            /**< Request mode, 0= ASCII, 1=ASAM-MC2 filename with path and
                                extension, 2= URL where the ASAM-MC2 file can be
                                found, 3=ASAM-MC2 file to upload, 128-255= user defined*/
        uint32_t *recvbuffer,/**< recv buffer pointer, min *len bytes long*/
        int *recvbufferlen   /**< recv buffer length*/
        );

    /** Clear DAQ Entry
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_clear_daq_list(
        cs_xcp_t *data,       /**< XCP stack structure*/
        int daq               /**<  DAQ Entry ID*/
        );

    /** Set DAQ Read/Write Pointer
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_set_daq_ptr(
        cs_xcp_t *data,       /**< XCP stack structure*/
        int daq,              /**< DAQ Entry ID (zero based)*/
        int odt,              /**< ODT ID (zero based)*/
        int odt_entry         /**< ODT Entry ID (zero based)*/
        );

    /** Write ODT entry infomation, entry has to be preselected by cs_xcp_set_daq_ptr
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_write_daq(
        cs_xcp_t *data,         /**< XCP stack structure*/
        uint8_t bit_offset,     /**< Bit offset,  for non bit values => 0xFF*/
        int size,               /**< Size in AG*/
        uint8_t addr_extension, /**< Address extension*/
        uint32_t address        /**< Address*/
        );

    /** Get DAQ processor information
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_get_daq_processor_info(
        cs_xcp_t *data,         /**< XCP stack structure*/
        cs_xcp_a2l_daq_proc_t *rv   /**< Result data structure, maybe NULL*/
        );

    /** Get DAQ resolution information
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_get_daq_resolution_info(
        cs_xcp_t *data,         /**< XCP stack structure*/
        cs_xcp_a2l_daq_proc_t *rv   /**< Result data structure, maybe NULL*/
        );
    /** Free DAQ lists
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_free_daq(
        cs_xcp_t *data         /**< XCP stack structure*/
        );

    /** Alloc DAQ lists
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_alloc_daq(
        cs_xcp_t *data,         /**< XCP stack structure*/
        int count               /**< Number of DAQ lists to alloc*/
        );

    /** Alloc ODT list space for DAQ entry
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_alloc_odt(
        cs_xcp_t *data,         /**< XCP stack structure*/
        int daq,                /**< DAQ list number*/
        int count               /**< Number of ODTs to alloc*/
        );

    /** Alloc ODT Entry list space for ODT
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_alloc_odt_entry(
        cs_xcp_t *data,   /**< XCP stack structure*/
        int daq,          /**< DAQ list number*/
        int odt,          /**< ODT number*/
        int count         /**< Number of ODT Entries to alloc*/
        );

    /** Configure DAQ list mode
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_set_daq_list_mode(
        cs_xcp_t *data, /**< XCP stack structure*/
        int daq,        /**< DAQ list number*/
        int ev,         /**< Event trigger of DAQ list*/
        cs_xcp_daq_dir_t direction, /**< DAQ direction*/
        int pid_off,    /**< 1 => disable PID transmission*/
        int timestamp,  /**< 1 => enable timestamp*/
        int prescaler,  /**< Prescaler >= 1*/
        uint8_t priority/**< DAQ priority, highest 0xff*/
        );

    /** Start/Stop/Select DAQ list
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_start_stop_daq_list(
        cs_xcp_t *data,   /**< XCP stack structure*/
        int daq,          /**< DAQ list number*/
        int mode          /**< Mode 0=>Stop 1=> Start 2=>Select*/
        );

    /** Start/Stop all or selected DAQ lists
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_start_stop_synch(
        cs_xcp_t *data,   /**< XCP stack structure*/
        int mode          /**< Mode 0=>Stop all 1=> Start selected 2=>Stop seelected*/
        );

    /** Get timestamp of slaveand calculate timejitter
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_get_daq_clock(
        cs_xcp_t *data,   /**< XCP stack structure*/
        double *jitter    /**< Result pointer, Jitter in s, may be NULL*/
        );

    /** Get information for an specific DAQ channel
        \retval 0 No failure
        \retval 1 Failure
    */
    int cs_xcp_cmd_get_daq_event_info(
        cs_xcp_t *data,  /**< XCP stack structure*/
        int event,       /**< Event number*/
        cs_xcp_a2l_event_t *ev /**< Feedback result pointer, may be NULL*/
        );
#ifdef __cplusplus
}
#endif

#endif
