/*  XCP CAN Stack for SocketCAN
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_xcp_a2l.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing the XCP A2L API.
 *
 * This file contain the XCP A2L API.
*/

#ifndef CS_XCP_A2L_H
#define CS_XCP_A2L_H

#include "cs_defines.h"
#include "cs_a2l.h"
#include "cs_xcp_codes.h"
#include "cs_xcp_can.h"

/** \def CS_XCP_A2L_EVENT_MAX
    Maximal number of events
*/
#define CS_XCP_A2L_EVENT_MAX 24

/** \enum cs_xcp_a2l_timeout_t
    Different timeouts
*/
typedef enum{
    XCP_A2L_TO_T1 = 0,            /**< Timeout T1 (connect)*/
    XCP_A2L_TO_T2,                /**< Timeout T2*/
    XCP_A2L_TO_T3,                /**< Timeout T3*/
    XCP_A2L_TO_T4,                /**< Timeout T4*/
    XCP_A2L_TO_T5,                /**< Timeout T5*/
    XCP_A2L_TO_T6,                /**< Timeout T6*/
    XCP_A2L_TO_T7,                /**< Timeout T7*/
    XCP_A2L_TO_MAX                /**< Timeout Array size definition*/
}cs_xcp_a2l_timeout_t;

/** \enum cs_xcp_a2l_daq_ts_size_t
    DAQ timestamp size
*/
typedef enum{
    XCP_A2L_TS_SZ_NONE = 0x00, /**< Timestamp size none, not supported*/
    XCP_A2L_TS_SZ_BYTE = 0x01, /**< Timestamp 1 Byte*/
    XCP_A2L_TS_SZ_WORD = 0x02, /**< Timestamp 2 Byte*/
    XCP_A2L_TS_SZ_DWORD = 0x04, /**< Timestamp 4 Byte*/
}cs_xcp_a2l_ts_size_t;

/** \enum cs_xcp_a2l_odt_overload_t
    Overload inducation type
*/
typedef enum{
    XCP_A2L_ODT_OVERLOAD_NONE = 0x00, /**< No Overload signal supported*/
    XCP_A2L_ODT_OVERLOAD_MSB = 0x01,  /**< Overload indicated by MSB in PID*/
    XCP_A2L_ODT_OVERLOAD_EV = 0x02,   /**< Overload indicated by event package*/
}cs_xcp_a2l_odt_overload_t;

/** \enum cs_xcp_a2l_odt_opt_t
    ODT Optimisation type
*/
typedef enum{
    XCP_A2L_ODT_OPT_DEFAULT = 0x00,  /**< Optimisation type default*/
    XCP_A2L_ODT_OPT_16 = 0x01,       /**< Optimisation type 16*/
    XCP_A2L_ODT_OPT_32 = 0x02,       /**< Optimisation type 32*/
    XCP_A2L_ODT_OPT_64 = 0x03,       /**< Optimisation type 64*/
    XCP_A2L_ODT_OPT_AL = 0x04,       /**< Optimisation type alignment*/
    XCP_A2L_ODT_OPT_MAX = 0x05,      /**< Optimisation type max entry size*/
}cs_xcp_a2l_odt_opt_t;

/** \enum cs_xcp_a2l_odt_addr_extension
    Allowed address extension type
*/
typedef enum{
    XCP_A2L_ODT_ADDR_EXT_FREE = 0x00, /**< Address extension can be different within one and the same ODT*/
    XCP_A2L_ODT_ADDR_EXT_ODTFIX = 0x01, /**< Address extension has to be the same within one ODT*/
    XCP_A2L_ODT_ADDR_EXT_FIX = 0x03,  /**< Address extension has to be the same for all DAQ*/
}cs_xcp_a2l_odt_addr_extension;

/** \enum cs_xcp_a2l_odt_iden_t
    ODT identification field type
*/
typedef enum{
    XCP_A2L_ODT_IDENT_ABS = 0x00,    /**< Absolute ODT number*/
    XCP_A2L_ODT_IDENT_RAB = 0x01,    /**< Relative ODT, absolute DAQ list number (Byte)*/
    XCP_A2L_ODT_IDENT_RAW = 0x02,    /**< Relative ODT, absolute DAQ list number (Word)*/
    XCP_A2L_ODT_IDENT_RAWA = 0x03,   /**< Relative ODT, absolute DAQ list number (Word, aligned)*/
}cs_xcp_a2l_odt_ident_t;

/** \enum cs_xcp_a2l_event_consistency_t
    Event consistency
*/
typedef enum{
    XCP_A2L_EVENT_CONSISTENCY_ODT = 0x00,  /**< Consistency on ODT level (default)*/
    XCP_A2L_EVENT_CONSISTENCY_DAQ = 0x01,  /**< Consistency on DAQ list level*/
    XCP_A2L_EVENT_CONSISTENCY_EVENT = 0x02,  /**< Consistency on Event Channel level*/
}cs_xcp_a2l_event_consistency_t;


/** \struct cs_xcp_a2l_event_t
    Event data store structure
*/
typedef struct{
    int STIM_supported;            /**< 1 => STIM direction is supported*/
    int DAQ_supported;             /**< 1 => DAQ direction is supported*/
    cs_xcp_a2l_event_consistency_t consistency; /**< Event consistency*/
    int max_daq_list;              /**< Maximal number of DAQ lists for this event*/
    int name_length;               /**< Event channel name length*/
    char name[CS_CHARBUF];     /**< Storage for event name*/                    
    char shortname[CS_CHARBUF];/**< Storage for event short name*/                    
    cs_a2l_ts_unit_t ts_unit;      /**< Cycle period unit*/
    uint8_t period_units;          /**< Sample time in ts_units*/
    double period;                 /**< Sample time in s*/
    uint8_t priority;              /**< Event priority*/
}cs_xcp_a2l_event_t;

/** \struct cs_xcp_a2l_daq_proc_t
    DAQ processor information
*/
typedef struct{
    cs_xcp_a2l_ts_size_t ts_size;       /**< Timestamp size*/
    cs_a2l_ts_unit_t ts_unit;       /**< Timestamp unit*/
    uint8_t ts_fixed;               /**< 1 => Timestamp is fixed*/
    uint8_t daq_odt_entry_gr_size;  /**< DAQ ODT entry granularity size*/
    uint8_t daq_odt_entry_max_size; /**< DAQ ODT entry max size in Byte*/
    uint8_t stim_odt_entry_gr_size; /**< STIM ODT entry granularity size*/
    uint8_t stim_odt_entry_max_size;/**< STIM ODT entry max size in Byte*/
    cs_xcp_a2l_odt_overload_t overload_mode; /**< Overload signal mode */
    uint8_t pid_off_supported;      /**< 1 => PID off is supported*/
    uint8_t daq_static;             /**< 1 => Static DAQ list configuration (0 => dynamic)*/
    uint8_t prescaler_supported;    /**< 1 => Prescaler is supported*/
    uint8_t resume_supported;       /**< 1 => Resume is supported*/
    uint8_t bit_stim_supported;     /**< 1 => Bitwise STIM is supported*/
    uint8_t ts_supported;           /**< 1 => Timestamp is supported*/
    cs_xcp_a2l_odt_opt_t odt_opt;       /**< ODT optimation type*/
    cs_xcp_a2l_odt_addr_extension odt_ext; /**< ODT address extension*/
    cs_xcp_a2l_odt_ident_t odt_ident;   /**< ODT identification type*/
    uint16_t daq_max;               /**< Max number of DAQ lists*/
    uint16_t ev_max;                /**< Max number of event channels*/
    uint16_t daq_min;               /**< Total number of predefined DAQ lists*/
    cs_xcp_a2l_event_t event[CS_XCP_A2L_EVENT_MAX];/**< Event list storage array, size defined by ev_max*/
}cs_xcp_a2l_daq_proc_t;

/** \struct cs_xcp_a2l_commode_t
    Communication data storage structure
*/
typedef struct{
    cs_a2l_bo_t BO;          /**< Byteorder 1 => Motorolla, 0 => Intel*/
    cs_a2l_ag_t AG;          /**< Adress granulation type*/
    uint8_t SBM;             /**< 1 => Slave block mode is available*/
    uint8_t optional;         /**< 1 => Optional features are available*/
    uint8_t IBM;             /**< 1 => Interleaved block mode is available*/
    uint8_t MBM;             /**< 1 => Master block mode is available*/
    uint8_t max_bs;          /**< Max Block size in master block mode*/
    uint8_t min_st;          /**< Min seperation time in units 100 micro seconds*/
    uint8_t queque;          /**< Slave queque size in interleaved mode*/
    uint8_t driver_version;  /**< Slave driver version*/
    int max_DTO;             /**< Max DTO pakages length*/
    int max_CTO;             /**< Max CTO pakages length*/
    uint32_t timeout[XCP_A2L_TO_MAX];/**< Different timeouts*/
    int prot_version;        /**< Protocol version*/
    int trans_version;       /**< Transport layer version*/
    cs_a2l_feature_t optfeature[XCP_CMD_CONNECT]; /**< Flags for optional features*/
    char seedkey[CS_CHARBUF];
}cs_xcp_a2l_commode_t;


/** \struct cs_xcp_a2l_t
    Storage structure for A2L-file parsing
*/
typedef struct{
  cs_a2l_t a2l;           /**< A2L file interface*/
  cs_xcp_a2l_commode_t commode;
  cs_xcp_a2l_daq_proc_t  daq_proc; /**< DAQ processor information*/
}cs_xcp_a2l_t;

#ifdef __cplusplus
extern "C" {
#endif

  /** Init working structure*/
  void cs_xcp_a2l_init(
		       cs_xcp_a2l_t *data  /**< Working storage structure*/
		       );
  
  
  /** Open A2L-File and tries to map it to memory
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_xcp_a2l_open(
		      cs_xcp_a2l_t *data,  /**< Working storage structure*/
		      char *filename,       /**< Filename to open*/
		      char *project,       /**< Project to open*/
		      char *module,        /**< Module to open*/
		      char *interface      /**< Interface to open*/
		      );
  
  
  /** Close A2L-File and reset working structure
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_xcp_a2l_close(
		       cs_xcp_a2l_t *data  /**< Working storage structure*/
		       );
  
  /** Read Protocol section and configure xcp commode structure
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_xcp_a2l_GetProtocoll(
			      cs_xcp_a2l_t *data, /**< Working storage structure*/
			      cs_xcp_a2l_commode_t *mode /**< Communication data storage structure*/
			      );
  
  
  /** Read XCP_ON_CAN section and configure xcp interface structure
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_xcp_a2l_GetPhysInterfaceCAN(
				     cs_xcp_a2l_t *data,   /**< Working storage structure*/
				     cs_xcp_can_t *ifcan   /**< Interface data storage structure*/
				     );
  /** Read out DAQ relevant information
      \retval 0 No failure
      \retval 1 failure
  */
  int cs_xcp_a2l_GetDAQ(
			cs_xcp_a2l_t *data,                /**< Working storage structure*/
			cs_xcp_a2l_daq_proc_t *daq             /**< Daq Processor storage structure*/
			);
  
  /** Init DAQ Processor storage structure
   */
  void cs_xcp_a2l_init_daqproc(
			   cs_xcp_a2l_daq_proc_t *proc  /**< DAQ Processor storage structure*/
			   );

  /** Get DAQ processor information
      \retval DAQ processor information
  */
  cs_xcp_a2l_daq_proc_t cs_xcp_a2l_data_get_processor_info(
							   cs_xcp_a2l_t *data /**< XCP A2L stack structure*/ 
							   );
  
  /** Get communication mode information
      \retval Communication information
  */
  cs_xcp_a2l_commode_t cs_xcp_a2l_data_get_commode(
						   cs_xcp_a2l_t *data  /**< XCP A2L stack structure*/ 
						   );
  
  /** Reset communication mode structure
   */
  void cs_xcp_a2l_data_reset_commode(
				     cs_xcp_a2l_commode_t  *mode/**< XCP stack structure*/ 
				     );

  /** Set communication mode information based on external source
      \retval 0 No failure
      \retval 1 Failure
  */  
  int cs_xcp_a2l_data_set_commode(
				  cs_xcp_a2l_t *data,    /**< XCP stack structure*/
				  cs_xcp_a2l_commode_t *mode /**< Communication data storage structure*/
				  );

  /** Print out DAQ processor information to STDOUT
   */
  void cs_xcp_a2l_print_processor_info(
				   cs_xcp_a2l_daq_proc_t *daq_proc /**< DAQ processor information pointer*/
				   );

  /** Print out communication mode information to STDOUT
   */
  void cs_xcp_a2l_print_commode(
				cs_xcp_a2l_commode_t *commode /**< Communication information pointer*/
				);
  
  /** Print out CAN interface data to STDOUT
   */
  void cs_xcp_a2l_print_PhysInterfaceCAN(
					 cs_xcp_can_t *ifcan  /**< Interface data pointer*/
					 );

  /** Get atomic data size
      \retval Data size in bytes
  */
  int cs_xcp_a2l_get_ag_size(
			     cs_xcp_a2l_t *data            /**< XCP A2L stack structure*/ 
			     );
  
#ifdef __cplusplus
}
#endif

#endif
