/*  XCP CAN Stack for SocketCAN
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_xcp_api.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing the XCP high level API.
 *
 * This file contain the XCP high level API.
 */

#ifndef CS_XCP_API_H
#define CS_XCP_API_H

#include "cs_xcp.h"

#ifdef __cplusplus
extern "C" {
#endif

/** Load basic data from A2L file
    \retval 0 No failure
    \retval 1 failure
*/
    int cs_xcp_api_load(
        cs_xcp_t *data,                       /**< Working storage structure*/
        char * filename,                      /**< A2L filename (with path)*/
        char* project,                        /**< Project section name*/
        char* module,                         /**< Module section name*/
        char *iface,                          /**< XCP Interface name*/
        uint32_t stim                         /**< Configure message seperation time in ms*/ 
        );

/** Open physical interface
    \retval 0 No failure
    \retval 1 failure
*/
    int cs_xcp_api_open(
        cs_xcp_t *data,                       /**< Working storage structure*/
        cs_xcp_ifacetype_t iftype,            /**< Interface type*/
        char *iface                           /**< System and type dependent interface name, ex for XCP_CAN "can0"*/
        );

/** Close physical interface
    \retval 0 No failure
    \retval 1 failure
*/
    int cs_xcp_api_close(
        cs_xcp_t *data                        /**< Working storage structure*/
        );

/** Search for Signal and note it for DAQ transmission 
    \retval 0 No failure
    \retval 1 failure
*/
    int cs_xcp_api_addSignalDAQ(
        cs_xcp_t *data,          /**< Working storage structure*/
        char* signalname,        /**< Signal name to search for*/
        char *ev_name,           /**< Event channel name to register the signal*/
        int prescaler,           /**< DAQ prescaler*/
        double *exvalue,         /**< Optional: External value reference, 
                                    to store last received value,
                                    maybe NULL
                                 */
        double *exts,            /**< Optional: External timestamp reference,
                                    to store last received timestamp value,
                                    maybe NULL
                                 */
        void (*extcallback)(double value, double ts, void *priv), /**< Optional: Receive callback function for new value
                                                                     Arguments: value -> signal value
                                                                     ts    -> signal timestamp
                                                                     priv  -> private data structure,
                                                                     maybe NULL
                                                                  */
        void *priv               /**< Optional: Private data structure storage pointer, 
                                    for use in callback function call,
                                    maybe NULL
                                 */
        );

/** Connect to slave and optional get status
    \retval 0 No failure
    \retval 1 failure
*/
    int cs_xcp_api_connect(
        cs_xcp_t *data,              /**< Working storage structure*/
        cs_xcp_status_t *status      /**< Optional storage pointer for status request, maybe NULL*/
        );

/** Configure DAQ transmission, this is done via dynamic configureation
    \retval 0 No failure
    \retval 1 failure
*/
    int cs_xcp_api_configureDAQ(
        cs_xcp_t *data               /**< Working storage structure*/
        );

/** Do a synch start of all configured DAQ lists
    \retval 0 No failure
    \retval 1 failure
*/
    int cs_xcp_api_startDAQ(
        cs_xcp_t *data               /**< Working storage structure*/
        );

/** Do a stop of all configured DAQ lists
    \retval 0 No failure
    \retval 1 failure
*/
    int cs_xcp_api_stopDAQ(
        cs_xcp_t *data               /**< Working storage structure*/
        );

/** Do message processing, has to be called every cylce
    \retval 0 No failure
    \retval 1 failure
*/
    int cs_xcp_api_process(
        cs_xcp_t *data               /**< Working storage structure*/
        );

    /** Set teimout factor for DAQ ODT recv messages.
        An timeout happens when the last ODT recv timestamp  plus
        (DAQ period multiplied by fac_timeout) is lesser than actual time
        \retval 0 No timeout
        \retval 1 There was an failure or timeout
    */
    int cs_xcp_api_setDAQTimeout(
        cs_xcp_t *data,                           /**< Working storage structure*/
        double fac                                /**< Timeout factor, has to be >=1.0, < 1.0 the check is deactivated*/
        );
    
    /** Check for DAQ ODT receive timeout
        \retval 0 No timeout
        \retval 1 There was an failure or timeout
    */
    int cs_xcp_api_hasDAQTimeout(
        cs_xcp_t *data                        /**< Working storage structure*/
        );
    
#ifdef __cplusplus
}
#endif

#endif
