/*  XCP CAN Stack for SocketCAN
    Copyright (C) 2014 Andreas Stewering-Bone

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>. */

/**
 * @file cs_xcp_can.h
 * @author Andreas Stewering-Bone
 * @date 9 Oct 2014
 * @brief File containing the XCP CAN API.
 *
 * This file contain the XCP CAN low level API.
 * This functions should only be used in XCP API functions.
 * Actually the CAN API is based on SocketCAN.
*/

#ifndef CS_XCP_CAN_H
#define CS_XCP_CAN_H

#include <inttypes.h>

/** \def CS_XCP_CAN_DLC
    Maximum datalength of CAN physicallayer
*/
#define CS_XCP_CAN_DLC 8

/** \def CS_XCP_CAN_MAXDAQ
    Maximum number of user defined DAQ CAN ids
*/
#define CS_XCP_CAN_MAXDAQ 0xFF

/** \def CS_XCP_CAN_CONFDAQ
    Maximum number of user configured DAQ CAN ids
*/
#define CS_XCP_CAN_CONFDAQ (CS_XCP_CAN_MAXDAQ - 2)

/** \def CS_XCP_CAN_UNDEF
    Not defined DAQ CAN id
*/
#define CS_XCP_CAN_UNDEF 0xFFFFFFFF

/** \def CS_XCP_CAN_CHARBUF
    Charbuf size
*/
#define CS_XCP_CAN_CHARBUF 255

/** \def CS_XCP_CAN_TXFULL
    Return code, if TX queque is full
*/
#define CS_XCP_CAN_TXFULL 2

typedef enum{
  XCP_DAQ_DIR_NONE = 0,      /**< Direction undefined*/
  XCP_DAQ_DIR_DAQ,           /**< Direction Slave -> Master*/
  XCP_DAQ_DIR_STIM,          /**< Direction Master -> Slave*/
}cs_xcp_daq_dir_t;


/** \struct cs_xcp_can
    XCP CAN Storage structure, predecleartion
*/
struct cs_xcp_can;
/** \struct cs_xcp_can_t
    XCP CAN Storage structure
*/
typedef struct cs_xcp_can cs_xcp_can_t; 


#ifdef __cplusplus
extern "C" {
#endif

  /** Basic initialisation of interface data structure
      \retval not NULL No failure
      \retval NULL Failure
  */
  /*************************************************************************************************/
  cs_xcp_can_t *cs_xcp_can_init();
  /*************************************************************************************************/
  
  /** Basic configuration of interface
      \retval 0 No failure
      \retval 1 Failure
  */
  /*************************************************************************************************/
  int cs_xcp_can_configure(
			   cs_xcp_can_t *iface, /**< Interface data storage structure pointer*/ 
			   uint32_t cmd_id,     /**< Cmd target CAN id*/
			   uint32_t res_id,     /**< Response target CAN id*/
			   uint32_t bcast_id    /**< Broadcast CAN id*/
			   );
  /*************************************************************************************************/

  /** Get CAN interface version
      \retval Version
      \retval 0 Failure
  */
  /*************************************************************************************************/  
  int cs_xcp_can_getVersion(
			    cs_xcp_can_t *iface /**< Interface data storage structure pointer*/ 
			    );
  /*************************************************************************************************/

  /** Get CAN interface name
      \retval Name
      \retval NULL Failure
  */
  /*************************************************************************************************/  
  char *cs_xcp_can_getName(
			   cs_xcp_can_t *iface /**< Interface data storage structure pointer*/ 
			   );
  /*************************************************************************************************/

  /** Set CAN interface version
   */
  /*************************************************************************************************/  
  void cs_xcp_can_setVersion(
			     cs_xcp_can_t *iface, /**< Interface data storage structure pointer*/
			     int version          /**< CAN interface version*/
			     );
  /*************************************************************************************************/
  
  /** Get CAN Response ID
      \retval CAN-Id
      \retval 0 Failure
  */
  /*************************************************************************************************/  
  uint32_t cs_xcp_can_getRespId(
				cs_xcp_can_t *iface /**< Interface data storage structure pointer*/ 
				);
  /*************************************************************************************************/

  /** Set CAN Response ID
   */
  /*************************************************************************************************/  
  void cs_xcp_can_setRespId(
			    cs_xcp_can_t *iface, /**< Interface data storage structure pointer*/
			    uint32_t id          /**< CAN Id*/
			    );
  /*************************************************************************************************/

    /** Get CAN Command ID
      \retval CAN-Id
      \retval 0 Failure
    */
  /*************************************************************************************************/  
  uint32_t cs_xcp_can_getCmdId(
				cs_xcp_can_t *iface /**< Interface data storage structure pointer*/ 
				);
  /*************************************************************************************************/

  /** Set CAN Command ID
    */
  /*************************************************************************************************/  
  void cs_xcp_can_setCmdId(
			   cs_xcp_can_t *iface, /**< Interface data storage structure pointer*/
			   uint32_t id          /**< CAN Id*/
			   );
  /*************************************************************************************************/

  /** Get CAN Broadcast ID
      \retval CAN-Id
      \retval 0 Failure
    */
  /*************************************************************************************************/  
  uint32_t cs_xcp_can_getBcastId(
				 cs_xcp_can_t *iface /**< Interface data storage structure pointer*/ 
				 );
  /*************************************************************************************************/

    /** Set CAN Broadcast ID
    */
  /*************************************************************************************************/  
  void cs_xcp_can_setBcastId(
			     cs_xcp_can_t *iface, /**< Interface data storage structure pointer*/
			     uint32_t id          /**< CAN Id*/
			     );
  /*************************************************************************************************/
  
/** Open CAN interface, set filter ids
    \retval 0 No failure
    \retval 1 Failure
*/
/*************************************************************************************************/
int cs_xcp_can_open(
		    cs_xcp_can_t *iface,   /**< Interface data storage structure pointer*/ 
		    char *ifcan            /**< Interface name*/
		    );
/*************************************************************************************************/

/** Close CAN interface,
    \retval 0 No failure
    \retval 1 Failure
*/
/*************************************************************************************************/
int cs_xcp_can_close(
		     cs_xcp_can_t *iface   /**< Interface data storage structure pointer*/ 
		     );
/*************************************************************************************************/

/** Is CAN interface connected,
    \retval 0 not connected or failure
    \retval 1 connected
*/
/*************************************************************************************************/
int cs_xcp_can_isConnected(
			   cs_xcp_can_t *iface   /**< Interface data storage structure pointer*/ 
			   );

/*************************************************************************************************/

/** Set optional CAN id for DAQ and STIM packages
    \retval 0 No failure
    \retval 1 Failure
*/
/*************************************************************************************************/
int cs_xcp_can_setdaqid(
			cs_xcp_can_t *iface, /**< Interface data storage structure pointer*/ 
			int daq,             /**< DAQ number*/
			uint32_t daq_id,     /**< DAQ CAN Id, CS_XCP_CAN_UNDEF, to set only Resp id*/
			cs_xcp_daq_dir_t dir /**< DAQ direction*/
			);
/*************************************************************************************************/

/** Send prepared CAN package
    \retval 0 No failure
    \retval 1 Failure
*/
/*************************************************************************************************/
int cs_xcp_can_send(
		    cs_xcp_can_t *iface,     /**< Interface data storage structure pointer*/ 
		    uint32_t id,             /**< CAN id*/
		    uint8_t *data,           /**< Storage data array*/
		    int dl                   /**< Data length*/
		    );
/*************************************************************************************************/

/** Receive CAN package, this function is blocking until timeout is reached
    \retval 0 No failure
    \retval 1 Failure
*/
/*************************************************************************************************/
int cs_xcp_can_recv(
		    cs_xcp_can_t *iface,   /**< Interface data storage structure pointer*/ 
		    uint32_t *id,          /**< Received CAN id*/
		    uint8_t *data,         /**< Storage data array, minimum 8 Byte long*/
		    int *dl,               /**< Data length storage pointer*/
		    uint32_t timeout,      /**< Timeout in ms*/
		    struct timeval *tv_recv/**< Recv timestamp*/ 
		    );
/*************************************************************************************************/

/** Receive CAN package, this function is non blocking
    \retval 0 No failure
    \retval 1 Failure
*/
/*************************************************************************************************/
int cs_xcp_can_recv_nonblock(
			     cs_xcp_can_t *iface,   /**< Interface data storage structure pointer*/ 
			     uint32_t *id,          /**< Received CAN id*/
			     uint8_t *data,         /**< Storage data array, minimum 8 Byte long*/
			     int *dl,               /**< Data length storage pointer*/
			     struct timeval *tv_recv/**< Recv timestamp*/ 
			     );
/*************************************************************************************************/

#ifdef __cplusplus
}
#endif

#endif
